/****************************************************************************/
/*

DESCRIPTION

	Target-specific single-instruction trace breakpoint implementation for
	ARM.
	
AUTHORS

	Copyright 2002, 2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/****************************************************************************/

/*
** TRK includes
*/

#include "trk_prefix.h"
#include "usr_put.h"		// must be included before msgcmd.h
#include "target.h"
#include "step.h"
#include "break.h"
#include "targimpl.h"
#include "nubassrt.h"

/*
** Target includes
*/

#include "arm_instr.h"

bool DecodeBranch( InstructionType instruction, void *address,
				   bool *conditional_ptr, void **destination_ptr,
				   bool *link_ptr,
				   bool *delay_slot_ptr,
				   ui32 processId, ui32 threadId);

/****************************************************************************/
/*
 *	TargetSetupTrace
 *
 *	Sets up single-instruction trace.
 *	If status indicates an active step command, the appropriate
 *	step-into/step-over mode is used.  Otherwise, step-into is assumed.
 */
/****************************************************************************/
DSError TargetSetupTrace( StepStatus *status,
							ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	InstructionType instruction;
	Breakpoint *bp;
	BreakpointId bpId;
	Breakpoint *found;
	BreakpointId foundId;
	InstructionType temp;
	void *address;
	bool branch;
	bool conditional;
	bool link;
	bool bp_next;
	bool bp_dest;
	bool stepOver;
	bool delay_slot;
	void *destination;
	int next_offset;
	unsigned int temp_addr;
	
	int breakpoint_index = 0;

	__puts( "TargetSetupTrace:" );
	
	/*
	** If a host-generated step command is active, use it. 
	** Otherwise, assume step-over (it's not always safe
	** to step-into unknown code).
	*/

	stepOver = (status->active) ? status->stepOver : true;

	/*
	** Get the instruction at the current PC.  Note: The lowest bit
	** of the PC register indicates whether the instruction is a MIPS16
	** instruction.  The function XTargetGetPC() filters out that bit,
	** so use XTargetAccessExtended1Single().
	*/

	 err = XTargetAccessDefaultSingle( TRK_DEFAULT_PC, (DefaultType *)&temp_addr, true/*Ritu*/
	
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
													, processId, threadId
#endif

	address = (void *)temp_addr;

	err = XTargetAccessInstruction( &instruction, address, -1,
								kUserMemory, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								, processId, threadId
#endif
								 );

	if (err == kNoError)
	{
		/*
		** Determine if the instruction is a branch.  If so, extract various
		** branch properties needed to set trace breakpoints.
		*/

		branch = DecodeBranch( instruction, address, 
							   &conditional, &destination,
							   &link, &delay_slot,
							   processId, threadId );

		__puts( "  instr: " );
		__puthex8( (ui32)instruction.i32 );
		__puts( "\r\n" );
		__puts( "  br: " );
		__puthex2( branch );
		__puts( "  cond: " );
		__puthex2( conditional );
		__puts( "  link: " );
		__puthex2( link );
		__puts( "  dslot: " );
		__puthex2( delay_slot );
		__puts( "  dest: " );
		__puthex8( (ui32)destination );
		__puts( "\r\n" );

		bp_dest = false;
		bp_next = false;
		
		if (branch)
		{
			if (conditional || (stepOver && link))
			{
				/*
				** For conditional branches set a breakpoint
				** after the delay slot, in case it does not
				** take the brach.  Also, when stepping over
				** linking branches, set a breakpoint after
				** the delay slot.
				**
				*/
		
				next_offset =  sizeof(InstructionType);
				bp_next = true;
			}

			if (!stepOver || !link)		
			{
				/*
				** When not stepping over a linking branch, for all
				** branches set a breakpoint at the destination
				** address.
				*/
				
				bp_dest = true;
			}
							
		}
		else
		{
			/*
			** For non-branch instructions, simply set a breakpoint at
			** the next instruction.
			*/
	
			next_offset =  sizeof(InstructionType);
			bp_next = true;
	
		}
		
		/*
		** If required, set a breakpoint at the next instruction.
		*/
		if (err == kNoError && bp_next)
		{
			temp_addr = (unsigned int)address + next_offset;
			if (FindExecBreakpoint( &foundId, &found, temp_addr, 
								1, true, false, processId, threadId ) != kNoError)
			{
				err = GetFreeBreakpoint( &bpId, &bp );
				if (err == kNoError)
				{
					err = SetBreakpoint( bp, temp_addr,
										 1, false, DS_MSG_BREAK_CONDITION_EXECUTE,
										 0, 0,
										 processId, threadId );
					if (err == kNoError)
						status->traceBreakpoints[ breakpoint_index++ ] = bpId;
					else
						ReleaseBreakpoint( bpId );
				}
			}
		}

		/*
		** If required, set a breakpoint at the branch destination.
		*/
		if (err == kNoError && bp_dest)
		{
			temp_addr = (unsigned int)destination;
			if (FindExecBreakpoint( &foundId, &found, temp_addr, 
								1, true, false, processId, threadId ) != kNoError)
			{
				err = GetFreeBreakpoint( &bpId, &bp );
				if (err == kNoError)
				{
					err = SetBreakpoint( bp, temp_addr,
										 1, false, DS_MSG_BREAK_CONDITION_EXECUTE,
										 0, 0,
										 processId, threadId );
					if (err == kNoError)
						status->traceBreakpoints[ breakpoint_index++ ] = bpId;
					else
						ReleaseBreakpoint( bpId );
				}
			}
		}

	}

	return( err );
}

/****************************************************************************/
/*
 *	TargetCleanupTrace
 *
 *	Cleans up any previously setup single-instruction breakpoints.
 */
/****************************************************************************/
DSError TargetCleanupTrace( StepStatus *status,
							ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	DSError temp_err;
	Breakpoint *bp;
	
	signed int index;

	/*
	** It is important to clear the breakpoints in the opposite order
	** in which they were set.  If two breakpoints refer to the same
	** address, then the first one will have the original instruction
	** stored.  The second one will have stored the breakpoint instr.
	** from the first one.  If we restore in the opposite order, then
	** the original instruction will always be the last instruction
	** restored to that address.
	*/
	
	for (index = kTraceBreakpointCount - 1;
		 index >= 0;
		 index--)
	{
		/*
		** If there is an error, record it but continue the cleanup.
		*/
	
		if (status->traceBreakpoints[ index ] != kInvalidBreakpointId)
		{
			bp = GetBreakpoint( status->traceBreakpoints[ index ] );
			ASSERT( bp != NULL );
			temp_err = ClearBreakpoint( bp );
			ReleaseBreakpoint( status->traceBreakpoints[ index ] );
			if (temp_err != kNoError)
			{
				__puts( "TargetCleanupTrace error: " );
				__puthex8( temp_err );
				__puts( "  bp: " );
				__puthex8( index );
				__puts( "\r\n" );
			
				err = temp_err;
			}
			else 
				status->traceBreakpoints[ index ] = kInvalidBreakpointId;
		}
	}
	
	return( err );
}

DefaultType
shifted_reg_val (InstructionType instruction, int carry, DefaultType pc_val,
		 DefaultType status_reg, ui32 processId, ui32 threadId)
{
  	DefaultType res, shift, temp;
  	int rm = ARM_RM(instruction.i32);
  	DefaultType shifttype = ARM_DATA_SHIFT(instruction.i32);

  	if (instruction.i32 & 0x10)	// bit 4
    {
      	int rs = ARM_RS(instruction.i32);
      	
		XTargetAccessDefaultSingle(TRK_DEFAULT_CPSR, &temp, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									, processId, threadId
#endif
									);
										
      	shift = (rs == TRK_DEFAULT_PC ? pc_val + 8 : temp) & 0xFF;
    }
  	else
    	shift = ARM_DATA_C(instruction.i32);

	XTargetAccessDefaultSingle(rm, &temp, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								, processId, threadId
#endif
								);
								
  	res = (rm == TRK_DEFAULT_PC ? (pc_val + ((instruction.i32 & 0x10) ? 12 : 8)) : temp);

	switch (shifttype)
    {
    	case 0:			/* LSL */
      		res = shift >= 32 ? 0 : res << shift;
      		break;

   		case 1:			/* LSR */
      		res = shift >= 32 ? 0 : res >> shift;
      		break;

    	case 2:			/* ASR */
      		if (shift >= 32)
			shift = 31;
      		res = ((res & 0x80000000L) ? ~((~res) >> shift) : res >> shift);
      		break;

   	 	case 3:			/* ROR/RRX */
      		shift &= 31;
      		if (shift == 0)
				res = (res >> 1) | (carry ? 0x80000000L : 0);
     		else
				res = (res >> shift) | (res << (32 - shift));
      		break;
    }

  	return res & 0xffffffff;
}


/***************************************************************************/
/* 
 * Return number of 1-bits in VAL.
 *
 */
/***************************************************************************/
static int
bitcount (InstructionType val)
{
  int nbits;
  for (nbits = 0; val.i32 != 0; nbits++)
    val.i32 &= val.i32 - 1;		/* delete rightmost 1-bit in val */
  return nbits;
}


/****************************************************************************/
/*
 *	DecodeBranch
 *
 *	Determines whether the given instruction is a branch-type
 *	instruction.  If so, determines whether or not it is a
 *	conditional branch, the destination address, whether
 *	or not it is a linking branch, whether the destination
 *	address is MIPS16 code, and whether the branch is 
 *	followed by a delay slot.
 *
 *  If the caller does not wish to know any set of the output
 *	argument values, it may pass NULL for those arguments.
 *
 *  Args:
 *		instruction => ptr to the instruction (input argument)
 *		address => address of the instruction (input argument)
 * 		conditional_ptr => pointer to storage for conditional flag.
 *			May be NULL. (output argument)
 *		destination_ptr => pointer to storage for destination address
 *			May be NULL. (output argument)
 *		link_ptr => pointer to storage for link flag.  May be
 *			NULL. (output argument)
 *		delay_slot_ptr =>  pointer to storage for flag indicating
 *			whether the branch is followed by a delay slot.
 *			May be NULL. (output argument)
 *
 *	Returns true if the instruction is a branch-type and false
 *	otherwise.
 */
/****************************************************************************/
bool DecodeBranch( InstructionType instruction, void *address,
				   bool *conditional_ptr, void **destination_ptr,
				   bool *link_ptr,
				   bool *delay_slot_ptr,
				   ui32 processId, ui32 threadId)
{
	DSError err = kNoError;
	bool branch = false;
	bool conditional = false;
	bool link = false;
	DefaultType destination = 0;
	int cond_code;
	DefaultType temp, status;
	DefaultType pc_val = (DefaultType)address;
	
	DefaultType operand1, operand2, immval, rotate, base, offset;
	unsigned long rn;
	int c, outlength;
	InstructionType reglist;
			
	cond_code = ARM_COND(instruction.i32);
	if (cond_code == 0xF)
	{
		if (ARM_OPCODE( instruction.i32 ) == ARM_OP_BRANCH)
		{
		
			/*
			** The BLX(1) instruction is a special case.  It has a condition code of 1111,
			** but it is a non-conditional, PC-relative branch.  Get destination address
			** by combining offset encoded in instruction with PC.  Also, the LK bit in
			** this case actually becomes bit 1 of the destination addr.
			** Note: This version of the instruction always switches to Thumb mode.
			** Since the code does not yet support Thumb mode, chaos will ensue.
			*/

			destination = (DefaultType)ARM_INSTR_B_DEST( instruction.i32, address );
			if (ARM_B_LK( instruction.i32 ) != 0)
				destination |= 0x2;
			branch = true;
		}
		else
		{
			// This instruction would not be executed;
			branch = false;
			conditional = false;
		}
	}
	else 
	{
		// This is conditional if the conditional code is not 0xE;
		conditional = (cond_code != 0xE);

		err = XTargetAccessDefaultSingle(TRK_DEFAULT_CPSR, &status, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
										, processId, threadId
#endif
										);
	
		switch( ARM_OPCODE( instruction.i32 ) )
		{
			case ARM_OP_DATA_REG:
			case ARM_OP_DATA_IMM:
			
				// BX/BLX(2) instructions don't fit the pattern, so check for those first.
				// The destination comes from the register in the RM field.  Note that
				// bit 0 of the register will be set if switching to thumb mode.  However,
				// that bit should be filtered from the actual destination address.
				if (ARM_OP_BX_BLX2(instruction.i32))
				{
					branch = true;
					link = ARM_BX_LK(instruction.i32);
			    	err = XTargetAccessDefaultSingle( ARM_RM(instruction.i32), &temp, true
						#if DS_PROTOCOL == DS_PROTOCOL_RTOS
														, processId, threadId
						#endif
													);
					destination = temp & ~1;
					break;
				}
				
				if (ARM_RD(instruction.i32) != TRK_DEFAULT_PC)
					break;
				
				if (ARM_OP_MULTIPLY(instruction.i32))
					;// error ("Illegal update to pc in instruction");
					
				c = (status & ARM_FLAG_C) ? 1 : 0;
			    rn = ARM_RN(instruction.i32);

		    	err = XTargetAccessDefaultSingle( rn, &temp, true
				#if DS_PROTOCOL == DS_PROTOCOL_RTOS
													, processId, threadId
				#endif
													);

		    	operand1 = (rn == TRK_DEFAULT_PC) ? pc_val + 8 : temp;
		    	
		    	if (ARM_OPCODE(instruction.i32) == ARM_OP_DATA_IMM)
		    	{
		    		immval = ARM_DATA_IMM(instruction.i32);
		    		rotate = ARM_DATA_ROT(instruction.i32);
					operand2 = ((immval >> rotate) | (immval << (32 - rotate)))
					  & 0xffffffff;
		      	}
		      	else
			      	operand2 = shifted_reg_val (instruction, c, pc_val, status, processId, threadId);
			      	
			    switch (ARM_DATA_A(instruction.i32))
			    {
			      	case 0x0:	/*and */
						destination = operand1 & operand2;
						break;

				    case 0x1:	/*eor */
						destination = operand1 ^ operand2;
						break;

				    case 0x2:	/*sub */
						destination = operand1 - operand2;
						break;

				    case 0x3:	/*rsb */
						destination = operand2 - operand1;
						break;

				    case 0x4:	/*add */
						destination = operand1 + operand2;
						break;

				    case 0x5:	/*adc */
						destination = operand1 + operand2 + c;
						break;

				    case 0x6:	/*sbc */
						destination = operand1 - operand2 + c;
						break;

				    case 0x7:	/*rsc */
						destination = operand2 - operand1 + c;
						break;

				    case 0x8:
				    case 0x9:
				    case 0xa:
				    case 0xb:	/* tst, teq, cmp, cmn */
						destination = (DefaultType) (pc_val + 4);
						break;

				    case 0xc:	/*orr */
						destination = operand1 | operand2;
						break;

				    case 0xd:	/*mov */
						/* Always step into a function.  */
						destination = operand2;
						break;

				    case 0xe:	/*bic */
						destination = operand1 & ~operand2;
						break;

				    case 0xf:	/*mvn */
						destination = ~operand2;
						break;
			    }
			    
				branch = true;
				break;
			
			case ARM_OP_SINGLE_REG:
			case ARM_OP_SINGLE_IMM:
			
				if (ARM_RD(instruction.i32) != TRK_DEFAULT_PC || !ARM_LOAD(instruction.i32))
					break;
				
				if (ARM_SINGLE_BYTE(instruction.i32))
					;// error ("Illegal update to pc in instruction");
					
			    rn = ARM_RN(instruction.i32);

		    	err = XTargetAccessDefaultSingle( rn, &temp, true
				#if DS_PROTOCOL == DS_PROTOCOL_RTOS
													, processId, threadId
				#endif
													);

		    	base = (rn == TRK_DEFAULT_PC) ? pc_val + 8 : temp;
		    	
		    	if (ARM_SINGLE_PRE(instruction.i32))
		    	{
		    		// Pre-insdexing
					c = (status & ARM_FLAG_C) ? 1 : 0;
		    		if (ARM_OPCODE(instruction.i32) == ARM_OP_SINGLE_IMM)
			    		offset = ARM_SINGLE_IMM(instruction.i32);
			      	else
				      	offset = shifted_reg_val (instruction, c, pc_val, status, processId, threadId);
				      	
				    if (ARM_SINGLE_U(instruction.i32))
				    	base += offset;
				    else
				    	base -= offset;
				}
				 
		 		outlength = 4;
		 		err = OsTargetAccessMemory( &destination,(void*)base, &outlength,
											kUserMemory, true,
											processId, threadId);/*Ritu*/
				branch = true;
				break;
				
			case ARM_OP_BLOCK:
				
				// If this is not a load instruction or PC is not in the register list, continue.
				if (!ARM_LOAD(instruction.i32) || !((ui32)instruction.i32 & 0x00008000L))
					break;
					
				if (ARM_BLOCK_U(instruction.i32))
				{
					reglist.i32 = ARM_BLOCK_REGLIST(instruction.i32);
					offset = (bitcount(reglist) - 1) * 4;
					if (ARM_BLOCK_PRE(instruction.i32))
						offset += 4;
				}
				else if (ARM_BLOCK_PRE(instruction.i32))
					offset = -4;
					
		    	err = XTargetAccessDefaultSingle( ARM_RN(instruction.i32), &temp, true
				#if DS_PROTOCOL == DS_PROTOCOL_RTOS
													, processId, threadId
				#endif
													);
				
				temp += offset;
		 		outlength = 4;				
		 		err = OsTargetAccessMemory( &destination,(void *)temp, &outlength,
											kUserMemory, true,
											processId, threadId);
		
				branch = true;
				break;
				
			case ARM_OP_BRANCH:
				/*
				** PC-relative branches.  Get destination address
				** by combining offset encoded in instruction with
				** PC.  These are conditional.
				*/

				destination = (DefaultType)ARM_INSTR_B_DEST( instruction.i32, address );
				link = ARM_B_LK( instruction.i32 ) != 0;
				branch = true;
				break;
		}
	}

	if (conditional_ptr != NULL)
		*conditional_ptr = conditional;
		
	if (destination_ptr != NULL)
		*destination_ptr = (void *)destination;/*Ritu*/
		
	if (branch && !link)
	{
		// Make sure if this branch is a linking branch.
		XTargetAccessDefaultSingle(TRK_DEFAULT_LR, &temp, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									, processId, threadId
#endif
									);
		// If the next instruction's address after branch instruction is stored in FP register before branch,
		// This branch is a linking branch, 
		if (temp == (DefaultType)address + sizeof(InstructionType))/*Ritu*/
			link = true;
	}

	if (link_ptr != NULL)
		*link_ptr = link;

	return( branch );
}

