/******************************************************************************************/
/*
	
DESCRIPTION

	Communication services under Linux.  Currently CodeWarriorTRK assumes
	a UART driver API, so this code conforms to that API for both serial
	and ethernet communication.

COPYRIGHT

	Copyright 2000-2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"
#include "usr_put.h"
#include "dstypes.h"
#include "dserror.h"
#include "linux_comm.h"
#include "linux_main.h"
#include "linux_target.h"
#include "msgcmd.h"
#include "nubassrt.h"
#include "transport.h"
#include "rfc1662.h"
#include "ethconfig.h"
#include "targimpl.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <poll.h>
#include <termios.h>
#include <signal.h>

/****************************************************************************/
/*
	Global variables
*/
/****************************************************************************/
int gCommInFd = -1;				/* input file descriptor for the debug connection */
int gCommOutFd = -1;			/* output file descriptor for the debug connection */
static char *gCommSpec = NULL;	/* string specification for the debug connection */
bool gPersistent = true;		/* try to survive broken connections? */
bool gCloseOnRead0 = true;		/* close connection when read returns 0? */

/*
** TRK defines a maximum message size.  The checksum could add up to 4 bytes.
** Each byte in the message and checksum could potentially be escaped and
** consume 2 bytes in the write buffer.  Finally, the start and end frame flags
** require 1 byte each.
*/
#define COMM_BUF_SIZE	(((DS_MAXMESSAGESIZE + 4) * 2) + 2)

static char gWriteBuf[ COMM_BUF_SIZE ];
static int gWritePos = 0;
char gReadBuf[ COMM_BUF_SIZE ];
int gReadPos = 0;
int gReadCount = 0;

/****************************************************************************/
/* 
 *	Define character-based transport layer.  For linux, this file
 *	implements that layer.
 */
/****************************************************************************/
static DSError InitCommLinux( LayerParams *params );
static DSError TerminateCommLinux( LayerParams *params );
static DSError ReadCommLinux( LayerParams *params, char* c);
static DSError FlushCommLinux( LayerParams *params );
static DSError WriteCommLinux( LayerParams *params, char c );
static CharLayerApi transportLinuxApi =
{
	InitCommLinux,				// init
	TerminateCommLinux,			// terminate
	ReadCommLinux,				// read
	WriteCommLinux,				// write
	FlushCommLinux,				// flush
	NULL						// irq
};
static CharLayer transportLinuxLayer =
{
	NULL,						/* no params for this layer */
	&transportLinuxApi,			/* api implementation for this layer */
};

/****************************************************************************/
/* 
	Define framing layer & parameters (used to frame messages).
	CodeWarriorTRK interfaces directly with this layer.  This layer uses
	the character-based layer for the actual communication.
*/
/****************************************************************************/
static Rfc1662Params transportRfc1662Params =
{
	&transportLinuxLayer		/* pointer to character-based sub-layer */
};
static MsgLayer transportRfc1662Layer =
{
	&transportRfc1662Params,	/* params for this layer */
	&gRfc1662Api,				/* api implementation for this layer */
};

/****************************************************************************/
/* Define the number of available communication ports                       */
/****************************************************************************/
#define TRK_TRANSPORT_NUM_COMM_PORTS			1
unsigned char gTrkTransportNumCommPorts			= TRK_TRANSPORT_NUM_COMM_PORTS;

/****************************************************************************/
/*
 * Define the number of communication ports which can remain active
 * at one time.
 */
/****************************************************************************/
unsigned char gTrkTransportNumCommPortsActive	= 1;

/****************************************************************************/
/*
 * gTransportPrimaryIndex is the index of the primary port.
 * Primary port is the fastest available port which CodeWarriorTRK will
 * use to send big messages (length > TRK_TRANSPORT_MSG_THRESHOLD (0x1000)).
 *
 * For example, if CodeWarriorTRK supports two types of communications ports:
 * serial (index = 0) and ethernet (index == 1), then
 * gTransportPrimaryIndex should be defined as 1.
 */
/****************************************************************************/
unsigned char gTrkTransportPrimaryIndex			= 0;

/****************************************************************************/
/*
 *	Define transport table
 *	
 *	In this case, there is only a single entry consisting of the rfc1662
 *	framing layer.  There is no direct reference to the char-based interface
 *	at this level.  However, the parameters for the framing layer point
 *	to the char-based layer underneath (see above).
 */
/****************************************************************************/
Transport gTransportTable[TRK_TRANSPORT_NUM_COMM_PORTS] =
{
	{
		true,						/* active */
		false,						/* used */
		kSerial,					/* port type */
		kChannel_0,					/* channel number */
		&transportRfc1662Layer		/* pointer to top communication layer */
	}
};

/****************************************************************************/
/*
	TrkSetComm
	
	Set the communication channel.  This must be called before InitCommLinux()
	or any of the other transport API functions in this file.
*/
/****************************************************************************/
DSError TrkSetComm( char *spec )
{
	/*
	** Free the previously-stored specification, if any.
	*/
	if (gCommSpec != NULL)
		free( gCommSpec );
	gCommSpec = strdup( spec );
	return( kNoError );
}

/****************************************************************************/
/*
	OpenTcpComm
	
	Open a TCP connection as a server on the specified port.
	Returns the file descriptor for the new connection, or -1 on failure.
*/
/****************************************************************************/
static int OpenTcpComm( int port )
{
	bool err = false;
	bool connected = false;
	int fd = -1;
	int sd = -1;
	struct sockaddr_in saddr;
	int optval;
	struct sigaction act;
	
	/*
	** Get the initial socket descriptor.
	*/
	err = (-1 == (sd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP )));
	if (err)
		__perror( "Failed to open socket" );
	
	/*
	** Bind to the specified port.
	*/
	if (!err)
	{
		/*
		** This option permits reuse of the port after the connection
		** is closed.  This must be done before bind().
		*/
		optval = 1;
		setsockopt( sd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof( optval ) );
		
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons( port );
		saddr.sin_addr.s_addr = INADDR_ANY;
		err = (-1 == (bind( sd, (struct sockaddr *)&saddr, sizeof( saddr ) )));
		if (err)
			__perror( "Failed to bind socket" );
	}
	/*
	** Before forking child processes, explicity ignore SIGCHLD signals.
	** This will prevent zombies from accumulating without requiring
	** wait().
	*/
	
	sigemptyset( &act.sa_mask );
	act.sa_flags = 0;
	act.sa_handler = SIG_IGN;
	sigaction( SIGCHLD, &act, NULL );
	/*
	** The initial TRK process simply loops, listening for connections.
	** On each new connection, fork a new TRK process to use that
	** connection.  Each child will exit this loop with the connected
	** flag set to true.  The initial process should only exit this loop
	** if there is an error.
	*/
#ifndef __uClinux__	
	while (!err && !connected)
	{
#endif
		/*
		** Listen for a connection from the debugger.  This will block
		** until the debugger attempts to connect.
		*/
		
		if (!err)
		{
			err = (-1 == listen( sd, 1 ));
			if (err)
				__perror( "Failed to listen for connection" );
		}
		
		/*
		** Accept the connection from the debugger.
		*/
		
		if (!err)
		{
			err = (-1 == (fd = accept( sd, NULL, 0 )));
			if (err)
				__perror( "Failed to accept connection" );
		}
		
#ifndef __uClinux__	
		if (!err)
		{
			switch (fork())
			{
				case -1: /* error */
					err = true;
					__perror( "Failed to fork process for new connection" );
					break;
					
				case 0:	 /* child */
					connected = true;
					gPersistent = false; /* the child will die when the connection breaks */
					gCloseOnRead0 = true; /* read returns 0 if connection closed */
					break;
					
				default: /* parent */
					close( fd ); /* only the child needs this */
					break;
			}
		}
	}
#endif
	/*
	** Set appropriate options on the new connection.
	*/
	
	if (!err)
	{
		/*
		** Use keepalive probes to detect long-dead connections.
		*/
		optval = 1;
		setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof( optval ) );
		
		/*
		** This option eliminates delay when sending small packets.
		*/
		optval = 1;
		setsockopt( fd, SOL_TCP, TCP_NODELAY, &optval, sizeof( optval ) );
		
		/*
		** Use non-blocking I/O.
		*/
		
		fcntl( fd, F_SETFL, O_NONBLOCK );
	}
	/*
	** The original socket descriptor won't be reused, so close it now.
	*/
		
	if (sd != -1)
		close( sd );
		
	return( fd );
}

/****************************************************************************/
/*
	OpenSerialComm
	
	Open a serial connection on the specified device.
	Returns the file descriptor for the new connection, or -1 on failure.
*/
/****************************************************************************/
static int OpenSerialComm( char *dev )
{
	bool err;
	int fd = -1;
	struct termios tios;
	
	err = (-1 == (fd = open( dev, O_RDWR | O_NONBLOCK )));
	if (err)
		__perror( "Failed to open serial device" );
	/*
	** Set the appropriate attributes.  Standard for CodeWarriorTRK: ignore
	** BREAK, raw IO, 8 data bits, no parity, 1 stop bit, hardware
	** flow control if possible.  Don't wait for input if it isn't
	** available.
	*/
		 
	if (!err)
	{
		
		/*
		** For a serial connection, read often returns 0.  Don't close the
		** connection when that happens.
		*/
		gCloseOnRead0 = false;
	
		err = (-1 == tcgetattr( fd, &tios ));
		
		if (!err)
		{
			tios.c_iflag &= ~(BRKINT | ICRNL | IGNCR | IMAXBEL | INLCR | INPCK |
					ISTRIP | IUCLC | IXANY | IXOFF | IXON | PARMRK);
			tios.c_iflag |= IGNBRK | IGNPAR;
			tios.c_oflag &= ~(BSDLY | CRDLY | FFDLY | NLDLY | OCRNL | OFDEL | 
					OFILL | OLCUC | ONLCR | ONLRET | ONOCR | OPOST |
					TABDLY | VTDLY);
			tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB);
			tios.c_cflag |= CLOCAL | CREAD | CS8 | HUPCL;
			tios.c_lflag &= ~(ECHO | ECHOCTL | ECHOE | ECHOK | ECHOKE |
					ECHONL | ECHOPRT | FLUSHO | ICANON | IEXTEN | ISIG | NOFLSH | 
					PENDIN | TOSTOP | XCASE);
			tios.c_cc[VMIN] = 0;
			tios.c_cc[VTIME] = 0;
			
			err = (-1 == tcsetattr( fd, TCSANOW, &tios ));
		}
		
		if (err)
			__perror( "Failed to set serial attributes" );
	}
	
	/*
	** Cleanup.
	*/
	
	if (err && fd != -1)
		close( fd );
	return( fd );
}

/****************************************************************************/
/*
	InitCommLinux
	
	Initialize the communication channel.
	For Linux, this opens the communication port as specified by the global
	gCommSpec.  If gCommSpec specifies a serial device, the current baud rate
	for that device will not be changed.
*/
/****************************************************************************/
static DSError InitCommLinux( LayerParams *params )
{
	char *s;
	
	/*
	** Open the new connection.  If the spec is simply a '|' character,
	** use stdin/stdout as pipes to the debugger.  If the spec contains
	** a ':' character, then it is a TCP connection spec in the form
	** :<tcp port>.  Otherwise it should be the name of a serial device. 
	*/
	
	if (gCommSpec[0] == '|')
	{
		/*
		** Use non-blocking I/O.
		*/
		
		fcntl( STDOUT_FILENO, F_SETFL, O_NONBLOCK );
		fcntl( STDIN_FILENO, F_SETFL, O_NONBLOCK );
		gPersistent = false;
		gCloseOnRead0 = true; /* read returns 0 if connection closed */
		gCommInFd = STDIN_FILENO;
		gCommOutFd = STDOUT_FILENO;
	}
	else
	{
		s = strchr( gCommSpec, ':' );
		
		if (s != NULL)
			gCommInFd = gCommOutFd = OpenTcpComm( atoi( s + 1 ) );
		else
		
			gCommInFd = gCommOutFd = OpenSerialComm( gCommSpec );
	}
		
	if (gCommInFd == -1 || gCommOutFd == -1)
	{
		perror( "Failed to open debug connection" );
		return( kTransportConfigError );
	}
	else
	{
		printf( "Connected on %s\n", gCommSpec );
	}
		
	return( kNoError );
}

/****************************************************************************/
/*
	TerminateCommLinux
	
	Close the communication channel.
*/
/****************************************************************************/
static DSError TerminateCommLinux( LayerParams *params )
{
	/*
	** Close the previous connection, if any.
	*/
	
	if (gCommInFd != -1)
		close( gCommInFd );
	if (gCommOutFd != -1)
		close( gCommOutFd );
	printf( "Disconnected\n" );
	return kNoError;
}

/****************************************************************************/
/*
 *	TrkReconnectComm
 *
 *	Reconnect the communication link if it is broken.
 */
/****************************************************************************/
#ifdef __uClinux__
extern int wakePipe[2];
extern void SigWake( int s /*, siginfo_t *i, void *p */);
#endif
static DSError TrkReconnectComm( LayerParams *params )
{
#ifdef __uClinux__
	struct sigaction act;
	sigset_t sset;
	int i;
#endif	
	/*
	** Shutdown and re-initialize.  This is probably overkill, but why not.
	*/
	TerminateCommLinux( params );
#ifndef __uClinux__	
	return( InitCommLinux( params ) );
#else
	InitCommLinux( params );
	
	/*
	** This should prevent us from dying from real-time signals when
	** debugging threaded apps.
	*/
	
	sigemptyset( &sset );
	for (i = SIGRTMIN; i <= SIGRTMAX; i++)
		sigaddset( &sset, i );
	sigprocmask( SIG_BLOCK, &sset, NULL );
	
	/*
	** Setup SIGCHLD signals to wakeup TRK when it is idle.  This will
	** ensure that the select() routine (used by TrkSelectIo() in Idle())
	** will return when a child process stops or dies.
	**
	** Open the wakeup pipe and set both ends to non-blocking mode.
	*/
	pipe( wakePipe );
	fcntl( wakePipe[0], F_SETFL, O_NONBLOCK );
	fcntl( wakePipe[1], F_SETFL, O_NONBLOCK );
	sigemptyset( &act.sa_mask );
	act.sa_flags = 0;
	act.sa_handler = SigWake;
	sigaction( SIGCHLD, &act, NULL );
	
	/*
	** Don't die when pipes break.
	*/
	signal( SIGPIPE, SIG_IGN );
	
	return( kNoError);	
#endif
}

/****************************************************************************/
/*
	ReadCommLinux
	
	Poll for one character from the comm. channel.  If none are available,
	return immediately.
	
*/
/****************************************************************************/
static DSError ReadCommLinux( LayerParams *params, char* c)
{
	/*
	** If there is nothing left in the read buffer, try to read more.
	*/
	
	if (gReadPos >= gReadCount)
	{
		gReadPos = 0;
		errno = 0;
		gReadCount = read( gCommInFd, gReadBuf, COMM_BUF_SIZE );
		
		if ((gReadCount < 0 && errno != EAGAIN) ||
			(gReadCount == 0 && gCloseOnRead0))
		{
			/*
			** If the connection is broken: re-connect if persistent, otherwise
			** cleanup and die.
			*/
			if (gPersistent)
			{
				TrkReconnectComm( params );
			}
			else
			{
				__puts( "Connection broken.  Exiting...\r\n" );
				CleanupTarget();
				exit( 0 );
			}
			/*
			** SIGIO doesn't seem to be reliable on PPC unless
			** we do at least one read after reconnect.
			*/
			gReadCount = read( gCommInFd, gReadBuf, COMM_BUF_SIZE );
		}
	}
	/*
	** If there is now something in the read buffer, get the next
	** character and update the buffer position.
	*/
	if (gReadPos < gReadCount)
	{
		*c = gReadBuf[ gReadPos++ ];
		return kNoError;
	}
	
	/*
	** If an actual error occurs, it is currently indistinguishable from
	** a lack of input.
	*/
	return( kTransportNoData );
}

/****************************************************************************/
/*
	FlushCommLinux
	
	If output is buffered, force it out now.
*/
/****************************************************************************/
static DSError FlushCommLinux( LayerParams *params )
{
	DSError err = kNoError;
	bool retry;
	if (gWritePos != 0)
	{
		/*
		** Keep writing until it's all gone.  A poll() or select() call should
		** probably be added here somewhere...
		*/
		do
		{
			retry = false;
			errno = 0;
			if (write( gCommOutFd, gWriteBuf, gWritePos ) < 0)
			{
				if (errno != EAGAIN)
				{
					__perror( "Write failed" );
					err = kTransportError;
				}
				else
				{
					retry = true;
				}
			}
		} while (retry);
		
		gWritePos = 0;
	}
	
	return( err );
}

/****************************************************************************/
/*
	WriteCommLinux
	
	Write one character to the communication channel.
*/
/****************************************************************************/
static DSError WriteCommLinux( LayerParams *params, char c )
{
	ASSERT( gWritePos < COMM_BUF_SIZE );
	gWriteBuf[ gWritePos++ ] = c;
	return( kNoError );
}

/****************************************************************************/
/*
** Function Name:		TrkSetEthConfig
**
** Description:			Sets the current ethernet configuration.
**
** Input:
**		config:			Ptr to the new configuration values.
**
** Output:
**		return:			error code (kNoError if succeeded)
**
*/
/****************************************************************************/
DSError TrkSetEthConfig( const TrkEthConfig *config )
{
	/* Stub.  Not implemented for linux. */
	return( kUnsupportedError );
}
