/******************************************************************************************/
/*
	
DESCRIPTION

	Linux ELF helper routines.

COPYRIGHT

	Copyright 2001-2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "nubassrt.h"
#include "target.h"
#include "targimpl.h"
#include "linux_elf.h"

/*
** Linux includes
*/
#include <string.h>
#include <fcntl.h>

/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/


/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/


/****************************************************************************/
/*
	Function Definitions
*/
/****************************************************************************/

/****************************************************************************/
/*
	TrkOpenElfLinux
	
	Opens the given file and obtains an ELF handle for it.
	
	Returns kNoError and fills in the fields of tlElf if succesful.
	
*/
/****************************************************************************/

DSError TrkOpenElfLinux( char *name, TrkLinuxElf *tlElf )
{
	DSError err = kNoError;
	
	ASSERT( tlElf != NULL );

	/*
	** Open the given file in read-only mode.
	*/
	
	__puts( "Opening ELF file " );
	__puts( name );
	__puts( "\r\n" );
	tlElf->fFd = open( name, O_RDONLY );
	if (tlElf->fFd == -1)
	{
		err = kOsError;
		__perror( "Failed to open elf file" );
	}
	
	/*
	** Get an ELF handle from the libelf library.
	*/
	
	if (err == kNoError)
	{
		/*
		** Must set the version before using libelf functions.
		*/
		
		elf_version( EV_CURRENT );
		
		tlElf->fElf = elf_begin( tlElf->fFd, ELF_C_READ, NULL );
		if (tlElf->fElf == NULL)
		{
			err = kOsError;
			__elf_perror( "Failed to get ELF handle" );
		}
	}
	
	return( err );
}

/****************************************************************************/
/*
	TrkCloseElfLinux
	
	Closes the given file and releases its ELF handle.
	
*/
/****************************************************************************/

DSError TrkCloseElfLinux( TrkLinuxElf *tlElf )
{
	ASSERT( tlElf != NULL );
	
	elf_end( tlElf->fElf );
	close( tlElf->fFd );
	
	return( kNoError );
}

/****************************************************************************/
/*
	TrkGetElfSectionRangeLinux
	
	Get the memory range for the specified section from the given ELF handle.
*/
/****************************************************************************/
void TrkGetElfSectionRangeLinux( Elf *elf, char *name, void *base,
									PCType *secStart, PCType *secEnd )
{
	Elf_Scn *esec=NULL;
	Elf32_Shdr *eshdr=NULL;
	size_t size;

	esec = TrkFindElfSectionLinux( elf, name );
	if (esec != NULL)
		eshdr = elf32_getshdr( esec );
		
	if (eshdr != NULL)
	{
		size = eshdr->sh_size;
		*secStart = (PCType)(eshdr->sh_addr + (ui8 *)base);
		*secEnd = (PCType)((ui8 *)(*secStart) + size - 1);
		__puthex8( *secStart );
		__puts( "-" );
		__puthex8( *secEnd );
		__puts( "\r\n" );
	}
}
									
/****************************************************************************/
/*
	TrkFindElfSectionLinux
	
	Finds the section with the given name in the given ELF file.
	
*/
/****************************************************************************/

Elf_Scn *TrkFindElfSectionLinux( Elf *elf, char *name )
{
	DSError err = kNoError;
	Elf32_Ehdr *ehdr;
	Elf32_Shdr *eshdr;
	Elf_Scn *esec = NULL;
	char *esName;
	int i;
	bool found = false;

	/*
	** Get the ELF header, which tells us how many sections there
	** are and which section number contains the section names.
	*/

	ehdr = elf32_getehdr( elf );
	if (ehdr == NULL)
	{
		err = kOsError;
		__elf_perror( "Failed to get ELF header" );
	}
	
	/*
	** Simple iteration over all sections comparing the names.
	*/

	for (i = 0; i < ehdr->e_shnum && !found && err == kNoError; i++)
	{
		/*
		** Get the section corresponding to the current index.
		*/
		esec = elf_getscn( elf, i );
		if (esec == NULL)
		{
			err = kOsError;
			__elf_perror( "Failed to get ELF section" );
		}
			
		/*
		** Get the header for the current section.
		*/
		if (err == kNoError)
		{
			eshdr = elf32_getshdr( esec );
			if (eshdr == NULL)
			{
				err = kOsError;
				__elf_perror( "Failed to get ELF section header" );
			}
		}
		
		/*
		** Get the name of the current section.
		*/
		if (err == kNoError)
		{
			esName = elf_strptr( elf, ehdr->e_shstrndx, eshdr->sh_name );
			if (esName == NULL)
			{
				err = kOsError;
				__elf_perror( "Failed to get ELF section name" );
			}
		}
		
		/*
		** Compare the current section name with the requested name.
		*/
		
		if (err == kNoError)
		{
			if (strcmp( esName, name ) == 0)
				found = true;
		}
	}
	
	return( (found) ? esec : NULL );
}

/****************************************************************************/
/*
	TrkFindElfSymbolLinux
	
	Finds the symbol with the given name in the given ELF file.
	
*/
/****************************************************************************/

Elf32_Sym *TrkFindElfSymbolLinux( Elf *elf, const char *name )
{
	DSError err = kNoError;
	Elf_Scn *strtabScn;
	Elf_Scn *symtabScn;
	size_t strtabI;
	Elf_Data *edata;
	Elf32_Sym *esyms;
	Elf32_Sym *result = NULL;
	char *symName;
	int count;
	int i;
	bool found = false;

	/*
	** Locate the ".strtab" and ".symtab" sections.  These contain
	** the symbol names and attributes, respectively.
	*/
	
	strtabScn = TrkFindElfSectionLinux( elf, ".strtab" );
	symtabScn = TrkFindElfSectionLinux( elf, ".symtab" );
	if (strtabScn == NULL || symtabScn == NULL)
	{
		err = kUnsupportedError;
		__elf_perror( "No symbols found in ELF file" );
	}
	
	/*
	** For the ".strtab" section, get the index.  For the ".symtab"
	** section, get the contents.
	*/
	
	if (err == kNoError)
	{
		strtabI = elf_ndxscn( strtabScn );
		
		edata = elf_getdata( symtabScn, NULL );
		if (edata == NULL)
		{
			err = kOsError;
			__elf_perror( "Failed to get ELF symbol table data" );
		}
	}

	/*
	** Simple iteration over symbol table, comparing names.
	*/
	
	if (err == kNoError)
	{
		esyms = edata->d_buf;
		count = edata->d_size / sizeof( *esyms );

		for (i = 0; i < count && !found && err == kNoError; i++)
		{
			symName = elf_strptr( elf, strtabI, esyms[i].st_name );
			if (symName == NULL)
			{
				err = kOsError;
				__elf_perror( "Failed to get ELF symbol name" );
			}
			
			if (err == kNoError && strcmp( symName, name ) == 0)
			{
				found = true;
				result = &(esyms[i]);
			}
		}
	}

	return( (found) ? result : NULL );
}


/****************************************************************************/
/*
	TrkGetElfSymbolOffset
	
	Gets the offset of the given symbol relative to the beginning of the ELF
	object.
		
*/
/****************************************************************************/

Elf32_Addr TrkGetElfSymbolOffset( Elf *elf, Elf32_Sym *esym )
{
	return( esym->st_value );
}
