/******************************************************************************************/
/*
	
DESCRIPTION

	Linux share-object handling routines.

COPYRIGHT

	Copyright 2000-2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "linux_so.h"
#include "targimpl.h"
#include "break.h"
#include "nubassrt.h"
#include "linux_elf.h"
#include "linux_notify.h"		// for DoOsNotifyDllCreated()

/*
** Linux includes
*/
#include <link.h>
#include <string.h>
#include <stdlib.h>		// for malloc
#include <linux/limits.h>


/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

#define MAX_FILENAME_SIZE	1024

static char *solib_break_names[] =
{
  "r_debug_state",
  "_r_debug_state",
  "_dl_debug_state",
  "rtld_db_dlactivity",
  NULL
};


/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/

/****************************************************************************/
/*
	Function Definitions
*/
/****************************************************************************/

/****************************************************************************/
/*
	TrkFindSoDebugPtrLinux
	
	Set breakpoints needed to detect shared-object events.
	
	Returns kNoError if successful, kUnsupportedError if not enough information
	was available (perhaps this is not a dynamically-linked executable?), or
	kOsError for unexpected errors returned by the OS.
*/
/****************************************************************************/

static DSError TrkFindSoDebugPtrLinux( ui32 processId, struct r_debug **soDebugPtr )
{
	DSError err = kNoError;
	char exePath[ PATH_MAX ];
	TrkLinuxElf tlElf;
	Elf_Scn *esec;
	Elf32_Shdr *eshdr = NULL;
	ElfW(Dyn) *dynamicBase;
	ElfW(Dyn) *dynamicBuf = NULL;
	int i;
	size_t size;
#if FOR_MIPS_ARCHITECTURE	
	void *temp;
#endif
	
	/*
	** This assumes ELF.
	*/

	tlElf.fFd = -1;
	
	/*
	** Get an ELF handle for the executable.  There is a link to it in 
	** "/proc/##/exe", where ## is the process ID.
	*/
	sprintf( exePath, "/proc/%lu/exe", processId );
	err = TrkOpenElfLinux( exePath, &tlElf );
		
	/*
	** Get the size and location of the .dynamic section and read its
	** contents from the target process.  Then search through the .dynamic
	** section for an entry with the DT_DEBUG tag.  That will point us to
	** the dynamic loader's r_debug structure.
	*/

	if (err == kNoError)
	{
		esec = TrkFindElfSectionLinux( tlElf.fElf , ".dynamic" );
		if (esec != NULL)
			eshdr = elf32_getshdr( esec );

		if (esec == NULL || eshdr == NULL)
			err = kUnsupportedError;
	}
	
	if (err == kNoError)
	{
		size = eshdr->sh_size;
		dynamicBase = (ElfW(Dyn) *)eshdr->sh_addr;
		dynamicBuf = (ElfW(Dyn) *)malloc( size );
		if (dynamicBuf == NULL)
			err = kOsError;
	}

	if (err == kNoError)
	{
		/*
		** Read the .dynamic section into a local buffer.
		*/
		__puts( ".dynamic section found at " );
		__puthex8( (ui32)dynamicBase );
		__puts( "\r\n" );
		err = OsTargetAccessMemory( dynamicBuf,
			dynamicBase,
			&size, kDebuggerMemory, true,
			processId, 0 );
	}

	/*
	** Search the .dynamic section for the DT_DEBUG tag.
	*/
	if (err == kNoError)
	{
		err = kUnsupportedError;
		for (i = 0; dynamicBuf[ i ].d_tag != DT_NULL; i++)
		{
#if FOR_MIPS_ARCHITECTURE	
			if (dynamicBuf[ i ].d_tag == DT_MIPS_RLD_MAP ) 
			{
				temp = (struct r_debug *)dynamicBuf[ i ].d_un.d_ptr;
				size = sizeof( *soDebugPtr );
				err = OsTargetAccessMemory( soDebugPtr, temp, &size,
					kDebuggerMemory, true, processId, 0 );
				__puts( "r_debug struct found at " );
				__puthex8( (ui32)*soDebugPtr );
				__puts( "\r\n" );

				break;
			}
#else
			if (dynamicBuf[ i ].d_tag == DT_DEBUG)
			{
				err = kNoError;
				*soDebugPtr = (struct r_debug *)dynamicBuf[ i ].d_un.d_ptr;
				__puts( "r_debug struct found at " );
				__puthex8( (ui32)*soDebugPtr );
				__puts( "\r\n" );

				break;
			}
#endif
		}
	}
	
	/*
	** If the value is 0, then the entry in the .dynamic section has not yet 
	** been initialized and we don't have enough information.
	*/
	if (*soDebugPtr == 0)
		err = kUnsupportedError;
	
	/*
	** Cleanup.
	*/
	if (dynamicBuf != NULL)
		free( dynamicBuf );
		
	if (tlElf.fFd != -1)
		TrkCloseElfLinux( &tlElf );
		
	return( err );
}



/****************************************************************************/
/* TrkTrimInvisibleRangeList
	This function trims invisible range list such that only first four 
	nodes(which corr. to main's .plt section and loader's .plt,.got amd .text section)
	are left in the list.All other nodes are deleted from the list.
	Count is the total number of nodes present in the list.
*/
/****************************************************************************/
void TrkTrimInvisibleRangeList(ui32 *Count,InvisibleRangeInfo **InvRangeList)
{
	InvisibleRangeInfo *temp;
	ui32 Cnt;
	for (Cnt = *Count;Cnt > kSectionCount;Cnt--)
	{
		temp = *InvRangeList;
		*InvRangeList = (*InvRangeList)->next;
		free(temp);
		(*Count)-=1;
	}
}

/****************************************************************************/
/* TrkFreeInvisibleRangeList
	Frees memory for Invisible step  range list.
*/
/****************************************************************************/

void TrkFreeInvisibleRangeList(InvisibleRangeInfo **InvRangeList)
{
	InvisibleRangeInfo *temp;
	temp = *InvRangeList;
	while (*InvRangeList != NULL)
	{
		*InvRangeList = temp->next;
		free(temp);
		temp = *InvRangeList;
	}
}


/****************************************************************************/
/* TrkAllocInvisibleRangeNode
	Allocates memory for Invisible step  range list.
*/
/****************************************************************************/

static DSError TrkAllocInvisibleRangeNode(InvisibleRangeInfo **tempInvRangeList)
{
	*tempInvRangeList = (InvisibleRangeInfo *)malloc(sizeof(InvisibleRangeInfo));
	if (*tempInvRangeList == NULL)
		return kOsError;
	(*tempInvRangeList)->InvisibleStepRanges.fStart =(PCType)NULL;
	(*tempInvRangeList)->InvisibleStepRanges.fEnd = (PCType)NULL;
	(*tempInvRangeList)->next = NULL;	
	return kNoError;
}
 
/****************************************************************************/
/*
	TrkFindSoLoaderInfo
	
	Find the symbol that points to the break location needed to detect
	shared-object events, along with the memory ranges for the ".plt"
	section of the main program, the ".plt" section of the loader, and the
	".text" section of the loader.
	
	If the base address of the shared-object loader is known, it should be
	passed in via the soLoaderBase parameter.  Otherwise, that parameter
	should be NULL.  If NULL, then this routine will determine the base with
	the assumption that the current PC is at the entry point to the loader.
	If that assumption is not valid, do not call this function with NULL for that
	parameter.
	
	Returns kNoError if successful, kUnsupportedError if not enough information
	was available (perhaps this is not a dynamically-linked executable?), or
	kOsError for unexpected errors returned by the OS.

*/
/****************************************************************************/

static DSError TrkFindSoLoaderInfo( ui32 processId,
									ui32 threadId,
									void *soLoaderBase, 
									void **breakSymbolPtr,
									InvisibleRangeInfo **fInvisibleRangeList)
{
	DSError err = kNoError;
	char buffer[ 0x1000 + PATH_MAX ];
	TrkLinuxElf tlElf;
	Elf_Scn *esec;
	Elf32_Shdr *eshdr;
	Elf_Data *edata;
	Elf32_Ehdr *ehdr;
	Elf32_Sym *esym;
	int j;
	char *dlPath = NULL;
	void *base = NULL;
	LinuxProcessState *pState;
	/*Temp node for Invisible step range list*/
	InvisibleRangeInfo *tmpInvisibleRangeList;
	
	*fInvisibleRangeList = NULL;
	tmpInvisibleRangeList = *fInvisibleRangeList;
	tlElf.fFd = -1;

	/*
	** Get an ELF handle for the executable.  There is a link to it in 
	** "/proc/##/exe", where ## is the process ID.
	*/
	sprintf( buffer, "/proc/%lu/exe", processId );
	err = TrkOpenElfLinux( buffer, &tlElf );
		
	/*
	** Get the address and size of the ".plt" section in the executable.
	*/
		
	if (err == kNoError)
	{
		__puts( "main .plt: " );
		/*Allocate node for Invisible step range which holds start and 
		end address of main .plt section*/
		err = TrkAllocInvisibleRangeNode(fInvisibleRangeList);
		if (err == kNoError)
		{
			TrkGetElfSectionRangeLinux( tlElf.fElf, ".plt", (void *)0,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fStart,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fEnd );
			(*fInvisibleRangeList)->next = tmpInvisibleRangeList;
			tmpInvisibleRangeList = *fInvisibleRangeList;
		}
	}
	
	/*
	** Find the .interp section and read in the contents.  It contains the name
	** of the dynamic linker file.
	*/
	
	if (err == kNoError)
	{
		esec = TrkFindElfSectionLinux( tlElf.fElf, ".interp" );
		if (esec != NULL)
			eshdr = elf32_getshdr( esec );
		if (esec == NULL || eshdr == NULL)
			err = kUnsupportedError;
	}
	
	if (err == kNoError)
	{
		edata = elf_getdata( esec, NULL );
		if (edata != NULL)
			dlPath = strdup( edata->d_buf );
		if (edata == NULL || dlPath == NULL)
			err = kUnsupportedError;
	}
	
	/*
	** Open the ELF file for the dynamic linker.
	*/
	
	if (err == kNoError)
	{
		__puts( "Opening dynamic linker file " );
		__puts( dlPath );
		__puts( "\r\n" );
		
		TrkCloseElfLinux( &tlElf );
		tlElf.fFd = -1;
		err = TrkOpenElfLinux( dlPath, &tlElf );
	}

	/*
	** If the base address of the loader was passed in, use it.  Otherwise
	** we assume we are stopped at the entry point to the dynamic linker
	** and calculate its base address by subtracting the unrelocated
	** entry point from the current PC.
	*/
	
	if (err == kNoError)
	{
		if (soLoaderBase == NULL)
		{
			ehdr = elf32_getehdr( tlElf.fElf );
			if (ehdr != NULL)
				base = (void *)(OsTargetGetPC( processId, threadId ) -
							ehdr->e_entry);
		}
		else
			base = soLoaderBase;

		
		/*
		** Get the addresses and sizes of the ".plt", ".text", and
		** ".got" sections in the loader.
		*/
		
		__puts( "loader .plt: " );
		/*Allocate node for Invisible step range which holds start and 
		end address of loader .plt section*/
		err = TrkAllocInvisibleRangeNode(fInvisibleRangeList);
		if (err == kNoError)
		{
			TrkGetElfSectionRangeLinux( tlElf.fElf, ".plt", base,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fStart,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fEnd );
			(*fInvisibleRangeList)->next = tmpInvisibleRangeList;
			tmpInvisibleRangeList = *fInvisibleRangeList;
		}
		__puts( "loader .text: " );
		/*Allocate node for Invisible step range which holds start and 
		end address of loader .text section*/
		err = TrkAllocInvisibleRangeNode(fInvisibleRangeList);
		if (err == kNoError)
		{
			TrkGetElfSectionRangeLinux( tlElf.fElf, ".text", base,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fStart,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fEnd );
			(*fInvisibleRangeList)->next = tmpInvisibleRangeList;
			tmpInvisibleRangeList = *fInvisibleRangeList;

		}

		__puts( "loader .got: " );
		/*Allocate node for Invisible step range which holds start and 
		end address of loader .got section*/
		err = TrkAllocInvisibleRangeNode(fInvisibleRangeList);
		if (err == kNoError)
		{
			TrkGetElfSectionRangeLinux( tlElf.fElf, ".got", base,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fStart,
				&(*fInvisibleRangeList)->InvisibleStepRanges.fEnd );
			(*fInvisibleRangeList)->next = tmpInvisibleRangeList;
		}
	}
	
	/*
	** Search for one of the symbols that refer to the function where
	** we want to set a breakpoint.
	**
	** The symbol value read from the file is not the final, relocated
	** address.  To get that, add the base address of the dynamic linker.
	*/

	if (err == kNoError)
	{
		err = kUnsupportedError;
		for (j = 0; solib_break_names[ j ] != NULL && err == kUnsupportedError; j++)
		{
			esym = TrkFindElfSymbolLinux( tlElf.fElf, solib_break_names[ j ] );
			if (esym != NULL)
			{
				err = kNoError;
				*breakSymbolPtr = (void *)(TrkGetElfSymbolOffset( tlElf.fElf, esym ) + (ui8 *)base);

				__puts( "solib symbol " );
				__puts( solib_break_names[ j ] );
				__puts( " found at " );
				__puthex8( (ui32)*breakSymbolPtr );
				__puts( "\r\n" );
			}
		}
		
		if (err == kUnsupportedError)
		{
			__puts( "solib break symbol not found.\r\n" );

			if (FindLinuxProcessStateId( processId, NULL, &pState ) == kNoError &&
				!pState->fDlWarned)
			{
				/* Print a warning for the user. */
				buffer[0] = '\0';
				TRK_strcat( buffer, "\nWarning: " );
				TRK_strcat( buffer, dlPath );
				if (pState->fExe != NULL)
				{
					TRK_strcat( buffer, " used by " );
					TRK_strcat( buffer, pState->fExe );
				}
				TRK_strcat( buffer, "\n         The dynamic loader does not contain the symbols needed for shared" );
				TRK_strcat( buffer, "\n         library debugging.  To enable shared library debugging, use an" );
				TRK_strcat( buffer, "\n         un-stripped version of this library.\n" );
				TrkPrintMessage( buffer );
				pState->fDlWarned = true;
			}
		}
	}
		

	/*
	** Cleanup.
	*/

	if (dlPath != NULL)
		free( dlPath );

	if (tlElf.fFd != -1)
		TrkCloseElfLinux( &tlElf );
		
	return( err );
}


/****************************************************************************/
/*
	TrkFindSoInfo
	
	Finds the .plt section start and end address of the input shared object.
	This is is function is called whenever any shared object is loaded/unloaded.
	This is called to compute invisible ranges for shared object. 
*/
/****************************************************************************/

DSError TrkFindSoInfo(char *sopath,
						     PCType *soPltStart,
						     PCType *soPltEnd)
{
	DSError err = kNoError;
	TrkLinuxElf tlElf;
	tlElf.fFd = -1;
	__puts("\r\nSONAME:");
	__puts(sopath);
	__puts("\r\n");

	err = TrkOpenElfLinux(sopath, &tlElf );
	
	/*
	** Get the address and size of the ".plt" section of the shared object.
	*/
		
	if (err == kNoError)
	{
		__puts( "so.plt: " );
		TrkGetElfSectionRangeLinux( tlElf.fElf, ".plt", (void *)0,
				soPltStart, soPltEnd );
	}
	
	TrkCloseElfLinux( &tlElf );
	return err;
	
}

/****************************************************************************/
/*
	TrkSetSoBreakpointsLinux
	
	Set breakpoints needed to detect shared-object events.
	
	Returns kNoError if successful, kUnsupportedError if not enough information
	was available (perhaps this is not a dynamically-linked executable?), or
	kOsError for unexpected errors returned by the OS.

*/
/****************************************************************************/

DSError TrkSetSoBreakpointsLinux( LinuxProcessState *pState,
									LinuxThreadState *tState )
{
	DSError err = kNoError;
	DSError err2 = kNoError;
	DSError *errPtr;
	struct r_debug rd;
	BreakpointId bpId;
	Breakpoint *bp;
	void *breakLoc;
	void *ldBase;
	void *temp;
	void **tempPtr;
	struct r_debug *soDebugPtr;
	size_t size;
		
	err = TrkFindSoDebugPtrLinux( pState->fProcessId, &soDebugPtr );
	if (err == kNoError)
	{
		/*
		** Read the r_brk field from the r_debug struct and set a breakpoint
		** at that location.
		*/
	
		pState->fSoDebugPtr = soDebugPtr;
		size = sizeof( rd );
		err = OsTargetAccessMemory( &rd, pState->fSoDebugPtr,
			&size, kDebuggerMemory, true,
			pState->fProcessId, 0 );
			
		breakLoc = (void *)rd.r_brk;

		/*
		** Get the other loader information using the base address of the
		** loader obtained from the r_debug struct.
		*/

		ldBase = (void *)rd.r_ldbase;
		tempPtr = &temp;
		errPtr = &err2;
	}
	else
	{
		/*
		** If the r_debug structure isn't available, find the symbol which
		** points to the function where we want to set the breakpoint.
		** We assume in this case that we are stopped at the entry point
		** to the loader.
		*/
		
		ldBase = NULL;
		tempPtr = &breakLoc;
		errPtr = &err;
	}
	
	*errPtr = TrkFindSoLoaderInfo( pState->fProcessId, tState->fThreadId,
				ldBase,tempPtr,
				&pState->fInvisibleStepRangeList);
	pState->fInvisibleStepRangeCount = kSectionCount;
	/*
	** Set the breakpoint.
	*/
	if (err == kNoError)
		err = GetFreeBreakpoint( &bpId, &bp );
		
	if (err == kNoError)
	{
		err = SetBreakpoint( bp, (unsigned int)breakLoc, 1,
			false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
			pState->fProcessId, DS_MSG_BREAK_THREAD_ID_ALL );
			
		if (err == kNoError)
		{
			__puts( "Shared object breakpoint set at " );
			__puthex8( (ui32)breakLoc );
			__puts( "\r\n" );
			pState->fSoBreakpointId = bpId;
		}
		else
		{
			ReleaseBreakpoint( bpId );
		}
	}

	return( err );
}

/****************************************************************************/
/*
	TrkGetSoListLinux
	
	Build a list of TrkSoInfo structures based on the provided address of 
	the link map in the target process.
*/
/****************************************************************************/
static DSError TrkGetSoListLinux( struct link_map *lmPtr, TrkSoInfo **soList,
							LinuxProcessState *pState, ui32 threadId )
{
	DSError err = kNoError;
	struct link_map lm;
	TrkSoInfo *soi;
	size_t size;
	size_t maxSize;
	size_t minSize;
	InvisibleRangeInfo *TempInvisibleRangeList;
	char buf[PATH_MAX];
	*soList = NULL;

	/*
	**Trim the list to avoid addition invisible ranges of  same .so multiple times.
	*/
	if (pState->fInvisibleStepRangeCount > kSectionCount)
	{
		TrkTrimInvisibleRangeList(&pState->fInvisibleStepRangeCount,&pState->fInvisibleStepRangeList);
	}

	for (;lmPtr != NULL && err == kNoError; lmPtr = lm.l_next)
	{
		buf[0] = '\0';

		/*
		** Read the next link_map struct.
		*/

		size = sizeof( lm );
		err = OsTargetAccessMemory( &lm, lmPtr,
			&size, kDebuggerMemory, true,
			pState->fProcessId, threadId );

		/*
		** Read the file name from the link_map struct.  Be sure
		** to NULL-terminate it in case we don't get a valid string.
		*/
		
		if (err == kNoError)
		{
			/*
			** If the name entry is NULL or the address is uninitialized,
			** skip it.
			*/
			
			if (lm.l_name == NULL || lm.l_addr == 0)
				continue;

			/*
			** Get the file name for the object. Reads are "all or nothing" and
			** we don't know how many bytes are in the name, so we simply try to
			** read MAX_FILENAME_SIZE number of bytes. If that turns out to not
			** be enough, we try a larger size. If we end up trying to read too 
			** many bytes (we go past the end of accessible memory), then we
			** go the other way. We hone in on a workable size using binary
			** interpolation.
			**
			** Of course, we could try reading byte by byte, but that could be
			** expensive if there are many entries in the so list and they
			** have long path names. In most cases, we'll get it right on
			** the first attempt.
			**/
			
			size = sizeof(buf);
			maxSize = size;
			minSize = 2;				// need at least one char + '\0' to be useful
			
			while (true)
			{
				size_t size_save = size;

				err = OsTargetAccessMemory( buf, lm.l_name,
					&size, kDebuggerMemory, true,
					pState->fProcessId, threadId );
					
				if (err == kNoError)
				{
					/*
					** If there's a null byte in the part of the buffer we populated with the read, 
					** then we know we got the entire string. Since reads are "all or nothing" and
					** the read succeeded, we know the null byte cannot be from uninitialized data 
					** in 'buf'
					*/
					bool bGotTheWholeString = false;
					int i;
					ASSERT(size == size_save);		// validates read was "all or nothing"
					for (i = 0; i < size; i++)
					{
						if (buf[i] == '\0')
						{
							bGotTheWholeString = true;
							break;
						}
					}

					if (bGotTheWholeString)
						break;	// we're done
					else
					{
						/* If the so name is larger than PATH_MAX there is something seriously wrong...*/
						if (size == sizeof(buf))
						{
							ASSERT(0);
							buf[0] = '\0';							
							break;
						}

						/*
						** Try a larger read. Make the last attempt size the 
						** lower bound; we know we can read at least that many.
						** maxSize continues to represent a size that is too big.
						** Somewhere inbetween these two lies the "sweet spot".
						*/
						minSize = size;
						size = (minSize + maxSize) / 2;
					}
				}
				else
				{
					/* OsTargetAccessMemory will set size to zero on a failed read; restore it */
					size = size_save;
					
					if (size <= minSize)
					{
						/* smallest possible read failed, so give up */
						ASSERT(minSize == 2);	// otherwise, a read that succeeded before failed now! Wouldn't make sense
						buf[0] = '\0';
						break;
					}
					else
					{
						/*
						** Try a smaller read. Make the last attempt size the 
						** upper bound since nothing equal or larger will 
						** succeed. minSize continues to represent either the value 2
						** or the largest size that we've been able to successfully 
						** read (but that obviously doesn't contain the whole so name,
						** or we wouldn't be here!).Somewhere inbetween these two lies
						** the "sweet spot".
						*/
						maxSize = size;
						size = (minSize + maxSize ) / 2;

						/*
						** It's possible (but unexpected) that through some sort of error
						** or corruption, the maximum number of readable bytes does
						** not contain a null byte. This will be the case if minSize
						** and maxSize end up differing by 1. We need to bail if this 
						** happens.
						*/
						if (minSize == (maxSize - 1))
						{
							ASSERT(0);
							buf[0] = '\0';
							break;
						}
						
						__puts( "Over-estimated so name size (disregard ptrace error).  Trying smaller size...\r\n" );
					}
				}
			}
		}

		/*
		** If an entry has no name, skip it.
		*/
		if (err == kNoError && buf[0] == '\0')
			continue;

		/*
		** The name might be a relative path (if nothing else, this is the case for
		** desktop Linux). It's relative to the executable path. Convert it to 
		** a full path.
		*/
		if (buf[0] != '/')
		{
			char * lastDelim = NULL;
			char temp[PATH_MAX];
			size_t length = 0;
			sprintf(temp, "/proc/%lu/exe", pState->fProcessId);
			length = readlink(temp, temp, sizeof(temp) - 1);		// readlink()'s return string does not have a null terminator
			if (length == -1)
			{
				ASSERT(0);	// huh? this should have worked...
				continue;
			}

			/* readlink()'s return string does not have a null terminator */
			temp[length] = '\0';	
			
			/* knock off the executable name */
			lastDelim = strrchr(temp, '/');		
			ASSERT(lastDelim);
			lastDelim[1] = '\0';
			
			/* add the relative path */
			strcat(temp, buf);
			
			/* and copy the complete result to 'buf' */
			strcpy(buf, temp);
		}


		/*
		** Allocate a new TrkSoInfo struct to represent the current link_map
		** struct and put it at the head of the list.
		*/
		
		if (err == kNoError)
		{
			soi = *soList;
			*soList = (TrkSoInfo *)malloc( sizeof( TrkSoInfo ) );
			if (*soList == NULL)
				err = kOsError;
		}


		/*
		** Fill in the fields for the new struct.
		*/
		if (err == kNoError)
		{
			(*soList)->l_next = soi;
			(*soList)->l_addr = lm.l_addr;
			(*soList)->l_name = strdup(buf);
			
			/*Compute invisible ranges for each shared object*/
			err = TrkAllocInvisibleRangeNode(&TempInvisibleRangeList);
			if (err == kNoError)
			{
				/* Finds .plt section start & end address of shared object*/
				/*these are offsets*/
				err = TrkFindSoInfo((*soList)->l_name,&TempInvisibleRangeList->InvisibleStepRanges.fStart,
					&TempInvisibleRangeList->InvisibleStepRanges.fEnd);

				if (err == kNoError)
				{
					/*add the library base address to get the true addresses for the range.*/
					TempInvisibleRangeList->InvisibleStepRanges.fStart += (*soList)->l_addr;
					TempInvisibleRangeList->InvisibleStepRanges.fEnd += (*soList)->l_addr;
					TempInvisibleRangeList->next = pState->fInvisibleStepRangeList;
					pState->fInvisibleStepRangeList = TempInvisibleRangeList;
					pState->fInvisibleStepRangeCount++;
				}

			}
		}
	}

	return( err );
}

/****************************************************************************/
/*
	TrkFindInSoListLinux
	
	Searches a TrkSoInfo list for an entry which matches the provided key.
*/
/****************************************************************************/
static TrkSoInfo *TrkFindInSoListLinux( TrkSoInfo *soList, TrkSoInfo *key )
{
	TrkSoInfo *soi;
	
	for (soi = soList; soi != NULL; soi = soi->l_next)
	{
		if (soi->l_addr == key->l_addr &&
			!strcmp( soi->l_name, key->l_name ))
			break;
	}
	
	return( soi );
}

/****************************************************************************/
/*
	TrkFindNameInSoListLinux
	
	Searches a TrkSoInfo list for an entry whose base name matches the
	specified name.
*/
/****************************************************************************/
static TrkSoInfo *TrkFindNameInSoListLinux( TrkSoInfo *soList, const char *name)
{
	TrkSoInfo *soi=NULL;
	char *baseName;
	
	for (soi = soList; soi != NULL; soi = soi->l_next)
	{
		/*
		** The base name is everything after the last '/', or the whole
		** string if there is no '/'.
		*/
		
		baseName = strrchr( soi->l_name, '/' );
		if (baseName == NULL)
			baseName = soi->l_name;
		else
			baseName++;
		
		if (!strcmp( baseName, name ))
			break;
	}
	
	return( soi );
}

/****************************************************************************/
/*
	TrkFindSoInProcessLinux
	
	Gets a TrkSoInfo object corresponding to the shared-object with the
	specified base name associated with given process.
*/
/****************************************************************************/
DSError TrkFindSoInProcessLinux( LinuxProcessState *pState, const char *name,
								TrkSoInfo **soi)
{
	*soi = TrkFindNameInSoListLinux( pState->fSoList, name );
	if (*soi == NULL)
		return( kNotFoundError );

	return( kNoError );
}

/****************************************************************************/
/*
	DoNotifyDllChanged
	
	Performs either a DLL created or DLL deleted notificiation depending
	on the parameters.
*/
/****************************************************************************/
static DSError DoNotifyDllChanged( ui32 offset, char *name, bool remaining,
								bool created, ui32 processId, ui32 threadId )
{
	if (created)
		return( DoOsNotifyDllCreated( offset, name, remaining, processId, threadId ) );
	else
		return( DoOsNotifyDllDeleted( name, remaining, processId, threadId ) );
}
/****************************************************************************/
/*
	TrkCompareSoListsLinux
	
	Compares an old TrkSoInfo list to the current one and generates the
	appropriate notification messages for added/removed entries.
*/
/****************************************************************************/

static DSError TrkCompareSoListsLinux(LinuxProcessState *pState, TrkSoInfo *oldSoList, TrkSoInfo *newSoList,
					int *notifyCount,ui32 processId, ui32 threadId )
{				
	DSError err = kNoError;
	TrkSoInfo *soi;
	TrkSoInfo *prevSoi = NULL;
	bool prevAdded = false;
	
	*notifyCount = 0;
		
	/*
	** Each time a notification goes out, we need to know if additional
	** notifications will follow.  Therefore each notification is 
	** postponed until the next one is found or none are left.
	*/
	
	
	/*
	** Look for any removed entries (in the old list but not the new).
	*/
	
	for (soi = oldSoList; soi != NULL && err ==  kNoError; soi = soi->l_next)
	{
		if (TrkFindInSoListLinux( newSoList, soi ) == NULL)
		{
			if (prevSoi != NULL)
			{
				err = DoNotifyDllChanged( prevSoi->l_addr, prevSoi->l_name, true,
						prevAdded,processId, threadId );
			}
			prevSoi = soi;
			prevAdded = false;
		}
	}

	/*
	** Look for any newly added entries (in the new list but not the old).
	*/
	
	for (soi = newSoList; soi != NULL && err ==  kNoError; soi = soi->l_next)
	{
		if (TrkFindInSoListLinux( oldSoList, soi ) == NULL)
		{
			if (prevSoi != NULL)
			{
				(*notifyCount)++;
				err = DoNotifyDllChanged( prevSoi->l_addr, prevSoi->l_name, true,
						prevAdded, processId, threadId );

			}
			prevSoi = soi;
			prevAdded = true;
		}
	}

	/*
	** Send out the final notification, if any.  This is the only one for which
	** the remaining flag will be false.
	*/
	
	if (prevSoi != NULL)
	{
		(*notifyCount)++;
		err = DoNotifyDllChanged( prevSoi->l_addr, prevSoi->l_name, false,
				prevAdded, processId, threadId );
	}

	return( err );	
}

/****************************************************************************/
/*
	TrkFreeSoListLinux
	
	Free the provided list of TrkSoInfo structures.
*/
/****************************************************************************/
void TrkFreeSoListLinux( TrkSoInfo *soList )
{
	if (soList != NULL)
	{
		TrkFreeSoListLinux( soList->l_next );
		free( soList->l_name );
		free( soList );
	}
}


/****************************************************************************/
/*
	TrkUpdateSoStateLinux
	
	Get the current shared object state for the process.
	
	If atBp is true, then the process has stopped at one of the shared object
	debug breakpoints.
	
	if notify is true, determine what events are occurring (e.g., DLL load/unload)
	and send the appropriate notifications to the host debugger.  If false, then
	NULL may be passed for notifyCount.
*/
/****************************************************************************/
DSError TrkUpdateSoStateLinux( LinuxProcessState *pState,
						LinuxThreadState *tState,
						bool atBp, bool notify,
						int *notifyCount )
{
	DSError err = kNoError;
	struct r_debug rd;
	size_t size;
	struct r_debug *soDebugPtr;
	TrkSoInfo *newSoList;
	TrkSoInfo *soi;

	if (notify)
	{
		ASSERT( notifyCount != NULL );
		*notifyCount = 0;
	}


	if (pState->fSoDebugPtr == NULL)
	{
		err = TrkFindSoDebugPtrLinux( pState->fProcessId, &soDebugPtr );
		if (err == kNoError)
			pState->fSoDebugPtr = soDebugPtr;
	}

	if (err == kNoError)
	{
		size = sizeof( rd );
		err = OsTargetAccessMemory( &rd, pState->fSoDebugPtr,
			&size, kDebuggerMemory, true,
			pState->fProcessId, 0 );
	}

	if (err == kNoError)
	{
		if (atBp)
		{
			__puts( "Status: " );
			switch( rd.r_state )
			{
				case RT_CONSISTENT:
					__puts( "RT_CONSISTENT\r\n" );
					break;
				case RT_ADD:
					__puts( "RT_ADD\r\n" );
					break;
				case RT_DELETE:
					__puts( "RT_DELETE\r\n" );
					break;
				default:
					__puts( "unknown\r\n" );
					break;
			}
		}
		
		err = TrkGetSoListLinux( rd.r_map, &newSoList, pState,
				tState->fThreadId );
			
	}
	
	/*
	** Compare the old and new lists to find added & removed items, sending 
	** appropriate notification mesages to the debugger.
	*/
	
	if (err == kNoError)
	{
		if (notify)
			err = TrkCompareSoListsLinux( pState, pState->fSoList, newSoList, notifyCount,
					pState->fProcessId,tState->fThreadId );
		
		/*
		** Free the old list and save the new one to compare against next time.
		*/

		TrkFreeSoListLinux( pState->fSoList );
		pState->fSoList = newSoList;
		__puts( "Shared objects:\r\n" );
		for (soi = newSoList; soi != NULL; soi = soi->l_next)
		{
			__puts( "  " );
			__puts( soi->l_name );
			__puts( ":" );
			__puthex8( soi->l_addr );
			__puts( "\r\n" );
		}
	}

	/*
	** If we haven't already sunk our claws into the process' thread
	** library, try again now.
	*/
	if (pState->fThreadAgent == NULL)
		TrkInitializeProcessThreadsLinux( pState );
	return( err );
}

