/******************************************************************************************/
/*

DESCRIPTION

	Linux thread handling routines.

COPYRIGHT

	Copyright 2000-2006, 2009 Freescale Semiconductor, Inc.
		
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "linux_thread.h"
#include "targimpl.h"
#include "break.h"
#include "nubassrt.h"
#include "msg.h"
#include "linux_msgtarg.h"
#include "linux_targimpl.h"
#include "linux_port_targimpl.h"
#include "linux_notify.h"
#include "linux_thread.h"
#include "linux_so.h"
#include "mainloop.h"
#ifndef __NotSupportSo__
#include "linux_elf.h"
#endif

/*
** Linux includes
*/

#include <thread_db.h>
#include <sys/procfs.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>				// for malloc/free

/*
** Set this to 1 to get detailed debug output when building the thread list.
*/
#define DEBUG_THREAD_LIST		(1)

typedef enum
{
  PS_OK,          /* generic "call succeeded" */
  PS_ERR,         /* generic. */
  PS_BADPID,      /* bad process handle */
  PS_BADLID,      /* bad lwp identifier */
  PS_BADADDR,     /* bad address */
  PS_NOSYM,       /* p_lookup() could not find given symbol */
  PS_NOFREGS      /* FPU register set not available for given lwp */
} ps_err_e;


/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

/****************************************************************************/
/*
 *	Storage for info about processes
 */
/***************************************************************************/

#define kThreadCount		100

static LinuxThreadState		gThreadObjects[ kThreadCount ];
static ObjectArray			gThreads;

/****************************************************************************/
/*
 *	Linked list element used to store a list of threads with
 *	associated information.  This is used to process the
 *	OsReadItem:ThreadItem message and is not the same as the list
 *	of threads we are currently debugging.
 */
/***************************************************************************/

typedef struct OsThreadItem
{
	struct OsThreadItem *next;
	ui32	id;
	ui32	libId;
	ui32	lwpId;
	ui8		state;
} OsThreadItem;


/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/

static DSError InitializeLinuxThreadState( ObjectHeader *ob );
static DSError TrkSetThreadEventBreakpoint( td_thragent_t *ta, td_event_e e,
							BreakpointId *bpId, ui32 processId );
static int RecurseAllThreads(OsLwpidItem* lwpidList, OsThreadItem** ptList, int* tid);

/****************************************************************************/
/*
	Function Definitions
*/
/****************************************************************************/

/****************************************************************************/
/*
	TrkInitializeThreadSupportLinux

	Do initial thread support setup, before debugging any processes.

*/
/****************************************************************************/
DSError TrkInitializeThreadSupportLinux()
{
	td_err_e tdErr;

#ifndef __NotSupportThread__
	tdErr = td_init();
#endif
	if (tdErr != TD_OK)
	{
		__puts( "td_init returned error " );
		__puthex8( tdErr );
		__puts( "\r\n" );
	}

	return( kNoError );
}

/****************************************************************************/
/*
	TrkInitializeProcessThreadsLinux

	For the given process, initializes thread debugging.

*/
/****************************************************************************/
DSError TrkInitializeProcessThreadsLinux( LinuxProcessState *pState )
{
	DSError err = kNoError;
#ifndef __NotSupportThread__
	td_err_e tdErr;
	td_thragent_t *ta;
	td_thr_events_t events;
	ObjectId pStateId;

	/* Abort if thread debugging is disallowed. */
	if (!pState->fAllowThreadDb)
		return( kNoError );

	/*
	** Attempt to get a thread agent for the process.
	*/

	pStateId = TrkGetLinuxProcessStateId( pState );
	ASSERT( pStateId != kInvalidObjectId );
	tdErr = td_ta_new( (struct ps_prochandle *)pStateId, &ta );

	if (tdErr == TD_OK)
	{
		pState->fThreadAgent = ta;

		/*
		** Unmask any events needed for debugging.  Note that these
		** must also be enabled in each thread in order to be reported.
		*/

		td_event_emptyset( &events );
		td_event_addset( &events, TD_CREATE );
		td_event_addset( &events, TD_DEATH );

		tdErr = td_ta_set_event( ta, &events );

		if (err != TD_OK)
		{
			__puts( "Failed to set global thread event mask." );
			err = kOsError;
		}
	}
	else
	{
		__puts( "Thread agent not created (not needed?) for " );
		__puthex8( pState->fProcessId );
		__puts( "\r\n" );
		err = kOsError;
	}

	/*
	** Set breakpoints for the various events.  This assumes that
	** the notification type is NOTIFY_BPT.
	*/


	if (err == kNoError)
		err = TrkSetThreadEventBreakpoint( ta, TD_CREATE,
			&(pState->fThreadCreateBreakpointId), pState->fProcessId );

	if (err == kNoError)
		err = TrkSetThreadEventBreakpoint( ta, TD_DEATH,
			&(pState->fThreadDeathBreakpointId), pState->fProcessId );

	#ifndef TRK_DELAY_ATTACHING_TO_PROCESS_THREADS
		/*
		 * Attaching to the process threads this early on is problematic on
		 * RedHat 9.0. The TRK crashes in the libthread_db call: td_ta_thr_iter,
		 * and trying to determine why has proven to be a very time consuming
		 * and fruitless exercise. What is certain, though, is that all goes
		 * smooth if we postpone calling td_ta_thr_iter. Perhaps the problem
		 * is that lthread_db isn't yet fully initialized at this point. Through
		 * much trial and error, I've been able to find a somewhat unexplainable
		 * point at which we can safely make the td_ta_thr_iter call. That point
		 * is after the first so load breakpoint notification where, funny enough,
		 * we detect no so has actually been loaded or unloaded (see where we
		 * set fReadyForDelayedProcThreadsAttach to true). Fortunately, there
		 * appears to be no side effect or lost functionality in delaying the
		 * TrkAttachProcessThreadsLinux, hence we do this regardless of the
		 * RedHat OS version. If this turns out to be a kernel issue that affects
		 * embedded targets then those project/targets should define the above
		 * macro.
		 *
		 * To reproduce the problem this works around, launch a debug session
		 * of a project/application that links in libpthread on RedHat 9.0
		 * (problem does not happen if libpthread is no linked in, since we do
		 * no otherwise use libthread_db.)
		 */
		if (err == kNoError)
			err = TrkAttachProcessThreadsLinux( pState );
	#endif

#endif
	return( err );
}

/****************************************************************************/
/*
	TrkSetThreadEventBreakpoint

	Sets a breakpoint to be triggered by the specified thread event.

*/
/****************************************************************************/
static DSError TrkSetThreadEventBreakpoint( td_thragent_t *ta, td_event_e e,
							BreakpointId *bpId, ui32 processId )
{
	DSError err = kNoError;
	Breakpoint *bp;
	td_err_e tdErr = TD_OK;
	td_notify_t notify;

	/*
	** Get the breakpoint address.
	*/

#ifndef __NotSupportThread__
	tdErr = td_ta_event_addr( ta, e, &notify );
#endif
	ASSERT( notify.type == NOTIFY_BPT );

	if (tdErr != TD_OK)
		err = kOsError;

	/*
	** Get the breakpoint object.
	*/

	if (err == kNoError)
		err = GetFreeBreakpoint( bpId, &bp );

	/*
	** Set the breakpoint.
	*/

	if (err == kNoError)
	{
		err = SetBreakpoint( bp, (unsigned int)notify.u.bptaddr, 1,
			false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
			processId, DS_MSG_BREAK_THREAD_ID_ALL );

		if (err != kNoError)
		{
			ReleaseBreakpoint( *bpId );
			*bpId = kInvalidBreakpointId;
		}
	}

	if (err != kNoError)
	{
		__puts( "Failed to set breakpoint for thread event: " );
		__puthex8( e );
		__puts( "\r\n" );
	}

	return( err );
}

/****************************************************************************/
/*
	TrkAttachThreadLinux

	Attach to a new thread & notify the debugger.

*/
/****************************************************************************/
DSError TrkAttachThreadLinux( const td_thrhandle_t *tHandle )
{
	DSError err = kNoError;
	td_thrinfo_t tInfo;
	LinuxProcessState *pState;
	ObjectId pStateId;
	ObjectId tStateId;
	ui32 threadId;
	int status;

	if (err == kNoError)
	{
#ifndef __NotSupportThread__
		if (td_thr_get_info ( tHandle, &tInfo ) != TD_OK)
			err = kOsError;
#endif
	}

	/*
	** Use ptrace to attach to the light-weight process associated
	** with the thread.  This assumes that each thread has its own
	** lw process.
	*/

	__puts( "attaching to new lwpid " );
	__puthex8( tInfo.ti_lid );
	__puts( "\r\n" );

	if (ptrace( PTRACE_ATTACH, tInfo.ti_lid ) < 0)
	{
		err = kOsError;
		__perror( "ptrace attach to lwpid failed" );
	}

	/*
	** Wait for the newly attached thread to stop.  If this call fails,
	** don't worry.  It happens - seemingly at random.
	*/

	if (err == kNoError)
	{
		waitpid( tInfo.ti_lid, &status, __WCLONE );
	}

	/*
	** Add a new thread state object for this thread so the debugger
	** can track it.  To do this, we need to extract the main process
	** ID, light-weight process ID, and debug thread ID from the
	** given thread handle.  The main process state object ID can be
	** obtained from the thread agent as the prochandle, and the
	** main process ID can be obtained from that.  The other two
	** come from the thread info structure.
	*/

	if (err == kNoError)
	{
#ifndef __NotSupportThread__
		if (td_ta_get_ph( tHandle->th_ta_p, (struct ps_prochandle **)&pStateId ) != TD_OK)
			err = kOsError;
#endif
	}

	if (err == kNoError)
	{
		pState = TrkGetLinuxProcessState( pStateId );
		ASSERT( pState );
		err = TrkAddThreadLinux( pStateId, tInfo.ti_tid, tInfo.ti_lid,
			&threadId, &tStateId );
	}

	/*
	** Enable event reporting for the following events.  This must be
	** done for each thread.
	*/

#ifndef __NotSupportThread__
	if (err == kNoError)
	{
		td_thr_event_enable( tHandle, TD_CREATE );
		td_thr_event_enable( tHandle, TD_DEATH );
		td_thr_event_enable( tHandle, TD_CATCHSIG );
	}
#endif

	/*
	** Notify the host debugger of the new thread and restart
	** its execution.
	*/

	if (err == kNoError)
	{
		err = DoOsNotifyThreadCreated( pState->fProcessId, threadId );
		PostContinueEvent( pState->fProcessId, threadId );
	}


	if (err != kNoError)
		__puts( "failed to attach to new thread\r\n" );

	return( err );
}

/****************************************************************************/
/*
	AttachThreadCallback

	Callback function for thread iteration.  Brings the given thread
	under the debugger's control.

*/
/****************************************************************************/
int AttachThreadCallback (const td_thrhandle_t *tHandle, void *cbData)
{
	DSError err = kNoError;
	td_thrinfo_t tInfo;
	LinuxProcessState *pState;
	LinuxThreadState *tState;
	ObjectId pStateId;

	/*
	** Get info on this thread, and get the associated process state
	** object from the thread agent.
	*/

#ifndef __NotSupportThread__
	if (td_thr_get_info ( tHandle, &tInfo ) != TD_OK)
		err = kOsError;
#endif

	__puts( "Thread: " );
	__puthex8( tInfo.ti_tid );
	__puts( " s:" );
	__puthex8( (ui32)tInfo.ti_startfunc );
	__puts( " st:" );
	__puthex8( tInfo.ti_state );
	__puts( " su:" );
	__puthex2( tInfo.ti_db_suspended );
	__puts( " t:" );
	__puthex2( tInfo.ti_type );
	__puts( " p:" );
	__puthex8( tInfo.ti_pri );
	__puts( "l:" );
	__puthex8( tInfo.ti_lid );
	__puts( "\r\n" );

	if (err == kNoError)
	{
#ifndef __NotSupportThread__
		if (td_ta_get_ph( tHandle->th_ta_p, (struct ps_prochandle **)&pStateId ) != TD_OK)
			err = kOsError;
#endif
	}

	if (err == kNoError)
	{
		pState = TrkGetLinuxProcessState( pStateId );
		ASSERT( pState );
		if (tInfo.ti_lid == pState->fProcessId)
		{
			/*
			** If the lwpid matches the main process ID, then this is
			** the default thread for the process.  There is already
			** a thread state object for this thread, but it was created
			** before getting any info from the thread library.  Get
			** it (it's the one w/thread ID 0), and update it with the
			** debug thread ID.  Also, we're already attached to this one,
			** so no need to do that again.  Finally, since the existence
			** of the default thread is implicit, no notification is
			** sent to the host debugger.
			*/
			err  = FindLinuxThreadStateId( pState->fProcessId, 0, NULL, &tState );

			if (err == kNoError)
				tState->fDbThreadId = tInfo.ti_tid;
		}
		else
		{
			/*
			** For all other threads, check to see if we already have
			** them in our list.  If not, attach to them and notify the
			** host debugger.
			*/

			if (FindLinuxThreadStateDbId( pState->fProcessId,
					tInfo.ti_tid, NULL, &tState ) != kNoError)
			{
				err = TrkAttachThreadLinux( tHandle );
			}
		}
	}

	/*
	** Enable event reporting for the following events.  This must be
	** done for each thread.
	*/

#ifndef __NotSupportThread__
	if (err == kNoError)
	{
		td_thr_event_enable( tHandle, TD_CREATE );
		td_thr_event_enable( tHandle, TD_DEATH );
		td_thr_event_enable( tHandle, TD_CATCHSIG );
	}
#endif

	/*
	** Returning anything other than 0 will abort the thread iteration.
	*/

	return( 0 );
}

/****************************************************************************/
/*
	TrkAttachProcessThreadsLinux

	For the given process, brings all of the threads under the debugger's
	control.

*/
/****************************************************************************/
DSError TrkAttachProcessThreadsLinux( LinuxProcessState *pState )
{
	td_err_e tdErr = TD_OK;
#ifndef __NotSupportThread__
	__puts( "TrkAttachProcessThreadsLinux\r\n" );

	if (pState->fThreadAgent != NULL)
	{
		if (tdErr == TD_OK)
		{
			/*
			** Iterate over all of the threads, adding each one to the list
			** of controlled threads.
			*/

			tdErr = td_ta_thr_iter (pState->fThreadAgent,
					AttachThreadCallback, NULL,
					TD_THR_ANY_STATE, TD_THR_LOWEST_PRIORITY,
					TD_SIGNO_MASK, TD_THR_ANY_USER_FLAGS);
		}
	}

#endif
	return( (tdErr == TD_OK) ? kNoError : kOsError );
}

/****************************************************************************/
/*
 *	PurgeThread
 *
 *	Purges all internally stored state for the given thread.  Do not
 *	attempt to alter the thread or associated process itself.  This
 *	function may be called after the thread and process has died, so
 *	fiddling with either one at this point could cause problems.
 */
/****************************************************************************/
DSError PurgeThread( ui32 processId, ui32 threadId )
{
	LinuxThreadState *tState;
	ObjectId i;

	/*
	** Release all breakpoints for the thread.
	**
	** Do not call ClearThreadBreakpoints() here.  That function will
	** try to un-set each breakpoint, which is no good if the thread
	** has already died.
	*/
//!!! This won't work.  If the breakpoint implementation is not
//!!! thread-local, then we need to actually remove the breakpoints
//!!! so they won't affect any of the threads that are still running.
//!!!	ReleaseThreadBreakpoints( processId, threadId );

	/*
	** If the specified thread is not in the list of controlled threads,
	** there is nothing left to do.
	*/
	if (FindLinuxThreadStateId( processId, threadId, &i, &tState ) != kNoError)
		return( kInvalidThreadId );

	/*
	** Attempt to detach.  Otherwise dead threads will remain as zombies as
	** long as the thread manager lives.
	*/

	ptrace( PTRACE_DETACH, tState->fLwProcessId, 0, 0 );

	/*
	** Release the thread state object.
	*/

	ReleaseArrayObject( &gThreads, i );

	return( kNoError );
}

/****************************************************************************/
/*
	TrkPurgeProcessThreadsLinux

	For the given process,  purges all internally stored state for the
	associated threads.  Do not attempt to alter the process or threads
	themselves.  This function may be called after the process/threads
	have died or after unattaching, so fiddling with the process/threads
	at this point could cause problems.

*/
/****************************************************************************/
DSError TrkPurgeProcessThreadsLinux( LinuxProcessState *pState )
{
	DSError err = kNoError;
	ObjectId i;
	LinuxThreadState *tState;

#ifndef __NotSupportThread__
	if (pState->fThreadAgent != NULL)
	{
		if (td_ta_delete( pState->fThreadAgent ) != TD_OK)
			return( kOsError );
	}
#endif
	/*
	** Iterate over all of the controlled threads.  For each one associated
	** with the given process, purge it.
	*/

	for (i = IterateObjectArrayFirst( &gThreads, (ObjectHeader **)&tState );
	     i != kInvalidObjectId && err == kNoError;
	     i = IterateObjectArrayNext( &gThreads, i, (ObjectHeader **)&tState ))
	{
		if (pState == TrkGetLinuxProcessState( tState->fProcessStateId ))
			err = PurgeThread( pState->fProcessId, tState->fThreadId );
	}


	return( kNoError );
}

/****************************************************************************/
/*
	TrkHandleThreadCreateEventLinux

	When a thread creation event occurs, get the info on the new
	thread and send a notification back to the host.

*/
/****************************************************************************/
DSError TrkHandleThreadCreateEventLinux( LinuxProcessState *pState )
{
	DSError err = kNoError;

#ifndef __NotSupportThread__
	/*
	** Find any new threads and attach to them.
	**
	** Note: The old implementation used td_ta_event_getmsg()
	** to get the TD_CREATE event message and attached to the
	** associated thread.  However, td_ta_event_getmsg() does
	** not reliably return the event message we are expecting.
	** The new implementation is less efficient but more reliable.
	*/

	err = TrkAttachProcessThreadsLinux( pState );
#endif
	return( err );
}

/****************************************************************************/
/*
	TrkHandleThreadDeathEventLinux

	When a TD_DEATH thread death event occurs, get the info on the dead
	thread and take appropriate action.

*/
/****************************************************************************/
DSError TrkHandleThreadDeathEventLinux( LinuxProcessState *pState )
{
	DSError err = kNoError;
	td_event_msg_t msg;
	td_thrinfo_t tInfo;
	LinuxThreadState *tState;

	/*
	** Get the message for the most recent event.  Verify that it
	** is a thread death event.
	*/
#ifndef __NotSupportThread__
	if (td_ta_event_getmsg( pState->fThreadAgent, &msg ) != TD_OK ||
		msg.event != TD_DEATH)
		err = kOsError;

	/*
	** Get info on the thread that caused the event.
	*/

	if (err == kNoError)
	{
		if (td_thr_get_info( msg.th_p, &tInfo ) != TD_OK)
			err = kOsError;
	}
#endif
	/*
	** Get the associated thread state object.
	*/

	if (err == kNoError)
		err = FindLinuxThreadStateDbId( pState->fProcessId, tInfo.ti_tid, NULL, &tState );

	/*
	** The thread debug library says the thread has died, but the associated
	** light-weight process is still alive.  At this point, simply store the
	** thread's exit code.  Only when the associated lw process actually dies
	** do we notify the host and remove the thread from the list of controlled
	** threads.
	*/

	if (err == kNoError)
		tState->fThreadResult = (void *)msg.msg.data;

	return( err );
}

/****************************************************************************/
/*
	TrkHandleThreadDeathLinux

	When a thread dies, update internal structures and send a notification
	back to the host.  If this is the last thread for a process, the process
	death is also handled.  The last two parameters are passed on to the
	process death handling routine if for that case.

	This function may be called for thread death due to either a TD_DEATH
	event or termination of the associated light-weight process.

*/
/****************************************************************************/
DSError TrkHandleThreadDeathLinux( void *threadResult,
								ui8 processDeathType, ui32 processDeathCode,
								ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	LinuxProcessState *pState;
	LinuxThreadState *tState;
	ObjectId i;
	bool bProcessDied = true;

	__puts( "Thread died: r" );
	__puthex8( (ui32)threadResult );
	__puts( " pdt" );
	__puthex8( processDeathType );
	__puts( " pdc" );
	__puthex8( processDeathCode );
	__puts( " " );
	__puthex8( processId );
	__puts( ":" );
	__puthex8( threadId );
	__puts( "\r\n" );


	/*
	** Find another thread associated with the same process.  If we
	** can't find one, then this was the last thread.  Treat that case
	** as a process death and call the appropriate handling routine
	** for that. Note that the thread list still contains an entry for
	** the dying thread so make sure not to consider that one.
	*/
	for (i = IterateObjectArrayFirst( &gThreads, (ObjectHeader **)&tState );
	     i != kInvalidObjectId && err == kNoError;
	     i = IterateObjectArrayNext( &gThreads, i, (ObjectHeader **)&tState ))
	{
		pState = TrkGetLinuxProcessState( tState->fProcessStateId );
		ASSERT( pState );

		if (pState->fProcessId == processId  &&  tState->fThreadId != threadId)
		{
			bProcessDied = false;
			break;
		}
	}

	/*
	** If this is the primary thread for the process and the light-weight process
	** is dead, then set the flag in the process state object indicating that the
	** primary thread is dead.
	*/

	err = FindLinuxThreadStateId( processId, threadId, NULL, &tState );
	if (err == kNoError)
	{
		pState = TrkGetLinuxProcessState( tState->fProcessStateId );
		ASSERT( pState );
		if (tState->fLwProcessId == pState->fProcessId)
			pState->fPrimaryThreadDead = true;

		/*
		** Get the stored thread exit result code if none was passed in.
		*/

		if (threadResult == NULL)
			threadResult = tState->fThreadResult;
	}

	if (bProcessDied)
	{
		err = TrkHandleProcessDeathLinux( processDeathType, processDeathCode,
			processId );
	}
	else
	{
		/*
		** Send out a kDSOSNotifyItemDeleted:kDSOSThreadItem message.
		*/

		err = DoOsNotifyThreadDeleted( (ui32)threadResult,
				processId, threadId );

		/*
		** Purge all stored state for the thread.
		*/

		PurgeThread( processId, threadId );
	}

	return( err );
}

/****************************************************************************/
/*
	InitializeThreadArray

	Initialize the thread status object array.

*/
/****************************************************************************/
DSError InitializeThreadArray()
{
	return( InitializeObjectArray( &gThreads, kThreadCount, sizeof( LinuxThreadState ),
									InitializeLinuxThreadState, &gThreadObjects ) );
}

/****************************************************************************/
/*
 *	InitializeLinuxThreadState
 *
 *	Initialize one of the LinuxThreadState objects.  This is called by
 *	the general-purpose object array routines, so the passed in type
 *	assumes the header only.  Note that the header itself should NOT be
 *	initialized by this routine, only the additional object-specific fields.
 */
/****************************************************************************/
static DSError InitializeLinuxThreadState( ObjectHeader *ob )
{
	LinuxThreadState *tState = (LinuxThreadState *)ob;

	ASSERT( tState );

	InitStepStatii( &tState->fStepStatii );
	tState->fContinueType = PTRACE_CONT;
	tState->fStopped = true;
	tState->fLastPc = 0;
	tState->fSignal = SIGSTOP;
	tState->fThreadResult = NULL;
	return kNoError;
}

/****************************************************************************/
/*
 *	MatchLinuxThreadStateId
 *
 *	Compare two thread state objects to determine if they have the same
 *	process & thread Ids.
 */
/****************************************************************************/
static bool MatchLinuxThreadStateId( ObjectHeader *ob1, ObjectHeader *ob2 )
{
	return( (((LinuxThreadState *)ob1)->fProcessStateId ==
			 ((LinuxThreadState *)ob2)->fProcessStateId ) &&
			(((LinuxThreadState *)ob1)->fThreadId ==
			 ((LinuxThreadState *)ob2)->fThreadId ) );
}

/****************************************************************************/
/*
 *	MatchLinuxThreadStateDbId
 *
 *	Compare two thread state objects to determine if they have the same
 *	process & debug thread IDs.
 */
/****************************************************************************/
static bool MatchLinuxThreadStateDbId( ObjectHeader *ob1, ObjectHeader *ob2 )
{
	return( (((LinuxThreadState *)ob1)->fProcessStateId ==
			 ((LinuxThreadState *)ob2)->fProcessStateId ) &&
			(((LinuxThreadState *)ob1)->fDbThreadId ==
			 ((LinuxThreadState *)ob2)->fDbThreadId ) );
}

/****************************************************************************/
/*
 *	MatchLinuxThreadStateLwId
 *
 *	Compare two thread state objects to determine if they have the same
 *	light-weight process ID.
 */
/****************************************************************************/
static bool MatchLinuxThreadStateLwId( ObjectHeader *ob1, ObjectHeader *ob2 )
{
	return( (((LinuxThreadState *)ob1)->fLwProcessId ==
			 ((LinuxThreadState *)ob2)->fLwProcessId ) );
}

/****************************************************************************/
/*
 *	MatchLinuxThreadStateLwAccessible
 *
 *	Compare two thread state objects to determine if they have the same
 *	process state ID and if the second one is accessible via ptrace().
 */
/****************************************************************************/
static bool MatchLinuxThreadStateLwAccessible( ObjectHeader *ob1, ObjectHeader *ob2 )
{
	errno = 0;
	ptrace( PT_READ_U, ((LinuxThreadState *)ob2)->fLwProcessId, 0, 0 );

	return( (((LinuxThreadState *)ob1)->fProcessStateId ==
			 ((LinuxThreadState *)ob2)->fProcessStateId ) &&
			 errno == 0);
}


/****************************************************************************/
/*
 *	MatchLinuxThreadState
 *
 *	Compare two thread state objects to determine if they have the same
 *	process state ID.
 */
/****************************************************************************/
static bool MatchLinuxThreadState( ObjectHeader *ob1, ObjectHeader *ob2 )
{
	return( (((LinuxThreadState *)ob1)->fProcessStateId ==
			 ((LinuxThreadState *)ob2)->fProcessStateId) );
}

/****************************************************************************/
/*
 *	FindLinuxThreadStateId
 *
 *	Find the thread state structure corresponding to the specified process
 *	and thread IDs.
 *
 *	The resultObId parameter can be NULL.
 */
/****************************************************************************/
DSError FindLinuxThreadStateId( ui32 processId,
									ui32 threadId,
									ObjectId *resultObId,
									LinuxThreadState **result )
{
	DSError err;
	ObjectId i;
	LinuxProcessState *pState;

	LinuxThreadState key;

	/*
	** Find the ProcessState object corresponding to the specified
	** processId and use its ID in the key.
	*/

	err = FindLinuxProcessStateId( processId, &i, &pState );

	if (err == kNoError )
	{
		key.fProcessStateId = i;
		key.fThreadId = threadId;
		err = FindArrayObject( &gThreads, &i, (ObjectHeader **)result,
								(ObjectHeader *)&key, &MatchLinuxThreadStateId );
		if (resultObId != NULL)
			*resultObId = i;
	}

	return( err );
}

/****************************************************************************/
/*
 *	FindLinuxThreadStateDbId
 *
 *	Find the thread state structure corresponding to the specified process
 *	and debug thread IDs.
 *
 *	The resultObId parameter can be NULL.
 */
/****************************************************************************/
DSError FindLinuxThreadStateDbId( ui32 processId,
									thread_t dbThreadId,
									ObjectId *resultObId,
									LinuxThreadState **result )
{
	DSError err;
	ObjectId i;
	LinuxProcessState *pState;

	LinuxThreadState key;

	/*
	** Find the ProcessState object corresponding to the specified
	** processId and use its ID in the key.
	*/

	err = FindLinuxProcessStateId( processId, &i, &pState );

	if (err == kNoError )
	{
		key.fProcessStateId = i;
		key.fDbThreadId = dbThreadId;
		err = FindArrayObject( &gThreads, &i, (ObjectHeader **)result,
								(ObjectHeader *)&key, &MatchLinuxThreadStateDbId );
		if (resultObId != NULL)
			*resultObId = i;
	}

	return( err );
}

/****************************************************************************/
/*
 *	FindLinuxThreadStateLwId
 *
 *	Find the thread state structure corresponding to the light-weight
 *	process ID.
 *
 *	The resultObId parameter can be NULL.
 */
/****************************************************************************/
DSError FindLinuxThreadStateLwId( lwpid_t lwProcessId,
									ObjectId *resultObId,
									LinuxThreadState **result )
{
	DSError err;
	ObjectId i;
	LinuxThreadState key;

	key.fLwProcessId = lwProcessId;
	err = FindArrayObject( &gThreads, &i, (ObjectHeader **)result,
							(ObjectHeader *)&key, &MatchLinuxThreadStateLwId );
	if (resultObId != NULL)
		*resultObId = i;

	return( err );
}

/****************************************************************************/
/*
 *	FindLinuxThreadStateLwAccessible
 *
 *	Find a thread state structure associated with the given process ID
 *	whose light-weight process is accesssible.
 *
 *	The resultObId parameter can be NULL.
 */
/****************************************************************************/
DSError FindLinuxThreadStateLwAccessible( ui32 processId,
									ObjectId *resultObId,
									LinuxThreadState **result )
{
	DSError err;
	ObjectId i;
	LinuxProcessState *pState;

	LinuxThreadState key;

	/*
	** Find the ProcessState object corresponding to the specified
	** processId and use its ID in the key.
	*/

	err = FindLinuxProcessStateId( processId, &i, &pState );

	if (err == kNoError )
	{
		key.fProcessStateId = i;
		err = FindArrayObject( &gThreads, &i, (ObjectHeader **)result,
								(ObjectHeader *)&key, &MatchLinuxThreadStateLwAccessible );
		if (resultObId != NULL)
			*resultObId = i;
	}

	return( err );
}

/****************************************************************************/
/*
 *	FindLinuxThreadStateLwAccessible
 *
 *	Find the first thread state structure associated with the given process ID
 *
 *	The resultObId parameter can be NULL.
 */
/****************************************************************************/
DSError FindLinuxThreadStateLwFirst( ui32 processId,
									ObjectId *resultObId,
									LinuxThreadState **result )
{
	DSError err;
	ObjectId i;
	LinuxProcessState *pState;

	LinuxThreadState key;

	/*
	** Find the ProcessState object corresponding to the specified
	** processId and use its ID in the key.
	*/

	err = FindLinuxProcessStateId( processId, &i, &pState );

	if (err == kNoError )
	{
		key.fProcessStateId = i;
		err = FindArrayObject( &gThreads, &i, (ObjectHeader **)result,
								(ObjectHeader *)&key, &MatchLinuxThreadState );
		if (resultObId != NULL)
			*resultObId = i;
	}

	return( err );
}

/****************************************************************************/
/*
 *	TrkAddThreadLinux
 *
 *	Add a new thread to the list of debugged threads.  The thread will
 *	be primarily associated with the process indicated by pStateId. The
 *	debug thread ID dbThreadId should be provided by the thread_db
 *	library.  The threadId parameter is output-only - it is
 *	generated by this function
 */
/****************************************************************************/
DSError TrkAddThreadLinux( ObjectId pStateId, thread_t dbThreadId,
							lwpid_t lwProcessId, ui32 *threadId,
							ObjectId *threadStateId )
{
	DSError err = kNoError;
	LinuxProcessState *pState;
	LinuxThreadState *tState;

	/*
	** Get the associated process state object.
	*/

	pState = TrkGetLinuxProcessState( pStateId );
	ASSERT( pState );

	/*
	** Get a free thread state object.  Then assign it the primary
	** process ID, generate and assign it the thread ID, and assign
	** it the provided debug thread ID and light-weight process ID.
	*/

	err = GetFreeArrayObject( &gThreads, threadStateId, (ObjectHeader **)&tState );
	if (err == kNoError)
	{
		tState->fProcessStateId = pStateId;
		tState->fThreadId = pState->fNextThreadId++;
		tState->fDbThreadId = dbThreadId;
		tState->fLwProcessId = lwProcessId;

		*threadId = tState->fThreadId;
	}

	return( err );
}

/****************************************************************************/
/*
 *	TrkKillProcessThreadsLinux
 *
 *	Kill all threads associated with the specified process.
 */
/****************************************************************************/
DSError TrkKillProcessThreadsLinux( ui32 processId )
{
	DSError err = kNoError;
	LinuxThreadState *tState;
	LinuxProcessState *pState;
	ObjectId i;

	/*
	** Iterate over all the threads in the list of controlled threads
	** and kill the ones associated with the specified process.
	*/

	for (i = IterateObjectArrayFirst( &gThreads, (ObjectHeader **)&tState );
	     i != kInvalidObjectId && err == kNoError;
	     i = IterateObjectArrayNext( &gThreads, i, (ObjectHeader **)&tState ))
	{
		pState = TrkGetLinuxProcessState( tState->fProcessStateId );
		ASSERT( pState );
		if (pState->fProcessId == processId)
			err = OsTargetKillThreadLinux( processId, tState->fThreadId );
	}


	return( err );
}

/****************************************************************************/
/*
 *	TrkGetLinuxThreadState
 *
 *	Get the thread state object associated with the specified thread state
 *	ID.
 */
/****************************************************************************/
LinuxThreadState *TrkGetLinuxThreadState( ObjectId i )
{
	return( (LinuxThreadState *)GetArrayObject( &gThreads, i ) );
}

/****************************************************************************/
/*
 *	TrkAccessDefaultSingleThreadLinux
 *
 *	R/W a register at the specified offset in user memory.  This
 *	implementation uses the thread_db library to access a register
 *	for a specific thread.
 */
/****************************************************************************/
DSError	TrkAccessDefaultSingleThreadLinux(unsigned int reg, DefaultType *value,
						bool read,
						LinuxThreadState *tState )
{
	/*
	** Simply route the call back to the process implementation using
	** the thread's light-weight process ID.
	** This assumes a 1-1 mapping between threads and lwpids.
	**
	** Note: the td_thr_get/setregs functions don't seem to work properly
	** under linux.
	*/

	return( TrkAccessDefaultSingleProcessLinux( reg, value, read,
				tState->fLwProcessId ) );
}

/****************************************************************************/
/*
 *	TrkAccessFPThreadLinux
 *
 *	R/W a range of floating point registers.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 *
 *	This implementation uses the thread_db library to access registers
 *	for a specific thread.
 */
/****************************************************************************/
DSError	TrkAccessFPThreadLinux(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						LinuxThreadState *tState )
{
	/*
	** Simply route the call back to the process implementation using
	** the thread's light-weight process ID.
	** This assumes a 1-1 mapping between threads and lwpids.
	*/

	return( TrkAccessFPProcessLinux( firstRegister, lastRegister, b,
			registerStorageSize, read, tState->fLwProcessId ) );
}

/****************************************************************************/
/*
 *	TrkContinueThreadLinux
 *
 *	Continue execution of the target thread, given a specific thread state.
 *	This assumes that the associated process is multi-threaded and has
 *	a thread agent.
 *
 */
/****************************************************************************/
DSError TrkContinueThreadLinux( LinuxThreadState *tState )
{
	/*
	** Simply route the call back to the process implementation using
	** the thread's light-weight process ID.
	** This assumes a 1-1 mapping between threads and lwpids.
	**
	** Note: the td_thr_dbresume function isn't implemented under Linux,
	** so don't bother with it.
	*/

	return( TrkContinueProcessLinux( tState->fLwProcessId, tState ) );
}

/****************************************************************************/
/*
 *	TrkStopThreadLinux
 *
 *	Stop the specified thread.
 *	This assumes that the associated process is multi-threaded and has
 *	a thread agent.
 */
/****************************************************************************/
DSError TrkStopThreadLinux( LinuxThreadState *tState )
{
	/*
	** Simply route the call back to the process implementation using
	** the thread's light-weight process ID.
	** This assumes a 1-1 mapping between threads and lwpids.
	**
	** Note: the td_thr_dbsuspend function isn't implemented under Linux,
	** so don't bother with it.
	*/

	return( TrkStopProcessLinux( tState->fLwProcessId ) );
}

/****************************************************************************/
/*
 *	TrkKillThreadLinux
 *
 *	"Takes care" of the specified thread.  This is a low-level function
 *	that kills the thread associated with the given thread state object.
 *	This assumes that the associated process is multi-threaded and has
 *	a thread agent.
 */
/****************************************************************************/
DSError TrkKillThreadLinux( LinuxThreadState *tState )
{
	DSError err = kNoError;

	/*
	** Get the light-weight process ID for the thread and kill it.
	** This assumes a 1-1 mapping between threads and lwpids.
	*/

	err = TrkKillProcessLinux( tState->fLwProcessId );

	return( err );
}

/****************************************************************************/
/*
	TrkFreeOsThreadList

	Free the given list.

*/
/****************************************************************************/
static void TrkFreeOsThreadList( OsThreadItem *list )
{
	OsThreadItem *curr;
	OsThreadItem *next;

	for (curr = list; curr != NULL; curr = next)
	{
		next = curr->next;
		free( curr );
	}
}

static DSThreadStateLinux ProcStateChar_To_DSThreadStateLinux(char c)
{
	DSThreadStateLinux state = kDSThreadUnknownLinux;

	switch (c)
	{
		case 'T':	state = kDSThreadStoppedLinux;	break;
		case 'R':	state = kDSThreadRunLinux;		break;
		case 'Z':	state = kDSThreadZombieLinux;	break;
		case 'S':	state = kDSThreadSleepLinux;	break;
		case 'D':	state = kDSThreadSleepLinux; 	break;
		
		default:
			ASSERT(0);
	}

	return state;
}

/****************************************************************************/
/*
	TrkBuildThreadList

	Build a new thread list based on the current system state.

*/
/****************************************************************************/
#ifdef __uClinux__
/*
 * See explanation in TrkBuildOsProcessList() for this customized uClinux behavior. That
 * function also has a uClinux variant.
 */
static OsThreadItem *TrkBuildOsThreadList( ui32 processId,
											int *listCount )
{
	OsThreadItem * tList = NULL;
	int i;
	LinuxThreadState * tState = NULL;
	LinuxProcessState * pState = NULL;
	bool abort = false;	


	*listCount = 0;

	if (FindLinuxProcessStateId(processId, NULL, &pState ) != kNoError)
	{
		char processStatus = '\0';
		if (GetProcessInfo(processId, NULL, &processStatus, NULL, NULL, NULL, NULL) == kNoError)
		{
			OsThreadItem * item = (OsThreadItem *) malloc(sizeof(OsThreadItem));	
			item->next = NULL;
			item->id = i;						// just use the index
			item->libId = 0;					// n/a; just set to zero
			item->lwpId = processId;
		
			item->state = ProcStateChar_To_DSThreadStateLinux(processStatus);
			
			*listCount = 1;
			return item;			
		}

		ASSERT(0);
		return NULL;
	}


	/*
	 * Iterate over all the threads in the list of controlled threads and add 
	 * the ones associated with the specified main process ID.
	 */
	for (i = IterateObjectArrayFirst(&gThreads, (ObjectHeader **)&tState);
	     i != kInvalidObjectId && !abort;
	     i = IterateObjectArrayNext(&gThreads, i, (ObjectHeader **)&tState))
	{
		pState = TrkGetLinuxProcessState(tState->fProcessStateId);
		ASSERT(pState);
		if (pState->fProcessId == processId)
		{
			char processStatus = '\0';
			OsThreadItem *temp = tList;
			(*listCount)++;
			tList = (OsThreadItem *) malloc(sizeof(OsThreadItem));
			if (tList == NULL)
			{
				tList = temp;
				abort = true;
				break;
			}
			
			tList->next = temp;
			tList->id = tState->fThreadId;
			tList->libId = tState->fDbThreadId;
			tList->lwpId = tState->fLwProcessId;

			if (GetProcessInfo(tList->lwpId, NULL, &processStatus, NULL, NULL, NULL, NULL) == kNoError)
				tList->state = ProcStateChar_To_DSThreadStateLinux(processStatus);			
			else
			{
				ASSERT(0);
				tList->state = kDSThreadUnknownLinux;
			}

			if (!abort)
			{
				__puts( "   id:" );
				__puthex8( tList->id );
				__puts( "   libid:" );
				__puthex8( tList->libId );
				__puts( "   lwpid:" );
				__puthex8( tList->lwpId );
				__puts( " s:" );
				__puthex2( tList->state );
				__puts( "\r\n" );
			}
		}
	}
	
	/* Cleanup */
	if (abort)
	{
		TrkFreeOsThreadList(tList);
		tList = NULL;
		*listCount = 0;
	}
	
	return tList;
}
#else
static OsThreadItem *TrkBuildOsThreadList( ui32 processId,
											int *listCount )
{
	LinuxThreadState* tState = NULL;
	LinuxProcessState* pState = NULL;

	extern OsLwpidItem* lwpidList;
	extern int g_lwpidCount;
	int lwpidCount = g_lwpidCount;
	int index;
	OsThreadItem* tList = NULL;
	int tid = 0;
	ObjectId i;
	bool abort = false;
	
	*listCount = 0;
	
	#if DEBUG_THREAD_LIST
		__puts( "Building thread list for process: " );
		__puthex8( processId );
	#endif

	/*
	 * We try first to see if the processes is in our list of controlled processess.
	 */
	if (FindLinuxProcessStateId(processId, NULL, &pState) == kNoError)
	{
		/*
	 	 * Iterate over all the threads in the list of controlled threads and add 
	 	 * the ones associated with the specified main process ID.
	 	 */
	
		for (i = IterateObjectArrayFirst( &gThreads, (ObjectHeader **)&tState );
			i != kInvalidObjectId && abort == false;
			i = IterateObjectArrayNext( &gThreads, i, (ObjectHeader **)&tState ))
		{
			pState = TrkGetLinuxProcessState(tState->fProcessStateId);
			ASSERT(pState);
			if (pState->fProcessId == processId)
			{
				char processStatus = '\0';
				OsThreadItem *temp = tList;
				(*listCount)++;
				
				tList = (OsThreadItem *) malloc(sizeof(OsThreadItem));
				if (tList == NULL)
				{
					tList = temp;
					abort = true;
					break;
				}
			
				tList->next = temp;
				tList->id = tState->fThreadId;
				tList->libId = tState->fDbThreadId;
				tList->lwpId = tState->fLwProcessId;
			
				if (GetProcessInfo(tList->lwpId, NULL, &processStatus, NULL, NULL, NULL, NULL) == kNoError)
					tList->state = ProcStateChar_To_DSThreadStateLinux(processStatus);			
				else
				{
					ASSERT(0);
					tList->state = kDSThreadUnknownLinux;
				}

				#if DEBUG_THREAD_LIST
					__puts( "id: " );
					__puthex8( tList->id );
					__puts( "libId: " );
					__puthex8( tList->libId );
					__puts( "lwpId: " );
					__puthex8( tList->lwpId );
					__puts( "state: " );
					__putchar( tList->state );
					__puts( "\r\n" );
				#endif
			}
		}
		return (tList);
	}

	/*
	 * If we reach here it means the process is not found among controlled processes.
	 */
	 

	/*
	 * Find the entry for which lwpid <=> processId
	 */
	for (index = 0; index < lwpidCount; index++)
	{
		if (lwpidList[index].pid == processId)
				  break;
	}

	if (index == lwpidCount)
	{
		free(lwpidList);
		return (NULL);
	}

	
	/*
	 * The thread list for a process always contains the main thread.
	 */	
	tList = (OsThreadItem*)malloc(sizeof(OsThreadItem));
	if (!tList)
	{
		free(lwpidList);
		return (NULL);
	}
	memset(tList, 0, sizeof(OsThreadItem));
	
	tList->id = tid++;
	tList->lwpId = lwpidList[index].pid;
	tList->state = lwpidList[index].state;
	tList->next = NULL;
	
	#if DEBUG_THREAD_LIST
		__puts( "id: " );
		__puthex8( tList->id );
		__puts( "libId: " );
		__puthex8( tList->libId );
		__puts( "lwpId: " );
		__puthex8( tList->lwpId );
		__puts( "state: " );
		__putchar( tList->state );
		__puts( "\r\n" );
	#endif

	
	if (RecurseAllThreads(lwpidList[index].threads, &tList, &tid) < 0)
	{
		free(lwpidList);
		free(tList);
		*listCount = 0;
		return (NULL);	
	}

	*listCount = tid;
	return (tList);

}
#endif

/****************************************************************************/
/*
	TargetAddThreadList

	Add thread list data to the given buffer, starting at the given
	index in the list.  If the start index is 0, refreshes the list.
	Otherwise, uses previously stored list (if the processId matches
	that of the previously stored list).

*/
/****************************************************************************/
DSError TargetAddThreadList( MessageBuffer *b, ui8 options, ui32 filter,
							ui32 startIndex,
                            ui32 *returnedCount,
                            ui32 *totalCount )
{
	DSError err = kNoError;
	unsigned int currentIndex;
	OsThreadItem *currentItem;
	ui32 processId = filter;

	static OsThreadItem *list = NULL;
	static unsigned int total = 0;
	static ui32 lastProcessId = 0;

	/*
	** If the requested start index is 0, then read a fresh thread
	** list.  Otherwise, it is an index into the last-read list
	** (the processId must match that of the last-read list).
	*/

	if (startIndex == 0)
	{
		TrkFreeOsThreadList( list );
		list = TrkBuildOsThreadList( processId, &total );
		lastProcessId = processId;
	}
	else if (processId != lastProcessId)
	{
		err = kInvalidProcessId;
	}

	if (err == kNoError)
	{
		*totalCount = total;
		*returnedCount = 0;

		if (startIndex < *totalCount)
		{
			/*
			** Find the item corresponding to the requested index.
			*/

			for (currentItem = list, currentIndex = 0;
				 currentItem != NULL && currentIndex < startIndex;
				 currentItem = currentItem->next, currentIndex++)
				;

			ASSERT( currentItem != NULL );

			/*
			** Add info to the buffer until we run out of threads or
			** buffer space.
			*/

			while (currentItem != NULL)
			{
				if (GetBufferRemaining( b ) <
					(TRK_MSG_TL_INFO_MIN_LENGTH_LINUX))
					break;

				err = MessageAdd_ui32( b, currentItem->id );
				if (err == kNoError)
					err = MessageAdd_ui32( b, currentItem->libId );
				if (err == kNoError)
					err = MessageAdd_ui32( b, currentItem->lwpId );
				if (err == kNoError)
					err = MessageAdd_ui8( b, currentItem->state );

				(*returnedCount)++;
				currentItem = currentItem->next;
			}
		}
	}

	return( err );
}

/****************************************************************************/
/*

	Callbacks required by the thread_db libraray.

*/
/****************************************************************************/
ps_err_e ps_pdaccess( struct ps_prochandle *pHandle,
                        psaddr_t addr, void *data, size_t size,
                        bool read)
{
	DSError err;
	size_t localSize;
	LinuxProcessState *pState;

	localSize = size;

	/*
	** The thread ID we pass in doesn't matter, which is convenient
	** since we have nothing to go on.
	*/

	pState = TrkGetLinuxProcessState( (ObjectId)pHandle );
	err = OsTargetAccessMemory( data, addr, &size, kUserMemory, read,
			pState->fProcessId, /*!!!*/0 );

	switch (err)
	{
		case kInvalidProcessId:
			return( PS_BADPID );
		case kNoError:
			return( PS_OK );
		default:
			return( PS_ERR );
	}
}

ps_err_e ps_pdread(struct ps_prochandle *pHandle,
                        psaddr_t addr, void *data, size_t size)
{
	__puts( "Called ps_pdread()\r\n" );
	return( ps_pdaccess( pHandle, addr, data, size, true ) );
}

ps_err_e ps_pdwrite(struct ps_prochandle *pHandle,
                        psaddr_t addr, const void *data, size_t size)
{
	__puts( "Called ps_pdwrite()\r\n" );
	return( ps_pdaccess( pHandle, (void *)addr, (void *)data, size, false ) );
}


ps_err_e ps_ptread(struct ps_prochandle *pHandle,
                        psaddr_t addr, void *data, size_t size)
{
	__puts( "Called ps_ptread()\r\n" );
	return( PS_OK );
}


ps_err_e ps_ptwrite(struct ps_prochandle *pHandle,
                        psaddr_t addr, const void *data, size_t size)
{
	__puts( "Called ps_ptwrite()\r\n" );
	return( PS_OK );
}



#ifndef __NotSupportSo__
ps_err_e ps_pglobal_lookup(struct ps_prochandle *pHandle,
        const char *oName, const char *sName, psaddr_t *sAddr)
{
	ps_err_e psErr = PS_OK;
	TrkLinuxElf tlElf;
	TrkSoInfo *soi;
	LinuxProcessState *pState;
	Elf32_Sym *esym;


	__puts( "Called ps_pglobal_lookup()\r\n" );

	tlElf.fFd = -1;

	pState = TrkGetLinuxProcessState( (ObjectId)pHandle );

	/*
	** To compute the relocated symbol value, we need the object base
	** address.  If we have that info, it will be in the shared-object
	** list.
	*/

	if (TrkFindSoInProcessLinux( pState, oName, &soi )
		!= kNoError)
	{
		psErr = PS_NOSYM;
		__puts( "Thread debug: so " );
		__puts( oName );
		__puts( " not found in process\r\n" );
	}


	/*
	** Get an ELF handle for the object.  Use the full path in the
	** shared-object info.
	*/
	if (psErr == PS_OK)
	{
		if (TrkOpenElfLinux( soi->l_name, &tlElf ) != kNoError)
		{
			psErr = PS_NOSYM;
			__puts( "Unable to open ELF for so " );
			__puts( soi->l_name );
			__puts( "\r\n" );
		}
	}

	/*
	** Search for one of the symbols that refer to the function where
	** we want to set a breakpoint.
	**
	** The symbol value read from the file is not the final, relocated
	** address.  To get that, add the base address of the shared object.
	*/

	if (psErr == PS_OK)
	{
		esym = TrkFindElfSymbolLinux( tlElf.fElf, sName );
		if (esym != NULL)
		{
			*sAddr = (psaddr_t)(TrkGetElfSymbolOffset( tlElf.fElf, esym ) +
					soi->l_addr);
			__puts( "psglobal symbol " );
			__puts( sName );
			__puts( " found at " );
			__puthex8( (ui32)*sAddr );
			__puts( "\r\n" );
		}
		else
		{
			psErr = PS_NOSYM;
			if (!pState->fThreadLibWarned)
			{
				char buffer[ 0x1000 ];
				__puts( "Unable to find symbol " );
				__puts( sName );
				__puts( "\r\n" );

				/* Print a warning for the user. */
				buffer[0] = '\0';
				TRK_strcat( buffer, "\nWarning: " );
				TRK_strcat( buffer, oName );
				if (pState->fExe != NULL)
				{
					TRK_strcat( buffer, " used by " );
					TRK_strcat( buffer, pState->fExe );
				}
				TRK_strcat( buffer, "\n         The thread library does not contain the symbols needed for multi-" );
				TRK_strcat( buffer, "\n         threaded debugging.  To enable multi-threaded debugging, use an" );
				TRK_strcat( buffer, "\n         un-stripped version of this library.\n" );
				TrkPrintMessage( buffer );
				pState->fThreadLibWarned = true;
			}
		}
	}


	/*
	** Cleanup.
	*/

	if (tlElf.fFd != -1)
		TrkCloseElfLinux( &tlElf );

	return( psErr );
}
#endif


ps_err_e ps_laccessregs(struct ps_prochandle *pHandle,
                        lwpid_t lid, const prgregset_t regs,
                        bool read)
{
	ps_err_e psErr = PS_OK;
	LinuxProcessState *pState;
	unsigned int reg;
	unsigned int regOffset;


	pState = TrkGetLinuxProcessState( (ObjectId)pHandle );
	ASSERT( pState );

	/*
	** Iterate through the TRK registers, accessing each at
	** its offset in the given thread_db register set.
	** This only works as long as all of the registers in the
	** prgregset_t are the same size as DefaultType.
	*/

	for (reg = 0; reg <= TRK_DEFAULT_MAX_REGISTER && psErr == PS_OK; reg++)
	{
		/*
		** Map CodeWarriorTRK register id to Linux thread_db register offset.
		*/

		regOffset = TrkDefaultIdToThreadDbOffset[ reg ];

		if (TrkAccessDefaultSingleProcessLinux( reg,
				(DefaultType *)(((ui8 *)regs) + regOffset),
				read, lid ) != kNoError)
			psErr = PS_ERR;
	}

	return( psErr );
}

ps_err_e ps_lgetregs(struct ps_prochandle *pHandle,
                        lwpid_t lid, prgregset_t regs)
{
	__puts( "Called ps_lgetregs()\r\n" );

	return( ps_laccessregs( pHandle, lid, regs, true ) );
}


ps_err_e ps_lsetregs(struct ps_prochandle *pHandle,
                        lwpid_t lid, const prgregset_t regs)
{
	__puts( "Called ps_lsetregs()\r\n" );

	return( ps_laccessregs( pHandle, lid, regs, false ) );
}


ps_err_e ps_lgetfpregs(struct ps_prochandle *pHandle,
                        lwpid_t lid, prfpregset_t *regs)
{
	__puts( "Called ps_lgetfpregs()\r\n" );
	return( PS_OK );
}


ps_err_e ps_lsetfpregs(struct ps_prochandle *pHandle,
                        lwpid_t lid, const prfpregset_t *regs)
{
	__puts( "Called ps_lsetfpregs()\r\n" );
	return( PS_OK );
}



pid_t ps_getpid (struct ps_prochandle *pHandle)
{
	LinuxProcessState *pState;
	__puts( "Called ps_getpid()\r\n" );

	pState = TrkGetLinuxProcessState( (ObjectId)pHandle );
	ASSERT( pState );
	return( pState->fProcessId );
}


ps_err_e ps_pstop (const struct ps_prochandle *pHandle)
{
	__puts( "Called ps_pstop()\r\n" );
	/*
	** This should be straight-forward.  However, linuxthreads
	** shouldn't use this function.  There isn't much point
	** in implementing this if it can't be tested, especially,
	** since the corresponding appropriate implementation for
	** ps_pcontinue() is not completely understood.
	*/
	return( PS_OK );
}


ps_err_e ps_pcontinue (const struct ps_prochandle *pHandle)
{
	__puts( "Called ps_pcontinue()\r\n" );

	/*
	** In order to implement this, we need a way to determine
	** the appropriate thread status object to use.  Most likely,
	** this would be determined by tracking the TD_SWITCHTO events
	** so we know which thread is currently assigned to the
	** process.  For linuxthreads, each thread is assigned to its
	** own lwpid, so this shouldn't be needed.
	*/
	return( PS_OK );
}



ps_err_e ps_lstop (const struct ps_prochandle *pHandle, lwpid_t lid)
{
	ps_err_e psErr = PS_OK;
	LinuxThreadState *tState;

	__puts( "Called ps_lstop()\r\n" );

	/*
	** Find the thread state object associated with the given light-weight
	** process ID.  In order for this to work correctly, there must be a
	** a static 1-1 mapping between the two.
	*/

	if (FindLinuxThreadStateLwId( lid, NULL, &tState ) != kNoError)
		psErr = PS_BADLID;

	ASSERT( tState->fProcessStateId == (ObjectId)pHandle );

	/*
	** Call the TRK function to stop the specified lid.
	*/

	if (psErr == PS_OK)
	{
		if (TrkStopProcessLinux( lid ) != kNoError)
			psErr = PS_ERR;
	}

	return( psErr );
}


ps_err_e ps_lcontinue (const struct ps_prochandle *pHandle, lwpid_t lid)
{
	ps_err_e psErr = PS_OK;
	LinuxThreadState *tState;

	__puts( "Called ps_lcontinue()\r\n" );

	/*
	** Find the thread state object associated with the given light-weight
	** process ID.  In order for this to work correctly, there must be a
	** a static 1-1 mapping between the two.
	*/

	if (FindLinuxThreadStateLwId( lid, NULL, &tState ) != kNoError)
		psErr = PS_BADLID;

	ASSERT( tState->fProcessStateId == (ObjectId)pHandle );

	/*
	** Call the TRK function to continue the specified lid.
	*/

	if (psErr == PS_OK)
	{
		if (TrkContinueProcessLinux( lid, tState ) != kNoError)
			psErr = PS_ERR;
	}

	return( psErr );
}

static int RecurseAllThreads(OsLwpidItem* lwpidList, OsThreadItem** ptList, int* tid)
{
	OsThreadItem* tItem = NULL;
	OsLwpidItem* lwpidItem = lwpidList;

	/*
	 * All threads created by other threads are part of the same 
	 * parent process. Add them in the processes list.
	 */
	while(lwpidItem)
	{
		tItem = (OsThreadItem*)malloc(sizeof(OsThreadItem));
		if (!tItem)
			return (-1);
		
		memset(tItem, 0, sizeof(tItem));

		tItem->id = (*tid)++;
		tItem->lwpId = lwpidItem->pid;
		tItem->state = lwpidItem->state;

		/*
		 * Add the new entry at the end of the list.
		 */

		tItem->next = *ptList;
		*ptList = tItem;

		#if DEBUG_THREAD_LIST
			__puts( "	id: " );
			__puthex8( tItem->id );
			__puts( "	libId: " );
			__puthex8( tItem->libId );
			__puts( "	lwpId: " );
			__puthex8( tItem->lwpId );
			__puts( "	state: " );
			__putchar( tItem->state );
			__puts( "\r\n" );
		#endif

		if (RecurseAllThreads(lwpidItem->threads, ptList, tid) <0)
		{
			free(tItem);
			return (-1);
		}

		lwpidItem = lwpidItem->next;
	}
	return (0);
}


