/******************************************************************************************/
/*
	
DESCRIPTION

	Linux timeout routines.


	A timeout implementation must define TimeoutType and provide
	functions or macros that conform to the following prototypes:

	void ResetTimeout( TimeoutType *timer_ptr );
		// Clear the given timer object and start timing.

	bool Timeout( TimeoutType *timer_ptr );
		// Test the given timer object.  If timeout has occurred,
		// return true, else return false.

COPYRIGHT

	Copyright 2002, 2006, 2009 Freescale Semiconductor, Inc.
		
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "target.h"
#include "linux_timeout.h"
#include "linux_comm.h"

/*
** Linux includes
*/
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

/*
** How many seconds to wait for the debugger to reply to a message
** before timing out.
*/
#define TIMEOUT_SECS	10

/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/


/****************************************************************************/
/*
	Function Definitions
*/
/****************************************************************************/

void ResetTimeout( TimeoutType *timer_ptr )
{
	time( timer_ptr );
}

bool Timeout( TimeoutType *timer_ptr )
{
	fd_set readFds;
	fd_set writeFds;
	fd_set exceptFds;
	struct timeval remaining;
	time_t elapsed;

	/*
	** If already timed out, return immediately.
	*/

	elapsed = time( NULL ) - (*timer_ptr);
	if (elapsed >= TIMEOUT_SECS)
		return( true );

	/*
	** If not already timed out, use select() to sleep until either:
	**  1) new input arrives
	**  2) an exception occurs on the input stream
	**  3) the timeout period expires
	*/

	FD_ZERO( &readFds );
	FD_ZERO( &writeFds );
	FD_ZERO( &exceptFds );

	FD_SET( gCommInFd, &readFds );
	FD_SET( gCommInFd, &exceptFds );

	remaining.tv_sec = TIMEOUT_SECS - elapsed;
	remaining.tv_usec = 0;
	
	select( gCommInFd + 1, &readFds, &writeFds, &exceptFds, &remaining );

	/*
	** No need to check for timeout again here.  Return false, and if
	** TRK didn't get the input it wanted, it will call Timeout() 
	** again.  The test at the top of this function is sufficient.
	*/
	
	return( false );
}
