/****************************************************************************/
/*

DESCRIPTION

	Target-specific implementation for Linux on PPC

AUTHORS

	Copyright 2002, 2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/****************************************************************************/

/*
** TRK includes
*/

#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "msg.h"
#include "targimpl.h"
#include "main_TRK.h"
#include "linux_targimpl.h"
#include "target.h"
#include "arm_reg.h"
#include "arm_instr.h"
#include "break.h"

/*
** Linux includes
*/

#include <errno.h>

#include <sys/ptrace.h>

#ifdef __GCC_3_3__
#include <asm/ptrace.h>		
#else
#include <asm-arm/ptrace.h>		
#endif

#include "arm_reg.h"	// Generic ARM Register Definitions

/*
** Define this to 1 to get detailed debugging output when accessing
** the floating point registers.
*/

#define DEBUG_FLOAT_ACCESS		(0)

/*
** Map from MetroTRK register ID to Linux register offset.
** Note: PT_R0, etc. are defined in <asm/ptrace.h>
*/
#define LINUX_INVALID_REG_ID			((U32)-1)

const int TrkDefaultIdToPtraceOffsetArray[ TRK_DEFAULT_MAX_REGISTER + 1 ] =
{
	0*4, /* it was PT_R0 etc., */	// GPR(0)
	1*4,
	2*4,
	3*4,
	4*4,
	5*4,
	6*4,
	7*4,
	8*4,
	9*4,
	10*4,	// GPR(11)
	11*4,	// FP
	12*4,	// IP
	13*4,	// SP
	14*4,	// LR
	15*4,	// PC
	16*4,	// CPSR
};

const int *TrkDefaultIdToPtraceOffset =
		TrkDefaultIdToPtraceOffsetArray;

const int *TrkDefaultIdToThreadDbOffset =
		TrkDefaultIdToPtraceOffsetArray;


/****************************************************************************/
/*
 *	TrkAccessFPProcessLinux
 *
 *	R/W a range of floating point registers.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 *
 *	NOTE:This is a low-level implementation that can be used
 *	to access the light-weight process for a thread.  This means that
 *	processId may NOT be the main process ID for the given thread.  This
 *	is important, because many of the TRK functions which accept a
 *	processId will ONLY accept the main process ID.  Avoid calling other
 *	TRK functions with the given process ID.
 */
/****************************************************************************/
DSError	TrkAccessFPProcessLinux(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId )
{
	return( kUnsupportedError );
}						


/****************************************************************************/
/*
 *	OsTargetAccessExtended1
 *
 *	R/W a range in the extended1 register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	OsTargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId,
						ui32 threadId )
{
#if __MWERKS__
	#pragma unused(firstRegister, lastRegister, b, read, processId, threadId)
#endif /* __MWERKS__ */
	
	*registerStorageSize = 0;
	
	return kUnsupportedError;		// no extended1 regs under Linux
}						

/****************************************************************************/
/*
 *	OsTargetAccessExtended2
 *
 *	R/W a range in the extended2 register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	OsTargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId,
						ui32 threadId )
{
#if __MWERKS__
	#pragma unused(firstRegister, lastRegister, b, read, processId, threadId)
#endif /* __MWERKS__ */
	
	*registerStorageSize = 0;
	
	return kUnsupportedError;		// no extended1 regs under Linux
}						


/****************************************************************************/
/*
 *	OsTargetGetPC
 *
 *	Get the PC for the specified thread.
 */
/****************************************************************************/
PCType OsTargetGetPC( ui32 processId, ui32 threadId )
{
	DefaultType value;

	OsTargetAccessDefaultSingle( TRK_DEFAULT_PC,
			&value, true, processId, threadId );

	return( (PCType)value );
}						



/****************************************************************************/
/*
 *	OsTargetAdjustBreakPC
 *
 *	When the target stops at a breakpoint, the PC may or may not be
 *	where we want it.  Adjust accordingly.
 *
 *  NOTE1: This should be called once after each breakpoint is hit,
 *		   before the PC is used for anything. In particular before
 *		   sending out a stop notification or performing a continue.
 *
 */
/****************************************************************************/
DSError OsTargetAdjustBreakPC( ui32 processId, ui32 threadId )
{
	/*
	** For PPC, nothing to do.
	*/
}

/****************************************************************************/
/*
 *	OsTargetSetBreakpoint
 *
 *	Sets a breakpoint according to the given breakpoint descriptor.
 */
/****************************************************************************/
DSError OsTargetSetBreakpoint( Breakpoint *b )
{
	DSError result = kNoError;
	InstructionType instr, test;
	
	/*
	** When fork() is detected, TRK must remove all of the parent's
	** breakpoints from the child.  However, some breakpoints may be removed
	** from the parent after the fork() but before TRK detects the fork().
	** This is bad because the child inherits the breakpoint, but TRK no
	** longer sees it in the parent so it can't remove the bp from the
	** child.  Use delayed-release to retain the removed bp info for a short
	** time in case this happens.  Set the fReleaseDelay to the number of 
	** delay cycles to wait before truly releasing a breakpoint.  A "cycle"
	** is target-defined.  In the Linux case, each time the process stops is
	** 1 cycle.  1 delay cycle *should* be enough for fork().
	*/
	
	//For i.MX21 Delay=1 cause "steping on fork() system call won't allow to proceed on Child"
	b->fReleaseDelay = 3;
	
	if (b->fHardware)
	{
		/*
		** No hardware break support yet.
		*/
		__puts( "Attempting to set HW breakpoint: unsupported\r\n" );
		result = kUnsupportedError;
	}
	else
	{
		test.i32 = ARM_BREAK;
	 	result = DefaultTargetSetSoftwareBreakpoint( b, test );   
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	OsTargetClearBreakpoint
 *
 *	Clear the breakpoint described in the given breakpoint object.
 */
/****************************************************************************/
DSError OsTargetClearBreakpoint( Breakpoint *b )
{
	DSError result = kNoError;
	InstructionType instr;
	
	if (b->fHardware)
	{
		/*
		** No hardware break support yet.
		*/
		result = kUnsupportedError;
	}
	else
	{
		result = DefaultTargetClearSoftwareBreakpoint( b );
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	OsTargetHideBreakpoint
 *
 *	Hide the breakpoint described in the given breakpoint object.
 */
/****************************************************************************/
DSError OsTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  )
{
	return( DefaultTargetHideBreakpoint( b, data, maxLen ) );
}

/****************************************************************************/
/*
 *	OsTargetAccessInstruction
 *
 *	Read instruction from memory, determining instruction size.  The
 *  storage pointed to by readData must be big enough to store the
 *  largest possible instruction.
 *
 *	Note: if maxSize is -1, the size is unlimited
 */
/****************************************************************************/
DSError OsTargetAccessInstruction(InstructionType* data,
								void* virtualAddress, size_t maxSize,
								MemoryAccessOptions accessOptions,
								bool read,
								ui32 processId, ui32 threadId)
{
	DSError err = kNoError;
	size_t length;
	size_t outlength;
	
	/*
	** Determine instruction size.  Instructions on PPC are 32 bits.
	** Bound it by maxSize.
	*/

	length = sizeof( InstructionType );
	if (maxSize != -1 && maxSize < length)
		length = maxSize;


	/*
	** Read/write the instruction in memory.  If we're doing this in the
	** MetroTRK process, then just use memcpy.
	*/
	
	outlength = length;
	if (processId != TRK_PROCESS_ID)
	{
		err = OsTargetAccessMemory( data, virtualAddress, &outlength,
									accessOptions, read,
									processId, threadId);
	}
	else
	{
		if (read)
			memcpy( data, virtualAddress, length );
		else
			memcpy( virtualAddress, data, length );
	}
	
	/*
	 * Verify that an entire instruction was read/written.
	 */
	 
	if (err == kNoError && length != outlength)
		err = kInvalidMemory;
		
	return( err );
}
