/****************************************************************************/
/*

DESCRIPTION

	Target-specific implementation for Linux on PPC
	
AUTHORS

	Copyright 2000-2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/****************************************************************************/

/*
** TRK includes
*/

#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "msg.h"
#include "targimpl.h"
#include "main_TRK.h"
#include "linux_targimpl.h"
#include "target.h"
#include "ppc_reg.h"
#include "ppc_instr.h"
#include "break.h"
#include "nubinit.h"

/*
** Linux includes
*/

#include <errno.h>
#include <sys/ptrace.h>

/*
** Define this to 1 to get detailed debugging output when accessing
** the floating point registers.
*/

#define DEBUG_FLOAT_ACCESS		(1)

/*
** Map from MetroTRK register ID to Linux register offset.
** Note: PT_R0, etc. are defined in <asm/ptrace.h>
*/
#define LINUX_INVALID_REG_ID			((U32)-1)

const int TrkDefaultIdToPtraceOffsetArray[ TRK_DEFAULT_MAX_REGISTER + 1 ] =
{
	PT_R0*4,	// GPR(0)
	PT_R1*4,
	PT_R2*4,
	PT_R3*4,
	PT_R4*4,
	PT_R5*4,
	PT_R6*4,
	PT_R7*4,
	PT_R8*4,
	PT_R9*4,
	PT_R10*4,
	PT_R11*4,
	PT_R12*4,
	PT_R13*4,
	PT_R14*4,
	PT_R15*4,
	PT_R16*4,
	PT_R17*4,
	PT_R18*4,
	PT_R19*4,
	PT_R20*4,
	PT_R21*4,
	PT_R22*4,
	PT_R23*4,
	PT_R24*4,
	PT_R25*4,
	PT_R26*4,
	PT_R27*4,
	PT_R28*4,
	PT_R29*4,
	PT_R30*4,
	PT_R31*4,	// GPR(31)
	PT_NIP*4,	// PC
	PT_LNK*4,	// LR
	PT_CCR*4,	// CR
	PT_CTR*4,	// CTR
	PT_XER*4	// XER
};

const int *TrkDefaultIdToPtraceOffset =
		TrkDefaultIdToPtraceOffsetArray;

const int *TrkDefaultIdToThreadDbOffset =
		TrkDefaultIdToPtraceOffsetArray;

/*
** The FPR's start at (PT_FPR0 * 4) and occupy 8 bytes each. 
*/

#define PTRACE_FPR_OFFSET( r )		((PT_FPR0 * 4) + ((r) * 8))


/****************************************************************************/
/*
 *	TrkAccessFPProcessLinux
 *
 *	R/W a range of floating point registers.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 *
 *	NOTE:This is a low-level implementation that can be used
 *	to access the light-weight process for a thread.  This means that
 *	processId may NOT be the main process ID for the given thread.  This
 *	is important, because many of the TRK functions which accept a
 *	processId will ONLY accept the main process ID.  Avoid calling other
 *	TRK functions with the given process ID.
 */
/****************************************************************************/
DSError	TrkAccessFPProcessLinux(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId )
{
	DSError err = kNoError;
	unsigned int reg;
	unsigned int offset;
	union {
		FloatType f;
		long      l[2];
	} value;

	if (firstRegister < TRK_FLOAT_MIN_REGISTER ||
		lastRegister > TRK_FLOAT_MAX_REGISTER)
		return( kInvalidRegister );

	/*
	** Iterate through the registers.  For each, read/write it
	** into/from the provided buffer.
	*/
	
#if DEBUG_FLOAT_ACCESS
	if (read)
		__puts( "Reading" );
	else
		__puts( "Writing" );
	__puts( " float regs " );
	__puthex4( firstRegister );
	__puts( ":" );
	__puthex4( lastRegister );
	__puts( "\r\n" );
#endif

	for (reg = firstRegister, *registerStorageSize = 0;
		 reg <= lastRegister && err == kNoError;
		 reg++)
	{
		*registerStorageSize += TRK_FLOAT_SIZE;
		
		errno = 0;
		
		if (read)
		{
			switch( reg )
			{
				case TRK_FLOAT_FPECR:	/* not accessible via ptrace */
					value.f = 0;
					break;
				case TRK_FLOAT_FPSCR:	/* a single long value */
					value.f = ptrace( PTRACE_PEEKUSER, processId, PT_FPSCR*4 );
					break;
				default:				/* FPRs are stored as pairs of longs */
					offset = PTRACE_FPR_OFFSET( reg );
					value.l[0] = ptrace( PTRACE_PEEKUSER, processId, offset );
					value.l[1] = ptrace( PTRACE_PEEKUSER, processId, offset + 4 );
					break;
			}
			err = MessageAdd_ui64( b, value.f );
		}
		else
		{
			err = MessageGet_ui64( b, &value.f );
			if (err == kNoError)
			{
				switch( reg )
				{
					case TRK_FLOAT_FPECR:	/* not accessible via ptrace */
						break;
					case TRK_FLOAT_FPSCR:	/* a single long value */
						ptrace(PTRACE_POKEUSER, processId, PT_FPSCR*4, (long)value.f );
						break;
					default:				/* FPRs are stored as pairs of longs */
						offset = PTRACE_FPR_OFFSET( reg );
						ptrace(PTRACE_POKEUSER, processId, offset, value.l[0] );
						ptrace(PTRACE_POKEUSER, processId, offset + 4, value.l[1] );
						break;
				}
			}
		}
		
		if (errno != 0)
		{
			err = kOsError;
#ifdef DEBUGIO_CUSTOM
			perror( "ptrace register access failed" );
#endif
		}

#if DEBUG_FLOAT_ACCESS
		__puts( "   " );
		__puthex16( value.f );
#ifdef DEBUGIO_CUSTOM
		printf( " (%f)", *((double *)&value.f) );
#endif
		__puts( "\r\n" );
#endif

	}

	return( err );
}						

extern
void displayRegParameters(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId,
						ui32 threadId );

/****************************************************************************/
/*
 *	OsTargetAccessExtended1
 *
 *	R/W a range in the extended1 register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	OsTargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId,
						ui32 threadId )
{

#ifdef PTRACE_GETEVRREGS

	DSError err = kNoError;
	LinuxProcessState *pState;
	LinuxThreadState *tState;
	Extended1_PPC_Linux extended1Regs;
	Extended1Type *start;
	int count;
	int i;
	unsigned long fake_addr = 0;

    displayRegParameters(firstRegister, lastRegister, b, registerStorageSize, read, processId, threadId);

	/*
	** Reject if the debugger tries to operate on MetroTRK.
	*/
	if (processId == getpid())
		return( kInvalidProcessId );

	if (firstRegister < DS_EXTENDED1_MIN_REGISTER_LINUX ||
		lastRegister > DS_EXTENDED1_MAX_REGISTER_LINUX)
		return( kInvalidRegister );

	if (firstRegister <= lastRegister)
	{
		/*
		** Get a pointer to the start of the requested register range and
		** calculate the size of the range in bytes.
		*/
		start = (Extended1Type *)&extended1Regs + firstRegister;
		
		//the ACC SPR is a 64-bit register located entierly in the Extended1 register set
		//so the count will be 2!
		if ( firstRegister == lastRegister && firstRegister == EPPC_E500_ACC_OFFSET )
		  count = 2;
		else
		  count = lastRegister - firstRegister + 1;
		
		(*registerStorageSize) = (*registerStorageSize) + count * sizeof(Extended1Type);
		
		
		if (FindLinuxThreadStateId( processId, threadId, NULL, &tState ) != kNoError)
			err = kInvalidThreadId;

		if (err == kNoError)
		{
			pState = TrkGetLinuxProcessState( tState->fProcessStateId );
			if ( pState == NULL )
			{
				err = kOsError;
				__perror( "GetProcessState returned NULL object" );
                return err;
			}

			/*
			** Use ptrace to read the SPE registers 
			*/

            errno = 0; 

            for ( i = 0; i < sizeof(Extended1_PPC_Linux) / sizeof(ui32); i++ )
            {
             	__puts(" EGPR");
             	__puthex2(i);
             	__puts(": ");
             	__puthex8( *(((ui32 *)&extended1Regs) + i) );          
             	__puts("\t");
             	if ( i % 5 == 4 )
             	  __puts("\n");
            }
            __puts("\n");
          
          
			/* Execute read before write because in the case of the PTRACE_SETEVRREGS, 
			** the ptrace() call will write all the extended registers!
			** Therefore the read extended registers will always be performed when an 
			** operation over these registers is needed!
			*/
			if (pState->fThreadAgent == NULL)
				ptrace(PTRACE_GETEVRREGS, processId, fake_addr, &extended1Regs);
			else
				ptrace(PTRACE_GETEVRREGS, tState->fLwProcessId, fake_addr, &extended1Regs);
			
			if ( errno == 0 )
			{
			  if (read)
			  {
			    __puts("Executing ptrace(PTRACE_GETEVRREGS)\n");
				err = MessageAddBlock_ui32( b, start, count );
			  }
			  else
			  {
			    __puts("Executing ptrace(PTRACE_SETEVRREGS)\n");
				err = MessageGetBlock_ui32( b, start, count );
				if (pState->fThreadAgent == NULL)
					ptrace(PTRACE_SETEVRREGS, processId, fake_addr, &extended1Regs);
				else
					ptrace(PTRACE_SETEVRREGS, tState->fLwProcessId, fake_addr, &extended1Regs);
			  }
			}

			if (errno != 0)
			{
				err = kOsError;
				__perror( "ptrace e500 extended register access failed" );
			}

            for ( i = 0; i < sizeof(Extended1_PPC_Linux) / sizeof(ui32); i++ )
            {
             	__puts(" EGPR");
             	__puthex2(i);
             	__puts(": ");
             	__puthex8( *(((ui32 *)&extended1Regs) + i) );          
             	__puts("\t");
             	if ( i % 5 == 4 )
             	  __puts("\n");
            }
            __puts("\n");
          
		}
	
	}
	
	return( err );

#else

#if __MWERKS__
	#pragma unused(firstRegister, lastRegister, b, read, processId, threadId)
#endif /* __MWERKS__ */
	
	*registerStorageSize = 0;
	
	return kUnsupportedError;		// no extended1 regs under Linux

#endif
}						


/****************************************************************************/
/*
 *	OsTargetAccessExtended2
 *
 *	R/W a range in the extended2 register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	OsTargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
						MessageBuffer *b, size_t* registerStorageSize,
						bool read,
						ui32 processId,
						ui32 threadId )
{
#if __MWERKS__
	#pragma unused(firstRegister, lastRegister, b, read, processId, threadId)
#endif /* __MWERKS__ */
	
	*registerStorageSize = 0;
	
	return kUnsupportedError;		// no extended1 regs under Linux
}						


/****************************************************************************/
/*
 *	OsTargetGetPC
 *
 *	Get the PC for the specified thread.
 */
/****************************************************************************/
PCType OsTargetGetPC( ui32 processId, ui32 threadId )
{
	DefaultType value;

	OsTargetAccessDefaultSingle( TRK_DEFAULT_PC,
			&value, true, processId, threadId );

	return( (PCType)value );
}						



/****************************************************************************/
/*
 *	OsTargetAdjustBreakPC
 *
 *	When the target stops at a breakpoint, the PC may or may not be
 *	where we want it.  Adjust accordingly.
 *
 *  NOTE1: This should be called once after each breakpoint is hit,
 *		   before the PC is used for anything. In particular before
 *		   sending out a stop notification or performing a continue.
 *
 */
/****************************************************************************/
DSError OsTargetAdjustBreakPC( ui32 processId, ui32 threadId )
{
	/*
	** For PPC, nothing to do.
	*/
}

/****************************************************************************/
/*
 *	OsTargetSetBreakpoint
 *
 *	Sets a breakpoint according to the given breakpoint descriptor.
 */
/****************************************************************************/
DSError OsTargetSetBreakpoint( Breakpoint *b )
{
	DSError result = kNoError;
	InstructionType instr;
	
	/*
	** When fork() is detected, TRK must remove all of the parent's
	** breakpoints from the child.  However, some breakpoints may be removed
	** from the parent after the fork() but before TRK detects the fork().
	** This is bad because the child inherits the breakpoint, but TRK no
	** longer sees it in the parent so it can't remove the bp from the
	** child.  Use delayed-release to retain the removed bp info for a short
	** time in case this happens.  Set the fReleaseDelay to the number of 
	** delay cycles to wait before truly releasing a breakpoint.  A "cycle"
	** is target-defined.  In the Linux case, each time the process stops is
	** 1 cycle.  1 delay cycle *should* be enough for fork().
	*/

	b->fReleaseDelay = 1;
	
	if (b->fHardware)
	{
		/*
		** No hardware break support yet.
		*/
		__puts( "Attempting to set HW breakpoint: unsupported\r\n" );
		result = kUnsupportedError;
	}
	else
	{
		result = DefaultTargetSetSoftwareBreakpoint( b, PPC_BREAK );
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	OsTargetClearBreakpoint
 *
 *	Clear the breakpoint described in the given breakpoint object.
 */
/****************************************************************************/
DSError OsTargetClearBreakpoint( Breakpoint *b )
{
	DSError result = kNoError;
	InstructionType instr;
	
	if (b->fHardware)
	{
		/*
		** No hardware break support yet.
		*/
		result = kUnsupportedError;
	}
	else
	{
		result = DefaultTargetClearSoftwareBreakpoint( b );
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	OsTargetHideBreakpoint
 *
 *	Hide the breakpoint described in the given breakpoint object.
 */
/****************************************************************************/
DSError OsTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  )
{
	return( DefaultTargetHideBreakpoint( b, data, maxLen ) );
}

/****************************************************************************/
/*
 *	OsTargetAccessInstruction
 *
 *	Read instruction from memory, determining instruction size.  The
 *  storage pointed to by readData must be big enough to store the
 *  largest possible instruction.
 *
 *	Note: if maxSize is -1, the size is unlimited
 */
/****************************************************************************/
DSError OsTargetAccessInstruction(InstructionType* data,
								void* virtualAddress, size_t maxSize,
								MemoryAccessOptions accessOptions,
								bool read,
								ui32 processId, ui32 threadId)
{
	DSError err = kNoError;
	size_t length;
	size_t outlength;
	
	/*
	** Determine instruction size.  Instructions on PPC are 32 bits.
	** Bound it by maxSize.
	*/

	length = sizeof( InstructionType );
	if (maxSize != -1 && maxSize < length)
		length = maxSize;


	/*
	** Read/write the instruction in memory.  If we're doing this in the
	** MetroTRK process, then just use memcpy.
	*/
	
	outlength = length;
	if (processId != getpid())
	{
		err = OsTargetAccessMemory( data, virtualAddress, &outlength,
									accessOptions, read,
									processId, threadId);
	}
	else
	{
		if (read)
			memcpy( data, virtualAddress, length );
		else
			memcpy( virtualAddress, data, length );
	}
	
	/*
	 * Verify that an entire instruction was read/written.
	 */
	 
	if (err == kNoError && length != outlength)
		err = kInvalidMemory;
		
	return( err );
}

unsigned int GetVersionRegister()
{
  FILE *file;
  unsigned int val = 0, val_hi, val_lo;
  unsigned char line[256];
  
  file = fopen("/proc/cpuinfo", "r");
  
  if ( file == NULL )
    return val;
  
  while ( !feof(file) )
  {
  	fgets(line, 255, file);
  	if ( (strstr(line, "revision") != NULL) && (strstr(line, "pvr") != NULL) )
  	{
  		sscanf(strstr(line, "pvr") + strlen("pvr"), "%x %x", &val_hi, &val_lo);
  		val = (val_hi << 16) | val_lo;
  		__puts("PVR_hi = ");
  		__puthex4(val_hi);
  		__puts("\tPVR_lo = ");
  		__puthex4(val_lo);
  		__puts("\tPVR = ");
  		__puthex8(val);
  		__puts("\n");
  		break;
   	}
  }
  
  fclose(file);
  return val;
}
