/******************************************************************************************/
/*

DESCRIPTION

	Target-independent breakpoint implementation.

COPYRIGHT

	Copyright 2000-2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "target.h"
#include "msgcmd.h"
#include "break.h"
#include "nubassrt.h"
#include "targimpl.h"

#ifndef _NO_SYSTEM_INCLUDES
#include <stdlib.h>
#endif

#if TRK_SUPPORT_BREAKPOINTS

static int kBreakpointCount = kInitialBreakpointCount;

/*
** Set this to 1 to see detailed information about normal breakpoint
** processing.  (error msgs output via __puts() when one of the
** DEBUGIO_ macros is defined will always be emitted regardless of
** whether DEBUG_BREAK is (1) or (0) .)
*/

#define DEBUG_BREAK	(1)


/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

typedef struct Breakpoints {
	DSMutex			fMutex;
	Breakpoint 		*pBreakpoints;
	BreakpointId 	fMaxUsed;
} Breakpoints;

Breakpoints gBreakpoints;

/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/

static Breakpoint *GetBreakpointImplementation(Breakpoint *bp);
static void SetBreakpointUsed(BreakpointId i, bool inUse);
static void SetBreakpointEnabled(Breakpoint *bp, bool enabled);
static bool GetBreakpointThreadLocalImp(Breakpoint *bp);

/****************************************************************************/
/*
	GetBreakpointImplementation
*/
/****************************************************************************/

static Breakpoint *GetBreakpointImplementation(Breakpoint *bp)
{
	Breakpoint *shared;
	ASSERT(bp);
	if (bp->fSharedId != kInvalidBreakpointId)
	{
		shared = GetBreakpoint( bp->fSharedId );
		ASSERT(shared);
		return(shared);
	}
	else
	{
		return(bp);
	}
}

/****************************************************************************/
/*
	SetBreakpointUsed
*/
/****************************************************************************/

static void SetBreakpointUsed(BreakpointId i, bool inUse)
{
	Breakpoint *b;
	BreakpointId j;
	
	b = GetBreakpoint( i );

	ASSERT(b);
	
	b->fInUse = inUse;
	
	/*
	** Update max used field.  If setting this one to true, 
	** then set new maxUsed if it is beyond the max.  If 
	** setting this one to false and it is the current max,
	** then find previous used.
	*/
	AcquireMutex( &gBreakpoints.fMutex );
	if (inUse)
	{
		if (gBreakpoints.fMaxUsed < i )
			gBreakpoints.fMaxUsed = i;
	}
	else
	{
		if (gBreakpoints.fMaxUsed == i)
		{
			/*
			** Assume no others are used, set the max below all valid ID's
			*/
			gBreakpoints.fMaxUsed = -1;
			
			/*
			** Work backwards from the previous max.  If a used bp is
			** found, make it the new max and stop.
			*/
			for (j = i - 1; j >= 0; --j)
			{
				b = GetBreakpoint( j );
				if (b->fInUse)
				{
					gBreakpoints.fMaxUsed = j;
					break;
				}
			}
		}
	}
	ReleaseMutex( &gBreakpoints.fMutex );
}

/****************************************************************************/
/*
	SetBreakpointEnabled
*/
/****************************************************************************/

static void SetBreakpointEnabled(Breakpoint *bp, bool enabled)
{
	ASSERT(bp);
	GetBreakpointImplementation( bp )->fEnabled = enabled;
}

/****************************************************************************/
/*
	GetBreakpointEnabled
*/
/****************************************************************************/

bool GetBreakpointEnabled(Breakpoint *bp)
{
	ASSERT(bp);
	return( GetBreakpointImplementation( bp )->fEnabled );
}

/****************************************************************************/
/*
	GetBreakpointThreadLocalImp
*/
/****************************************************************************/

static bool GetBreakpointThreadLocalImp(Breakpoint *bp)
{
	ASSERT(bp);
	return( GetBreakpointImplementation( bp )->fThreadLocalImp );
}

/****************************************************************************/
/*
	InitializeBreakpoint
*/
/****************************************************************************/

DSError InitializeBreakpoint( BreakpointId i )
{
	Breakpoint *bp;
	
	bp = GetBreakpoint( i );
	ASSERT( bp );
	
	InitializeMutex(&(bp->fMutex));
	AcquireMutex(&(bp->fMutex));
	
	SetBreakpointUsed(i, false);
	bp->fEnabled = false;
	bp->fPrevId = kInvalidBreakpointId;
	bp->fNextId = kInvalidBreakpointId;
	bp->fSharedId = kInvalidBreakpointId;
	bp->fReleaseDelay = 0;

	ReleaseMutex(&(bp->fMutex));

	return kNoError;
}

/****************************************************************************/
/*
	InitializeBreakpoints
*/
/****************************************************************************/

DSError InitializeBreakpoints(void)
{
	DSError err = kNoError;
	int i;
	Breakpoint *pointer;
	
	/*
	 * Here for the first time,a fixed count of BP structures
	 * created and the allocation is done using TRK_calloc(). 
	 * As the requests for more BP increase, we incrementally
	 * reassign by calling TRK_realloc() when the
	 * current limit is exhausted.  But all these upto a 
	 * maximum of 4000 (which is faily large)
	 * so that the limit is to some fixed count.
	 */
	pointer  = (void *)TRK_calloc( kBreakpointCount, sizeof(Breakpoint));
	if( pointer == NULL )
		return kBreakpointsFull;				/* Error case */
	else
		gBreakpoints.pBreakpoints = pointer; 
	
	InitializeMutex(&gBreakpoints.fMutex);
	gBreakpoints.fMaxUsed = -1;
	for (i = 0; i < kBreakpointCount && err == kNoError; i++)
		InitializeBreakpoint( i );
	return( err );
}


/****************************************************************************/
/*
	ReInitializeBreakpoints - If all breakpoints available are used, then 
	increase the breakpoint count "kBreakpointCount" by "kBreakpointCountIncrement". 
	After incrementing kBreakpointCount, call TRK_realloc memory function to allocate
	memory for the incremented breakpoint count and replace earlier breakpoint
	pointer by this pointer (which points to increased breakpoint memory area)
	
	When you enter this function, this should happen only when already
	the BPs limit got exhausted.  No further check is done in this routine
	due to this assumption.
	
	This returns failure only if the TRK_realloc() fails and otherwise it returns true.
*/
/****************************************************************************/

DSError ReInitializeBreakpoints(BreakpointId* resultBreakpointId,
								Breakpoint** resultBreakpoint)
{
	BreakpointId i = kBreakpointCount;
	
	Breakpoint* pointer;
	int j;
	
	/* Allocate more space for "kBreakpointCountIncrement" new breakpoints */
	kBreakpointCount += kBreakpointCountIncrement ;



	pointer  = (void *)TRK_realloc( gBreakpoints.pBreakpoints, (kBreakpointCount * sizeof(Breakpoint)) );
	if( pointer == NULL )
		return kBreakpointsFull;	/* Error case */
	else			
		gBreakpoints.pBreakpoints = pointer;

	// Initialize the newly created breakpoints
	for (j = i ; j < kBreakpointCount; j++)
		InitializeBreakpoint( j );
	
	return _GetFreeBreakpoint(resultBreakpointId, resultBreakpoint);
	
}

/****************************************************************************/
/*
	FreeBreakpoint
*/
/****************************************************************************/

void FreeBreakpoint(void)
{
	free(gBreakpoints.pBreakpoints);
}

/****************************************************************************/
/*
	GetFreeBreakpoint
*/
/****************************************************************************/

DSError _GetFreeBreakpoint(BreakpointId* resultBreakpointId,
							Breakpoint** resultBreakpoint)
{
	DSError result = kBreakpointsFull;		/* assume failure */
	BreakpointId i;
	Breakpoint* b;
	
	*resultBreakpoint = NULL;
	
	/*
		linear search through the breakpoints to find one that is not in use
	*/
	
	for (i = 0; i < kBreakpointCount; i++) {
		b = GetBreakpoint(i);
		AcquireMutex(&b->fMutex);
		
		if (b->fInUse == false) {
		
			/* initialize breakpoint */
			SetBreakpointUsed(i, true);
			b->fEnabled = false;
			b->fPrevId = kInvalidBreakpointId;
			b->fNextId = kInvalidBreakpointId;
			b->fSharedId = kInvalidBreakpointId;
			b->fReleaseDelay = 0;
			
			*resultBreakpoint = b;
			*resultBreakpointId = i;
			result = kNoError;
			
			i = kBreakpointCount;		/* terminate iteration */
		}
		
		ReleaseMutex(&b->fMutex);
	}
	
	/* If result is kBreakpointsFull and kBreakpointCount is equal to
       kBreakpointCountMax then return error = kBreakpointsFull
       
       But if result is kBreakpointsFull but kBreakpointCount is not equal
       to kBreakpointCountMax, then we call ReInitializeBreakpoints function
       where we try to increase the memory for breakpoints structure.
     */
       
       
	if( result == kBreakpointsFull )
	{
		if( kBreakpointCount >= kBreakpointCountMax )
			return result;		/* Error: No more allocation possible */
		else
		{
#if DEBUG_BREAK
				__puts( "ReInitializeBreakpoints:Old Count=" );
				__puthex8( kBreakpointCount );
				__puts( "\r\n" );
#endif

			result = ReInitializeBreakpoints(resultBreakpointId,resultBreakpoint );

#if DEBUG_BREAK
				__puts( "ReInitializeBreakpoints:New Count=" );
				__puthex8( kBreakpointCount );
				__puts( " Error=" );
				__puthex8( result);
				__puts( "\r\n" );
#endif
		}
	}
	
	return result;
}

/****************************************************************************/
/*
	GetBreakpoint
	
	Get a pointer to a single breakpoint structure
*/
/****************************************************************************/

Breakpoint *GetBreakpoint(BreakpointId breakpointId)
{
	Breakpoint* result = NULL;
	
	if (breakpointId >= 0 && breakpointId < kBreakpointCount) {
		result = &gBreakpoints.pBreakpoints[breakpointId];
	}
	
	return result;
}

/****************************************************************************/
/*
	ReleaseBreakpoint
	
	Mark the breakpoint as unused
*/
/****************************************************************************/

void _ReleaseBreakpoint(BreakpointId breakpointId)
{
	Breakpoint* b;
	
	if (breakpointId == kInvalidBreakpointId)
		return;
	
	if (breakpointId >= 0 && breakpointId < kBreakpointCount) {

		b = &gBreakpoints.pBreakpoints[breakpointId];

		/*
		** If fReleaseDelay is 0, the breakpoint can be released
		** immediately.
		**
		** If it is positive, then this is the initial release
		** attempt.  Don't release immediately, but start the
		** release countdown by negating the value.
		**
		** If it is negative, then the countdown has begun.
		** Increment it and if the result is zero release the
		** breakpoint.
		*/
		
		if (b->fReleaseDelay < 0)
			b->fReleaseDelay++;
			
		if (b->fReleaseDelay == 0)
		{
			UpdateSharedBreakpoints( breakpointId );

			AcquireMutex(&b->fMutex);
			SetBreakpointUsed(breakpointId, false);
			ReleaseMutex(&b->fMutex);		
		}
		else if (b->fReleaseDelay > 0)
		{
			b->fReleaseDelay = -b->fReleaseDelay;
		}
		
	}
}

/****************************************************************************/
/*
	CheckBreakpointExistence
	
	Check if a breakpoint already exists at a given address.
*/
/****************************************************************************/
DSError CheckBreakpointExistence(unsigned int start, unsigned int len,
						         bool hardware, BreakpointCond conditions,
						         unsigned count, ui8 targetOptions,
						         ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	BreakpointId foundId;
	Breakpoint *found;
		
	/*
	** If the execute condition is specified and this breakpoint overlaps
	** with an existing one, allow it but share with the existing
	** breakpoint.  This means that only the original breakpoint keeps
	** the restore information (saved instruction).
	*/
	if ((conditions & DS_MSG_BREAK_CONDITION_EXECUTE) &&
		FindExecBreakpoint( &foundId, &found, start, len, true, false, processId, threadId )
			== kNoError)
	{
		/*
		** Delayed release breakpoints are an irritating special case.  If one
		** of them overlaps the new one and the new one will be detected by 
		** ClearInheritedBreakpoints(), the delayed release one can actually be
		** released immediately.
		*/

		if (found->fReleaseDelay < 0 && found->fSharedId == kInvalidBreakpointId &&
			found->fProcessId == processId &&
			(found->fThreadId == threadId || threadId == DS_MSG_BREAK_THREAD_ID_ALL))
		{
		}
		else
		{
#if DEBUG_BREAK
				__puts( "CheckBreakpointExistence: New breakpoint confilicts w/existing breakpoint: " );
				__puthex8( foundId );
				__puts(" at address 0x");
				__puthex8(start);
				__puts( "\r\n" );
#endif
            
            if ( GetBreakpointEnabled(found) ) //return conflict if and only if the existing breakpoint is enabled
			{
              err = kBreakpointConflict;
			}
		}
	}
	
	return( err );
}

/****************************************************************************/
/*
	SetBreakpoint
	
	Set a breakpoint.
*/
/****************************************************************************/
DSError SetBreakpoint( Breakpoint *b, unsigned int start, unsigned int len,
						bool hardware, BreakpointCond conditions,
						unsigned count, ui8 targetOptions,
						ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	BreakpointId foundId;
	Breakpoint *found;
	
	ASSERT(b);
	
	/*
	** If the execute condition is specified and this breakpoint overlaps
	** with an existing one, allow it but share with the existing
	** breakpoint.  This means that only the original breakpoint keeps
	** the restore information (saved instruction).
	*/
	if ((conditions & DS_MSG_BREAK_CONDITION_EXECUTE) &&
		FindExecBreakpoint( &foundId, &found, start, len, true, false, processId, threadId )
			== kNoError)
	{
		/*
		** Delayed release breakpoints are an irritating special case.  If one
		** of them overlaps the new one and the new one will be detected by 
		** ClearInheritedBreakpoints(), the delayed release one can actually be
		** released immediately.
		*/
		if (found->fReleaseDelay < 0 && found->fSharedId == kInvalidBreakpointId &&
			found->fProcessId == processId &&
			(found->fThreadId == threadId || threadId == DS_MSG_BREAK_THREAD_ID_ALL))
		{
			found->fReleaseDelay = 0;
			ReleaseBreakpoint( foundId );
		}
		else
		{
			if (found->fSharedId != kInvalidBreakpointId)
			{
				b->fSharedId = found->fSharedId;
			}
			else
			{
				b->fSharedId = foundId;
				found->fSharedId = foundId;
			}

#if DEBUG_BREAK
				__puts( "SetBreakpoint: New breakpoint shared w/existing breakpoint: " );
				__puthex8( b->fSharedId );
				__puts( "\r\n" );
#endif
		}
	}

	if (err == kNoError)
	{	
		b->fStart = start;
		b->fLength = len;
		b->fCount = count;
		b->fHardware = hardware;
		b->fConditions = conditions;
		b->fProcessId = processId;
		b->fThreadId = threadId;
		b->fTargetOptions = targetOptions;
		err = TargetValidateBreakpoint(b);
		if (err == kNoError)
			err = EnableBreakpoint(b);
	}
	
	return( err );
}

/****************************************************************************/
/*
	ClearBreakpoint
	
	Clear (unset) a breakpoint that has been previously set.
*/
/****************************************************************************/
DSError ClearBreakpoint( Breakpoint *b )
{
	DSError err = kNoError;
	
	ASSERT(b);
	
	if (GetBreakpointEnabled( b ))
		err = DisableBreakpoint( b );
	
	return( err );
}

/****************************************************************************/
/*
	EnableBreakpoint
	
	Enable a breakpoint that has been previously set and disabled.
*/
/****************************************************************************/
DSError EnableBreakpoint( Breakpoint *b )
{
	DSError err = kNoError;
	
	ASSERT(b);
	
	AcquireMutex(&(GetBreakpointImplementation(b)->fMutex));
	if (!GetBreakpointEnabled( b ))
	{
		err = XTargetSetBreakpoint( GetBreakpointImplementation( b ) );

		if (err == kNoError)
			SetBreakpointEnabled( b, true );
	}
	ReleaseMutex(&(GetBreakpointImplementation(b)->fMutex));

	return( err );
}

/****************************************************************************/
/*
	DisableBreakpoint
	
	Disable a breakpoint that has been previously set.  Once a breakpoint
	has been disabled, it will no longer trigger when its condition is met.
	However, it retains all of the information required to re-enable it
	later.
*/
/****************************************************************************/
DSError DisableBreakpoint( Breakpoint *b )
{
	DSError err = kNoError;
	
	ASSERT(b);
	
	AcquireMutex(&(GetBreakpointImplementation(b)->fMutex));
	if (GetBreakpointEnabled( b ))
	{
		err = XTargetClearBreakpoint( GetBreakpointImplementation( b ) );

		if (err == kNoError)
			SetBreakpointEnabled( b, false );
	}
	ReleaseMutex(&(GetBreakpointImplementation(b)->fMutex));
	
	return( err );
}

/****************************************************************************/
/*
	FindExecBreakpoint
	
	Find an execution breakpoint that covers any part of the
	specified address range.  If there are multiple breakpoints which 
	meet this condition, this function will only find one of them.
	
	If the threadSpecific flag is false, then this function will return
	any breakpoint which could stop the given thread, even if that breakpoint
	is not specifically assigned to that thread.  If the threadSpecific
	flag is true, then this function will only return a breakpoint which
	is specifically assigned to the given thread.

*/
/****************************************************************************/
DSError FindExecBreakpoint( BreakpointId* resultBreakpointId,
							Breakpoint** resultBreakpoint,
							unsigned int start, unsigned int len,
							bool requireEnabled, bool threadSpecific,
							ui32 processId, ui32 threadId )
{
	DSError err = kBreakpointNotFound;
	BreakpointId i;
	Breakpoint *b;

#if DEBUG_BREAK
		__puts( "FindExecBreakpoint: s" );
		__puthex8( start );
		__puts( ":l" );
		__puthex8( len );
		__puts( " (p" );
		__puthex8( processId );
		__puts( ":t" );
		__puthex8( threadId );
		__puts( ")\r\n" );
#endif

	for (i = gBreakpoints.fMaxUsed; i >= 0 && err == kBreakpointNotFound; i--)
	{
		b = GetBreakpoint( i );
		
		/*
		** If we find a breakpoint that is in use, enabled, has the
		** execute condition, and the memory regions overlap, then
		** return that breakpoint.
		*/
		
#if DEBUG_BREAK
			__puts( "  checking: u" );
			__puthex2( b->fInUse );
			__puts( ":e" );
			__puthex2( GetBreakpointEnabled( b ) );
			__puts( ":x" );
			__puthex2( (b->fConditions & DS_MSG_BREAK_CONDITION_EXECUTE) != 0 );
			__puts( ":li" );
			__puthex2( GetBreakpointThreadLocalImp( b ) );
			__puts( " s" );
			__puthex8( b->fStart );
			__puts( ":l" );
			__puthex8( b->fLength );
			__puts( " (p" );
			__puthex8( b->fProcessId );
			__puts( ":t" );
			__puthex8( b->fThreadId );
			__puts( ")\r\n" );
#endif

		/*
		** Is it an enabled exec breakpoint?
		*/
		if (b->fInUse && (!requireEnabled || GetBreakpointEnabled( b )) &&
			(b->fConditions & DS_MSG_BREAK_CONDITION_EXECUTE))
		{
			/*
			** Does it match the generic criteria?
			*/
#ifdef __uClinux__			
			if ( ((!threadSpecific)||(b->fProcessId == processId)) &&
#else			
			if (b->fProcessId == processId &&  
#endif //__uClinux__			
				((!threadSpecific && !GetBreakpointThreadLocalImp( b )) ||
				 b->fThreadId == threadId || 
				 (!threadSpecific && b->fThreadId == DS_MSG_BREAK_THREAD_ID_ALL)) &&
				((start + len) > b->fStart && start < (b->fStart + b->fLength)))
			{
				err = kNoError;
			}
			else 
			{
			
				/*
	  			 ** If not, does it match target-specific criteria?
				 */
				if (!threadSpecific &&
					 	XTargetMatchExecBreakpoint( b, start, len
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
													, processId, threadId
#endif
												   ))
						err = kNoError;
			}
			
			
			if (err == kNoError)
			{
				*resultBreakpointId = i;
				*resultBreakpoint = b;
			}
		}
	}
	
	__puts( (err == kNoError) ? "  BP Found\r\n" : "  BP Not Found\r\n" );
	
	return( err );
}

/****************************************************************************/
/*
	ReleaseThreadBreakpoints
	
	Release all breakpoints for a specific thread.

	If the clear flag is true:
		- If the weak flag is false, each breakpoint will be cleared
		  (unset) before it is released.
		- If the weak flag is true, only the un-shared breakpoints
		  will be cleared (unset) before they are released.

	If the input threadId is DS_MSG_BREAK_THREAD_ID_ALL, then this will
	release all of the breakpoints for the specific process, regardless
	of thread.

*/
/****************************************************************************/
int ReleaseThreadBreakpoints( ui32 processId, ui32 threadId,
								bool clear, bool weak )
{
	BreakpointId i;
	Breakpoint *b;
	DSError err = kNoError;

#if DEBUG_BREAK
		__puts( "ReleaseThreadBreakpoints: " );
		__puthex8( processId );
		__puts( ":" );
		__puthex8( threadId );
		__puts( clear ? ",clear" : "" );
		__puts( weak ? ",weak" : "" );
		__puts( "\r\n" );
#endif

	for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
	{
		b = GetBreakpoint( i );
		
		/*
		** If we find a breakpoint that is in use and associated with
		** the specified thread, release it and continue searching.
		** Depending on the clear and weak flags, it may be cleared
		** before it is released.
		*/
		
		if (b->fInUse && b->fProcessId == processId &&
			(b->fThreadId == threadId || threadId == DS_MSG_BREAK_THREAD_ID_ALL))
		{
			if (clear && (!weak || b->fSharedId == kInvalidBreakpointId))
				ClearBreakpoint( b );
			b->fReleaseDelay=0; /*[rjeswani] Force releasing of breakpoints*/
			ReleaseBreakpoint( i );
			
		}
	}
	
	return (err);
}

/****************************************************************************/
/*
	ReleaseAllBreakpoints
	
	Release all breakpoints in the system.

	If the clear flag is true:
		- If the weak flag is false, each breakpoint will be cleared
		  (unset) before it is released.
		- If the weak flag is true, only the un-shared breakpoints
		  will be cleared (unset) before they are released.

*/
/****************************************************************************/
void ReleaseAllBreakpoints( bool clear, bool weak )
{
	BreakpointId i;
	Breakpoint *b;

#if DEBUG_BREAK
		__puts( "ReleaseAllBreakpoints\r\n" );
#endif

	for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
	{
		b = GetBreakpoint( i );

		/*
		** If we find a breakpoint that is in use, release it and
		** continue searching.
		** Depending on the clear and weak flags, it may be cleared
		** before it is released.
		*/
		
		if (b->fInUse)
		{
			if (clear && (!weak || b->fSharedId == kInvalidBreakpointId))
				ClearBreakpoint( b );
			ReleaseBreakpoint( i );
		}
	}
}

/****************************************************************************/
/*
	HideBreakpoints
	
	Given a buffer containing the contents of a region of memory in a target
	program, hide all of the breakpoints so that the buffer represents
	what the memory contents would be if there were no breakpoints set.

*/
/****************************************************************************/
DSError HideBreakpoints( void* data,
						unsigned int start,
						unsigned int len,
						ui32 processId,
						ui32 threadId )
{
	DSError err = kNoError;
	BreakpointId i;
	Breakpoint *b;
	unsigned int dOffset;
	unsigned int dMaxLen;

	for (i = gBreakpoints.fMaxUsed; i >= 0 && err == kNoError; i--)
	{
		b = GetBreakpoint( i );
		
		/*
		** If we find a breakpoint that is in use, enabled, and the memory
		** regions overlap, then hide that breakpoint.
		*/
		
		ASSERT(b);

		if (b->fInUse && GetBreakpointEnabled( b ) &&
			b->fProcessId == processId && 
			(!GetBreakpointThreadLocalImp( b ) || b->fThreadId == threadId ||
			 b->fThreadId == DS_MSG_BREAK_THREAD_ID_ALL) &&
			((start + len) > b->fStart && start < (b->fStart + b->fLength)))
			{
			
				/*
				** Get the start offset and max length of the hide point in
				** the data buffer.  Make sure it doesn't go outside
				** the boundaries of the buffer (no negative offset,
				** and offset + max length must not exceed the total buffer
				** length).
				*/
				if (start <= b->fStart)
				{
					dOffset = b->fStart - start;
					dMaxLen = len - dOffset;
				}
				else
				{
					dOffset = 0;
					dMaxLen = len;
				}
				
				err = XTargetHideBreakpoint(
						GetBreakpointImplementation( b ),
						(ui8 *)data + dOffset, dMaxLen );				
			}
	}
	
	
	return( err );
}

/****************************************************************************/
/*
	UpdateSharedBreakpoints
	
	The given breakpoint is about to be released.  Update accordingly 
	all breakpoints which share the given breakpoint.

*/
/****************************************************************************/
void UpdateSharedBreakpoints( BreakpointId releasedId )
{
	BreakpointId i;
	Breakpoint *b;
	Breakpoint *released;
	BreakpointId alternateId;
	Breakpoint *alternate = NULL;
	

	alternateId = kInvalidBreakpointId;
	
	released = GetBreakpoint( releasedId );
	ASSERT( released );

	/*
	** If not shared, there is nothing to update.
	*/
	
	if (released->fSharedId == kInvalidBreakpointId)
		return;

#if DEBUG_BREAK
		__puts( "UpdateSharedBreakpoints: " );
		__puthex8( releasedId );
		__puts( "\r\n" );
#endif

	if (released->fSharedId == releasedId)
	{
		/*
		** The one to be released is the keeper of the implementation and
		** enabled info.  Find an alternate to hold this info.
		*/
	
		for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
		{
			b = GetBreakpoint( i );
			ASSERT(b);

			/*
			** If this breakpoint is sharing the one that is
			** about to be released (and it is not the one that
			** is about to be released):
			*/
			if (b->fInUse && b->fSharedId == releasedId &&
				i != releasedId)
			{
				if (alternateId == kInvalidBreakpointId)
				{
					/*
					**	It is the first one found, it should
					**	become the new holder of the breakpoint
					**	implementation info and status.
					*/
					alternateId = i;
					alternate = b;
					
					alternate->fRestore = released->fRestore;
					alternate->fEnabled = released->fEnabled;
				}
				else
				{
					/*
					** For any others, redirect them to the first
					** alternate found.
					*/
					b->fSharedId = alternateId;
					alternate->fSharedId = alternateId;
				}
			}
				
		}
	}
	else if (released->fSharedId != kInvalidBreakpointId)
	{
		/*
		** The one to be released is sharing another.  If it is the last
		** one, then restore the sharee back to a normal breakpoint.
		*/

		for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
		{
			if (i != releasedId && i != released->fSharedId)
			{
				b = GetBreakpoint( i );
				ASSERT(b);
				if (b->fSharedId == released->fSharedId)
					break; /* found another one */
			}
		}
		if (i < 0) /* no other found */
		{
			b = GetBreakpoint( released->fSharedId );
			ASSERT(b);
			b->fSharedId = kInvalidBreakpointId;
		}
	}
}

/****************************************************************************/
/*
 *	DefaultTargetSetSoftwareBreakpoint
 *
 *	Sets a software breakpoint according to the given breakpoint descriptor,
 *	assuming a default implementation in which the breakpoint is set by 
 *	replacing the current instruction with a provided breakpoint instruction.
 */
/****************************************************************************/
DSError DefaultTargetSetSoftwareBreakpoint( Breakpoint *b,
											InstructionType instr )
{
	DSError result = kNoError;
	
	/*
	** Software breakpoints can only break on the execution condition.
	** Also, they only support a length of 1.
	*/
	
	if (((b->fConditions & (DS_MSG_BREAK_CONDITION_READ |
							DS_MSG_BREAK_CONDITION_WRITE |
							DS_MSG_BREAK_CONDITION_MODIFY)) != 0) ||
		(b->fLength != 1) || b->fHardware)
	{
		__puts( "Attempting to set SW breakpoint w/non-execute condition.\r\n" );
		result = kUnsupportedError;
	}

	if (result == kNoError)
	{
		/*
		** For software breakpoints, store the old instruction and
		** write out a breakpoint instruction at that address.
		*/
#if DEBUG_BREAK
			__puts( "Setting SW breakpoint at " );
			__puthex8( b->fStart );
			__puts( " (" );
			__puthex8( b->fProcessId );
			__puts( ":" );
			__puthex8( b->fThreadId );
			__puts( ")\r\n" );
#endif
		
		result = XTargetAccessInstruction( (InstructionType *)&b->fRestore,
					(void *)b->fStart,
					-1, kDebuggerMemory, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
					,b->fProcessId, b->fThreadId 
#endif
					);

		if (result == kNoError)
		{
			result = XTargetAccessInstruction( &instr, (void *)b->fStart,
						-1, kDebuggerMemory, false 
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						,b->fProcessId, b->fThreadId 
#endif
					);
		}
		
		/*
		** Assume that all threads in a process share the same memory.
		** This means that they will all be affected by this breakpoint,
		** since it was set by writing memory.  Therefore, this is not
		** a thread-local implementation.
		*/
		
		b->fThreadLocalImp = false;
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	DefaultTargetClearSoftwareBreakpoint
 *
 *	Clears a software breakpoint according to the given breakpoint descriptor,
 *	assuming a default implementation in which the breakpoint is set by 
 *	replacing the current instruction with a breakpoint instruction.
 */
/****************************************************************************/
DSError DefaultTargetClearSoftwareBreakpoint( Breakpoint *b )
{
	DSError result = kNoError;
	
	/*
	** For software breakpoints, just write back the original saved
	** instruction.
	*/
#if DEBUG_BREAK
		__puts( "Clearing SW breakpoint at " );
		__puthex8( b->fStart );
		__puts( " (" );
		__puthex8( b->fProcessId );
		__puts( ":" );
		__puthex8( b->fThreadId );
		__puts( ")\r\n" );
#endif

	result = XTargetAccessInstruction( (InstructionType *)&b->fRestore,
				(void *)b->fStart, 
				-1, kDebuggerMemory, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
				,b->fProcessId, b->fThreadId 
#endif
				);

	return( result );
}

/****************************************************************************/
/*
 *	DefaultTargetHideBreakpoint
 *
 *	Hide the breakpoint described in the given breakpoint object,
 *	assuming a default implementation in which the breakpoint is set by 
 *	replacing the current instruction with a breakpoint instruction.
 */
/****************************************************************************/
DSError DefaultTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  )
{
	DSError result = kNoError;
	
	if (!b->fHardware)
	{
		/*
		** For software breakpoints, write the saved instruction to the 
		** buffer.
		*/
#if DEBUG_BREAK
			__puts( "Hiding SW breakpoint at " );
			__puthex8( b->fStart );
			__puts( "\r\n" );
#endif

		result = XTargetAccessInstruction( (InstructionType *)&b->fRestore,
					data, maxLen,
					kDebuggerMemory, false 
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
					,TRK_PROCESS_ID, TRK_THREAD_ID 
#endif
					);

	}
	
	return( result );
}

/****************************************************************************/
/*
	AddSubBreakpoint
	
	Add a sub-breakpoint to an existing breakpoint.  This is used when
	a thread-global breakpoint is set on a group of threads but has a
	thread-local implementation.  Each of the sub-breakpoints represents
	a breakpoint set on a specific thread, while the main breakpoint 
	represents the entire group and is really just a placeholder in this
	case.
	
	This routine does not set or enable the new breakpoint.  It only
	sets up the structure and puts it in the appropriate sub-breakpoint
	list.
	
*/
/****************************************************************************/
DSError AddSubBreakpoint( BreakpointId mainBreakpointId, ui32 threadId,
							BreakpointId* subBreakpointId,
							Breakpoint** subBreakpoint )
{
	DSError err = kNoError;
	Breakpoint *b;
	
	b = GetBreakpoint( mainBreakpointId );
	ASSERT(b);
	
	/*
	** Get a new breakpoint and copy all of the parameters from the main
	** breakpoint.  The one exception is the thread ID.  Use the one that
	** was passed in rather than the one from the main breakpoint.
	*/

	err = GetFreeBreakpoint( subBreakpointId, subBreakpoint );
	
	if (err == kNoError)
	{
		(*subBreakpoint)->fStart = b->fStart;
		(*subBreakpoint)->fLength = b->fLength;
		(*subBreakpoint)->fHardware = b->fHardware;
		(*subBreakpoint)->fProcessId = b->fProcessId;
		(*subBreakpoint)->fThreadId = threadId;

		/*
		** Insert it at the head of the sub-breakpoint list under
		** the main breakpoint.
		*/

		(*subBreakpoint)->fNextId = b->fNextId;
		(*subBreakpoint)->fPrevId= mainBreakpointId;
		b->fNextId = *subBreakpointId;
	}

	return( err );
}


/****************************************************************************/
/*
	ClearInheritedBreakpoints
	
	For systems in which a "child" process/thread inherits breakpoints from
	a "parent" (e.g., after a fork() in unix).  For each breakpoint in the
	parent, remove the equivalent breakpoint from the child.
	
	To affect all threads in a process, use DS_MSG_BREAK_THREAD_ID_ALL for
	the thread IDs.
*/
/****************************************************************************/
DSError ClearInheritedBreakpoints( ui32 pProcessId, ui32 pThreadId, 
		ui32 cProcessId, ui32 cThreadId )
{
	DSError err = kNoError;
	ui32 sThreadId;
	ui32 sProcessId;
	bool sEnabled;
	int i;
	Breakpoint *b;
	
	for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
	{
		b = GetBreakpoint( i );
		
		/*
		** If we find a breakpoint which affects the parent thread,
		** retarget it to the child, disable it, retarget back to the
		** parent, and restore the parent's "enabled" state.
		**
		** Note1: It is important to do this even for breakpoints which
		** are disabled in the parent.  This will handle the case where
		** the breakpoint was disabled *after* it was inherited.  This is
		** particularly useful in combination with the fReleaseDelay field.
		** Info on released breakpoints can be retained for a time in case
		** they are inherited by a child and needs to be cleared from it.
		**
		*/
		
		if (b->fInUse && b->fSharedId == kInvalidBreakpointId &&
			b->fProcessId == pProcessId &&
			(b->fThreadId == pThreadId || pThreadId == DS_MSG_BREAK_THREAD_ID_ALL ||
			 b->fThreadId == DS_MSG_BREAK_THREAD_ID_ALL || !b->fThreadLocalImp))
		{
			/* save parent status */
			b = GetBreakpointImplementation( b );
			sProcessId = b->fProcessId;
			sThreadId = b->fThreadId;
			sEnabled = GetBreakpointEnabled( b );

			/* clear in child */
			b->fProcessId = cProcessId;
			b->fThreadId = cThreadId;
			SetBreakpointEnabled( b, true );
			DisableBreakpoint( b );
			
			/* restore parent status */
			b->fProcessId = sProcessId;
			b->fThreadId = sThreadId;
			SetBreakpointEnabled( b, sEnabled );
		}
	}

	return( err );
}

/****************************************************************************/
/*
	ReleaseDelayedBreakpoints
	
	If a target system uses the fReleaseDelay field to delay breakpoint
	release, then use this function to perform another release iteration on
	any breakpoints waiting for release.

	To affect all threads in a process, use DS_MSG_BREAK_THREAD_ID_ALL for
	the thread ID.
*/
/****************************************************************************/
DSError ReleaseDelayedBreakpoints( ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	int i;
	Breakpoint *b;

	for (i = gBreakpoints.fMaxUsed; i >= 0; i--)
	{
		b = GetBreakpoint( i );

		if (b->fInUse && b->fReleaseDelay < 0 && b->fProcessId == processId &&
			(b->fThreadId == threadId || threadId == DS_MSG_BREAK_THREAD_ID_ALL ||
			 b->fThreadId == DS_MSG_BREAK_THREAD_ID_ALL))
			ReleaseBreakpoint( i );	/* increments the count and releases if 0 */
	}
	
	return( err );
}

#endif /* #if TRK_SUPPORT_BREAKPOINTS */
