/******************************************************************************************/
/*

DESCRIPTION
	Breakpoints for CodeWarriorTRK.
	
	Breakpoints are allocated out of a pool of breakpoints, preallocated
	at initialization.
	
	Usage
	
		Call InitializeBreakpoints once to initialize the breakpoints. When a
		new breakpoint is desired, call "GetFreeBreakpoint" and a breakpoint ID
		plus a pointer to the Breakpoint struct is returned.  Call "ResetBreakpoint"
		before using the breakpoint.  When using the Breakpoint, acquire and
		release the mutex before and after using or modifying it.  Once the
		breakpoint is no longer being used, call ReleaseBreakpoint.
	Initialization
	
		DSError InitializeBreakpoints(void);
		
			Call once at the beginning of the program.
	
	Getting and Releasing
	
		DSError GetFreeBreakpoint(BreakpointId* returnedBreakpointId,
									Breakpoint** breakpoint);
		Breakpoint* GetBreakpoint(BreakpointId breakpointId);
		
		
			GetFreeBreakpoint will acquire a breakpoint if it is available.
			Returned values:
			kNoError	the returnedBreakpointId and breakpoint pointers refer
						to the allocated breakpoint
						
			kBreakpointsFull
						no more breakpoints are available.
			
			GetBreakpoint gets a pre-allocated Breakpoint pointer from the
			breakpoint's ID.  If the breakpoint was not allocated or the ID is out
			of range, GetBreakpoint retuns NULL.
			
			
		void ReleaseBreakpoint(BreakpointId id);
			
			When a breakpoint is no longer needed, it can be returned to
			the storage pool.
		
			
	Operations on Breakpoints
		SetBreakpoint()
		ClearBreakpoint()
		EnableBreakpoint()
		DisableBreakpoint()
		FindBreakpoint()
		
COPYRIGHT

	Copyright 2000-2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#ifndef _BREAK_H_
#define _BREAK_H_

/****************************************************************************/
/*
 *	Includes
 */
/****************************************************************************/

#ifndef _NO_SYSTEM_INCLUDES
#include <stddef.h>
#endif
#include "dstypes.h"
#include "dserror.h"
#include "mutex_TRK.h"
#include "target.h"
#include "msgcmd.h"
#include "usr_put_config.h"

/*
** By default, L1 and L2 (bare-board support) does not support breakpoints, and
** L3 (RTOS support) does.  Any target can override this by defining
** TRK_SUPPORT_BREAKPOINTS in target.h.
*/
#ifndef TRK_SUPPORT_BREAKPOINTS
	#if DS_PROTOCOL < DS_PROTOCOL_RTOS
		#define TRK_SUPPORT_BREAKPOINTS		(0)
	#else
		#define TRK_SUPPORT_BREAKPOINTS		(1)
	#endif
#endif

/*
** Set this to 1 to get debug messages showing the get/release calls
** for the message breakpoints.  This helps to track down leaks.
*/
#define DEBUG_BREAK_USAGE		(0)
#if DEBUG_BREAK_USAGE && (defined( DEBUGIO_SERIAL ) || defined( DEBUGIO_RAM ) || defined( DEBUGIO_CUSTOM ))
#define GetFreeBreakpoint( id, b )		( __puts( "Calling GetFreeBreakpoint from " ), __puts( __FILE__ ), __putchar( ':' ), __puthex8( __LINE__ ), __puts( "\r\n" ), _GetFreeBreakpoint( id, b ) )
	#define ReleaseBreakpoint( id )			( __puts( "Calling ReleaseBreakpoint from " ), __puts( __FILE__ ),  __putchar( ':' ), __puthex8( __LINE__ ), __puts( "\r\n" ), _ReleaseBreakpoint( id ) )
#else /* #if DEBUG_BREAK_USAGE */
	#define GetFreeBreakpoint( id, b )		_GetFreeBreakpoint( id, b )
	#define ReleaseBreakpoint( id )			_ReleaseBreakpoint( id )
#endif /* #if DEBUG_BREAK_USAGE */

/****************************************************************************/
/*
 *	Errors
 */
/****************************************************************************/
enum {								/* DSErroror; see dserror.h */
	kBreakpointsFull = 0x0b00,		/* no more available bp resources */
	kBreakpointNotFound,			/* could not find specified breakpoint */
	kBreakpointConflict				/* attempt to set bp which conflicts w/existing bp */
};

/****************************************************************************/
/*
 *	Types
 */
/****************************************************************************/

/*
 *	BreakpointId
 */
typedef int BreakpointId;
enum {
	kInvalidBreakpointId = -1
};

/*
** The following three constants go to-gether.
**
** The initial breakpoint capacity is "kInitialBreakpointCount."
** When breakpoints are full, the limit is increased  
** by an incremental amount "kBreakpointCountIncrement",
** till the value  reaches system limit "kBreakpointCountMax"
** So the user is requested to modify only "kBreakpointCountMax"
** henceforth.
**
*/
#define kInitialBreakpointCount (100) 

/*
** Amount by which "kBreakpointCount" is incremented when breakpoints
** are full
**
*/
#define kBreakpointCountIncrement (100) 

/* *** USER MODIFIABLE PARAMETER ***
** Maximum number of breakpoints that can be set in the system.  
** This can be modified if the limit exceeds this constant.
**
*/
#define kBreakpointCountMax (4000) 

/*
** Breakpoint conditions type.  At present they fit into a byte.
** We use the breakpoint condition flags defined in msgcmd.h.
*/
typedef ui8		BreakpointCond;

/*
 ** struct Breakpoint.  Note that the BreakpointRestore type is 
 ** target-specific & so must be defined somewhere in target.h
 **
 ** Note: The use of DS_MSG_BREAK_THREAD_ID_ALL in fThreadId and false
 ** in fThreadLocalImp should not be confused.  The fThreadId field 
 ** represents the intent of the debugger, while the fThreadLocalImp
 ** field represents the capability of the low-level implementation.
 */
typedef struct Breakpoint {
	DSMutex			fMutex;					/* synchronization */
	bool			fInUse;					/* true if breakpoint is not free */
	BreakpointId	fNextId;				/* ID of next breakpoint in list, if in a list */
	BreakpointId	fPrevId;				/* ID of previous breakpoint in list, if in a list */
	bool			fEnabled;				/* true if breakpoint is not disabled */
	unsigned int	fStart;					/* start address of breakpoint */
	unsigned int	fLength;				/* length of breakpoint memory range */
	unsigned int	fCount;					/* only stop after triggered this many times (currently unused) */
	bool			fHardware;				/* true if breakpoint uses specific hardware support */
	ui32			fProcessId;				/* ID of associated process */
	ui32			fThreadId;				/* ID of associated thread */
	bool			fThreadLocalImp;		/* true if the implementation is local to the thread */
	BreakpointCond	fConditions;			/* breakpoint trigger conditions (r,w,m,x) */
	BreakpointId	fSharedId;				/* ID of shared breakpoint */
	ui8				fTargetOptions;			/* target-specific options, not used by portable code */
	BreakpointRestore fRestore;				/* contains info needed to clear bp once set */
	int				fReleaseDelay;			/* delay release for how long? */
} Breakpoint;

/****************************************************************************/
/*
 *	Public Function Prototypes
 */
/****************************************************************************/
__extern_c
	/*
	 *	Breakpoint management
	 */
	bool GetBreakpointEnabled(Breakpoint *bp);
	DSError InitializeBreakpoint( BreakpointId i );
	DSError InitializeBreakpoints(void);
	DSError _GetFreeBreakpoint(BreakpointId* resultBreakpointId,
							Breakpoint** resultBreakpoint);
	Breakpoint* GetBreakpoint(BreakpointId breakpointId);
	void _ReleaseBreakpoint(BreakpointId breakpointId);
	
	/*
	 * operations on Breakpoints
	 */
	
	DSError CheckBreakpointExistence(unsigned int start, unsigned int len,
						bool hardware, BreakpointCond conditions,
						unsigned count, ui8 targetOptions,
						ui32 processId, ui32 threadId );
	DSError SetBreakpoint( Breakpoint *b, unsigned int start, unsigned int len,
						bool hardware, BreakpointCond conditions,
						unsigned count, ui8 targetOptions,
						ui32 processId, ui32 threadId );
	DSError ClearBreakpoint( Breakpoint *b );
	DSError EnableBreakpoint( Breakpoint *b );
	DSError DisableBreakpoint( Breakpoint *b );
	DSError FindExecBreakpoint( BreakpointId* resultBreakpointId,
								Breakpoint** resultBreakpoint,
								unsigned int start, unsigned int len,
								bool requireEnabled, bool threadSpecific,
								ui32 processId, ui32 threadId );
	int ReleaseThreadBreakpoints( ui32 processId, ui32 threadId,
									bool clear, bool weak );
	void ReleaseAllBreakpoints( bool clear, bool weak );
	DSError HideBreakpoints( void* data,
							unsigned int start,
							unsigned int len,
							ui32 processId,
							ui32 threadId );
	void UpdateSharedBreakpoints( BreakpointId releasedId );
	
	DSError DefaultTargetSetSoftwareBreakpoint( Breakpoint *b,
											InstructionType instr );
	DSError DefaultTargetClearSoftwareBreakpoint( Breakpoint *b );
	DSError DefaultTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  );
	DSError AddSubBreakpoint( BreakpointId mainBreakpointId, ui32 threadId,
							BreakpointId* subBreakpointId,
							Breakpoint** subBreakpoint );
	DSError ClearInheritedBreakpoints( ui32 pProcessId, ui32 pThreadId, 
							ui32 cProcessId, ui32 cThreadId );
	DSError ReleaseDelayedBreakpoints( ui32 ProcessId, ui32 ThreadId );
	DSError ReInitializeBreakpoints(BreakpointId* resultBreakpointId,
								Breakpoint** resultBreakpoint);
	void FreeBreakpoint(void);
__end_extern_c

#endif /* _BREAK_H_ */
