/******************************************************************************************/
/*
	
DESCRIPTION

	Default implementation to set/retrieve ethernet configuration.
	
	If TRK_CONFIG_NVRAM_ADDR is defined in target.h, then the
	ethernet configuration will be stored in non-volatile RAM at
	that address.
	
	Otherwise it will be stored in standard RAM and the configuration
	will be restored to default values each time the target system is
	powered on or reset.  The following example illustrates how to
	define the default values in target.h:
		#define TRK_IP_ADDR				0x0a5603d0	// "10.86.3.208"
		#define TRK_GATEWAY				0x00000000	// "0.0.0.0"
		#define TRK_SUBNET_MASK			0xffffff00	// "255.255.255.0"
		#define TRK_DNS_SERVER			0x00000000	// "0.0.0.0"
		#define TRK_PORT_NUM			6868
	
	The functions TrkGetEthConfig() and TrkSetEthConfig() can be used
	to get or set the configuration.
	
COPYRIGHT

	Copyright 2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"		/* must be included first */

#include "target.h"
#include "ethconfig.h"
#include "nubassrt.h"
#include "mem_TRK.h"

/****************************************************************************/
/* Location of global ethernet configuration.
/****************************************************************************/
#ifdef TRK_CONFIG_NVRAM_ADDR

	/*
	** If an NVRAM address is defined, refer directly to that.
	*/
	#define gTrkEthConfigPtr		((TrkEthConfig*)(TRK_CONFIG_NVRAM_ADDR))

#else /* #ifdef TRK_CONFIG_NVRAM_ADDR */

	static TrkEthConfig gTrkEthConfig =
	{
		0,	// reserved for options
		TRK_IP_ADDR,
		TRK_PORT_NUM,
		TRK_GATEWAY,
		TRK_SUBNET_MASK
	};
	
	#define gTrkEthConfigPtr	(&gTrkEthConfig)
#endif /* #ifdef TRK_CONFIG_NVRAM_ADDR */



/****************************************************************************/
/*
** Function Name:		TrkGetEthConfig
**
** Description:			Returns the current ethernet configuration.
**
** Input:
**		config:			Ptr to storage for the resulting config.
**
** Output:
**		return:			error code (kNoError if succeeded)
**		config:			If successful, configuration will be stored at *config.
**
*/
/****************************************************************************/
DSError TrkGetEthConfig( TrkEthConfig *config )
{
	ASSERT( config );
	TRK_memcpy( config, gTrkEthConfigPtr, sizeof( TrkEthConfig ) );
	return( kNoError );
}

/****************************************************************************/
/*
** Function Name:		TrkSetEthConfig
**
** Description:			Sets the current ethernet configuration.
**
** Input:
**		config:			Ptr to the new configuration values.
**
** Output:
**		return:			error code (kNoError if succeeded)
**
*/
/****************************************************************************/
DSError TrkSetEthConfig( const TrkEthConfig *config )
{
	ASSERT( config );
	TRK_memcpy( gTrkEthConfigPtr, config, sizeof( TrkEthConfig ) );
	return( kNoError );
}

