/******************************************************************************************/
/*
	
DESCRIPTION

	This file is to be used on the host side to extract needed data
	from the debug server's reply buffer. Pass in the starting
	address of data buffer (the data only, no framing byte).
	
COPYRIGHT

	Copyright 1999-2004, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "msgxtrct.h"
#include <string.h>

#ifdef USE_ALLOC_TRACKER
#include "AllocTracker.h"
#endif

extern bool	HostLittleEndian(void);
bool	HostLittleEndian(void)
{
	unsigned long test = 0x11223344;
	unsigned char* testPtr = (unsigned char *)&test;
	if (*testPtr == 0x11)
		return false;	// Host is big endian 
	else
		return true;	// Host is little endian
}


//----------------------------------------------------------------------
// Function name	: DSGetReplyMemData32xN
// Description	    : Grabs the Big Endian data from the replyBuffer and
//                    places it into destAddr buffer in host endian-ness
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of 32 bit integers to grab/swap
//----------------------------------------------------------------------
void	DSGetReplyMemData32xN_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	unsigned char *replyMemDataPtr = (unsigned char *)replyBuffer;
	replyMemDataPtr += 4+snOffset; // Mem Data starts at 4 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyMemDataPtr, count*sizeof(ui32));

	// Swap only if host is not big endian since the CodeWarriorTRK returns
	// all values in big endian
	if (HostLittleEndian())
		Swap4xN((unsigned long *)destAddr, count);
}

//----------------------------------------------------------------------
// Function name	: DSGetReplyRegData32xN
// Description	    : Grabs the Big Endian data from the replyBuffer and
//                    places it into destAddr buffer in host endian-ness
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of 32 bit integers to grab/swap
//----------------------------------------------------------------------
void	DSGetReplyRegData32xN_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	unsigned char *replyRegDataPtr = (unsigned char *)replyBuffer;
	replyRegDataPtr += 2+snOffset; // Reg Data starts at 2 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyRegDataPtr, count*sizeof(ui32));

	// Swap only if host is not big endian since the CodeWarriorTRK returns
	// all values in big endian
	if (HostLittleEndian())
		Swap4xN((unsigned long *)destAddr, count);
}

//----------------------------------------------------------------------
// Function name	: DSGetReplyRegData64xN
// Description	    : Grabs the Big Endian data from the replyBuffer and
//                    places it into destAddr buffer in host endian-ness
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of 32 bit integers to grab/swap
//----------------------------------------------------------------------
void	DSGetReplyRegData64xN_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	unsigned char *replyRegDataPtr = (unsigned char *)replyBuffer;
	replyRegDataPtr += 2+snOffset; // Reg Data starts at 2 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyRegDataPtr, count*8);

	// Swap only if host is not big endian since the CodeWarriorTRK returns
	// all values in big endian
	if (HostLittleEndian())
		Swap8xN((unsigned char *)destAddr, count);
}

//----------------------------------------------------------------------
// Function name	: DSGetReplyRegData128xN
// Description	    : Grabs the Big Endian data from the replyBuffer and
//                    places it into destAddr buffer in host endian-ness
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of 128 bit integers to grab/swap
//----------------------------------------------------------------------
void	DSGetReplyRegData128xN_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	unsigned char *replyRegDataPtr = (unsigned char *)replyBuffer;
	replyRegDataPtr += 2+snOffset; // Reg Data starts at 2 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyRegDataPtr, count*16);

	// Swap only if host is not big endian since the CodeWarriorTRK returns
	// all values in big endian
	if (HostLittleEndian())
		Swap16xN((unsigned char *)destAddr, count);
}


//----------------------------------------------------------------------
// Function name	: DSGetReplyMemData
// Description	    : Grabs the data from the replyBuffer and
//                    places it into destAddr buffer with no swapping
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of bytes to copy
//----------------------------------------------------------------------
void	DSGetReplyMemData_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	ui8* replyRegDataPtr = (ui8*)replyBuffer;
	replyRegDataPtr += 4+snOffset; // Reg Data starts at 2 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyRegDataPtr, count);

}


//----------------------------------------------------------------------
// Function name	: DSGetReplyFileData
// Description	    : Grabs the data from the replyBuffer and
//                    places it into destAddr buffer with no swapping
// Return type		: none
// Argument         : void *destAddr: where to put data
// Argument         : void *replyBuffer: the buffer received from CWDS
// Argument         : ui16 count: number of bytes to copy
//----------------------------------------------------------------------
void	DSGetReplyFileData_SN(void *destAddr, void *replyBuffer, ui16 count, ui8 snOffset)
{
	unsigned char *replyRegDataPtr = (unsigned char *)replyBuffer;
	replyRegDataPtr += 5+snOffset; // file Data starts at 7 bytes into reply buffer

	memcpy((unsigned char *)destAddr, replyRegDataPtr, count);

}

//----------------------------------------------------------------------
// Function name	: TRKSupports
// Description	    : Takes a pointer to a support mask array(acquired by sending a
//					  kDSSupportMask message) and a TRK command ID and returns whether the
//					  bit for that command is set in the support mask.
// Return type		: bool
// Argument         : DSSupportMask *supportMask: pointer to support mask array
// Argument         : MessageCommandID cmdId: 
//----------------------------------------------------------------------
bool TRKSupports(DSSupportMask *supportMask, MessageCommandID cmdId)
{
	bool supports = false;
	int index = cmdId / 8;
	if (index >= 0 && index < sizeof(DSSupportMask))
	{
		int shiftIndex = cmdId - (8 * index);
		ui8 shiftMask = 0x01;
		shiftMask = shiftMask << shiftIndex;
		ui8 keyByte = (*supportMask)[index];
		if (keyByte & shiftMask)
			supports = true;
	}
	return supports;
}

