/******************************************************************************************/
/*

DESCRIPTION

	C MSL interface to the support functions provided by CodeWarriorTRK.
	This interface is designed to be built as a separate library
	that can be linked with a target program.  The target program
	can make calls into this interface to obtain support services
	from CodeWarriorTRK.  The interface is designed to be used directly
	by MSL, without requiring any #includes from the CodeWarriorTRK
	sources.
	
	NOTE: CodeWarriorTRK must be running on the target system in order for
	these routines to function.
	
	Public functions:
	
		__read_console() - Read a specified number of characters
			from the console into a buffer.
			
		__write_console() - Read a specified number of characters 
			from a buffer to the console.
			
		__close_console() - Close the console.
			
		__read_file() - Read a specified number of characters
			from a file into a buffer.
			
		__write_file() - Read a specified number of characters 
			from a buffer to a file.
			
		__close_file() - Close the specified file.

		
		Function parameters:
		
			handle (__file_handle): Specifies the console or
				disk file on which to operate.  Disk file
				handles are assigned dynamically.  Console
				file handles are:
				
					stdin:	0
					stdout: 1
					stderr:	2
					
			buffer (unsigned char *): For writes, points to
				the data which is to be written.  For reads,
				points to storage for the data which will
				be read.
				
			count (size_t *): The desired number of characters
				to read/write.  The value after calling 
				a particular function indicates the actual
				number of characters read/written.
				
			idle_proc (__idle_proc): A function to be called
				while the current operation is idle.  Not used
				in the current implementation.
				
		Return value:  All of the public functions here return
			a value of type 'enum __io_results' as defined
			in stdio.h in the CodeWarrior MSL.

COPYRIGHT

	Copyright 1999, 2001-2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include <console_io.h>
#include <file_io.h>

#define __NEED_BOOL__		/* causes bool to be defined by dstypes.h */

#include "dstypes.h"
#include "targsupp.h"
#include "msgcmd.h"

static int __access_file( __file_handle handle, unsigned char * buffer, 
						  size_t * count, __ref_con ref_con, bool read );


static int __read_file( __file_handle handle, unsigned char * buffer, 
				 size_t * count, __ref_con ref_con );

static int __write_file( __file_handle handle, unsigned char * buffer, 
				  size_t * count, __ref_con ref_con);

static int __close_file( __file_handle handle );

/****************************************************************************/
/*
 *	__read_console
 *
 *	Read the specified amount of data from the console file specified by
 *	handle into the given buffer.
 */
/****************************************************************************/
int __read_console( __file_handle handle, unsigned char * buffer, 
					size_t * count, __ref_con ref_con)
{
	return( __read_file( 0, buffer, count, ref_con ) );
}


/****************************************************************************/
/*
 *	__write_console
 *
 *	Write the specified amount of data to the console file specified by
 *	handle from the given buffer.
 */
/****************************************************************************/
int __write_console( __file_handle handle, unsigned char * buffer, 
					 size_t * count, __ref_con ref_con )
{
	return( __write_file( 1, buffer, count, ref_con ) );
}

/****************************************************************************/
/*
 *	__close_console
 *
 *	Close the console file specified by handle from the given buffer.
 */
/****************************************************************************/
int __close_console( __file_handle handle )
{
	return( __close_file( handle ) );
}

/****************************************************************************/
/*
 *	__read_file
 *
 *	Read the specified amount of data from the file specified by
 *	handle into the given buffer.
 */
/****************************************************************************/
int __read_file( __file_handle handle, unsigned char * buffer, 
				 size_t * count, __ref_con ref_con )
{
	return( __access_file( handle, buffer, count, ref_con, true ) );
}


/****************************************************************************/
/*
 *	__write_file
 *
 *	Write the specified amount of data to the file specified by
 *	handle from the given buffer.
 */
/****************************************************************************/
int __write_file( __file_handle handle, unsigned char * buffer, 
				  size_t * count, __ref_con ref_con)
{
	return( __access_file( handle, buffer, count, ref_con, false ) );
}

/****************************************************************************/
/*
 *	__close_file
 *
 *	Close the file specified by handle from the given buffer.
 */
/****************************************************************************/
int __close_file( __file_handle handle )
{
	#pragma unused( handle )

	return( __no_io_error );
}

/****************************************************************************/
/*
 *	__access_file
 *
 *	Read/Write the specified amount of data from/to the file specified by
 *	handle to/from the given buffer.
 */
/****************************************************************************/
int __access_file( __file_handle handle, unsigned char * buffer, 
				   size_t * count, __ref_con ref_con, bool read )
{
	#pragma unused( ref_con )

	ui32 local_count;
	ui8 result;

	/*
	** Convert MSL types (loosely specified sizes and forms) into
	** types with well-defined sizes and forms.
	*/
	
	local_count = (ui32)*count;
	
	result = TRKAccessFile( (ui8)(read ? kDSReadFile : kDSWriteFile),
							(ui32)handle, &local_count,
							(ui8 *)buffer );
	
	*count = (size_t)local_count;
	
	/*
	** Convert the result into the MSL equivalent.
	*/
				   
	switch (result)
	{
		case kDSIONoError:
			return( __no_io_error );
		case kDSIOEOF:
			return( __io_EOF );
		case kDSIOError:
		default:
			return( __io_error );
	}
}


