/******************************************************************************************/
/*

DESCRIPTION

	TRK rfc1662 message framing
	
COPYRIGHT

	Copyright 2002, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "nubassrt.h"

#include "mainloop.h"
#include "nubinit.h"
#include "nubevent.h"
#include "msgbuf.h"
#include "msg.h"
#include "dispatch.h"
#include "rfc1662.h"
#include "msghndlr.h"
#include "mem_TRK.h"

/******************************************************************************/
#include "serframe.h"


#include "transport.h"
#include "transport_tbl.h"

/******************************************************************************/

/*
** DEBUG_FRAME_STATE causes the frame state-machine to put out debug
** information.  It can not be used when the debug information goes 
** out to the serial port, as it will interfere with the serial input.
*/

#define DEBUG_FRAME_STATE	(0)

/*
** DEBUG_PACKET causes the entire contents of each received packet
** to be written out as debug information.  It is OK to use this
** even when debug info goes out via the serial port.
*/

#define DEBUG_PACKET_IN		(0)

/*
** DEBUG_PACKET_OUT causes the entire contents of each sent packet
** to be written out as debug information.  It is OK to use this
** even when debug info goes out via the serial port.
*/

#define DEBUG_PACKET_OUT	(0)

/******************************************************************************/
/*
 *	Framing Types
 */
/******************************************************************************/

typedef enum ReceiverState {
	kWaitFlag,
	kFoundFlag,
	kInFrame,
	kInFrameEscaped,
	kFrameOverflow
} ReceiverState;

typedef struct FramingState {
	MessageBufferId	fBufferID;
	MessageBuffer*	fBuffer;

	ReceiverState	fReceiveState;
	FCSType			fFCS;
} FramingState;

#ifndef IS_SMP
static FramingState gFramingState;
#else
#define gFramingState (*perCpuFramingState)
#endif

/******************************************************************************/
/*
	static function prototypes
*/
/******************************************************************************/

#if __MWERKS__
#pragma mark Static Function Prototypes
#endif /* __MWERKS__ */

static DSError InitializeRfc1662( LayerParams *p );
static DSError TerminateRfc1662( LayerParams *p );
static DSError ReadRfc1662( LayerParams *params, MessageBufferId *bufferId );
static DSError WriteRfc1662( LayerParams *p, MessageBuffer *b );
static DSError IrqRfc1662( LayerParams *p );

static void DiscardFrame(FramingState* frameState);
static void RejectFrame(FramingState* framingState, DSReplyError replyError);
static bool ProcessFrame(FramingState* frameState);



/******************************************************************************/
/*
	global variables
*/
/******************************************************************************/

#if __MWERKS__
#pragma mark -
#pragma mark Global Variables
#endif /* __MWERKS__ */

MsgLayerApi gRfc1662Api =
{
	InitializeRfc1662,				/* init */
	TerminateRfc1662,				/* terminate */
	ReadRfc1662,					/* read */
	WriteRfc1662,					/* write */
	IrqRfc1662						/* irq */
};

#if __MWERKS__
#pragma mark -
#endif /* __MWERKS__ */

/******************************************************************************/
/*
 *	InitializeRfc1662
 */
/******************************************************************************/

DSError InitializeRfc1662( LayerParams *p )
{
#ifdef IS_SMP
	FramingState* perCpuFramingState = TRK_malloc(sizeof(FramingState));
	PRIVATE_DATA(COM_FRAME_STATE) = perCpuFramingState;
#endif

	gFramingState.fBufferID = kInvalidMessageBufferId;
	gFramingState.fReceiveState = kWaitFlag;

	/*
	** Initialize the sub-layer.
	*/
	
	return( LAYER_INIT( ((Rfc1662Params *)p)->nextLayer ) );
}


/******************************************************************************/
/*
 *	TerminateRfc1662
 */
/******************************************************************************/

DSError TerminateRfc1662( LayerParams *p )
{
	/*
	** Terminate the sub-layer.
	*/
#ifdef IS_SMP
	FramingState* perCpuFramingState = PRIVATE_DATA(COM_FRAME_STATE);
	TRK_free(perCpuFramingState);
	PRIVATE_DATA(COM_FRAME_STATE) = NULL;
#endif
	return( LAYER_TERMINATE( ((Rfc1662Params *)p)->nextLayer ) );
}

/******************************************************************************/
/*
	ReadRfc1662
	
	Poll the input stream for a completed message.
	
	The following is true when this routine exits:
	
		- All of the input was read.
			--OR--
		- A complete message was read.
	
*/
/******************************************************************************/
DSError ReadRfc1662( LayerParams *params, MessageBufferId *bufferId )
{
	DSError				err;
	char				c;
	Rfc1662Params		*lparams = (Rfc1662Params *)params;
	CharLayer			*nextLayer = lparams->nextLayer;
#ifdef IS_SMP
	FramingState* perCpuFramingState = PRIVATE_DATA(COM_FRAME_STATE);
#endif

	/* get input from the next layer */
	err = CHAR_LAYER_READ( nextLayer, &c );
	
	/*
	 *	read characters until an error occurs or until no characters
	 *	are available
	 */
	
	while (err == kNoError) {
	
#if DEBUG_FRAME_STATE
		__puts("char = [");
		__puthex2(c);
		__puts("] ");
#endif	
		/*
		 *	handle state machine
		 */
		switch (gFramingState.fReceiveState) {
		
			case kWaitFlag:
#if DEBUG_FRAME_STATE
				__puts("case kWaitFlag\r\n");
#endif
			
				/*
				 *	Start of a new packet
				 */
				 
				if (c == PPP_FLAG) {
					ASSERT(gFramingState.fBufferID == kInvalidMessageBufferId);

					/*
					 *	allocate a buffer and reset frame state
					 */
					err = GetFreeBuffer(&gFramingState.fBufferID, &gFramingState.fBuffer);
					ASSERT(err == kNoError);
					if (err == kNoError)
					{
					gFramingState.fFCS = PPPINITFCS;
					gFramingState.fReceiveState = kFoundFlag;	/* next state */
					}

				} else {
					/*
					 *	discard character
					 */
				}
				break;
				
				
			case kFoundFlag:
#if DEBUG_FRAME_STATE
				__puts("case kFoundFlag\r\n");
#endif

				if (c == PPP_FLAG)		/* discard multiple flags */
					break;
				
				
				gFramingState.fReceiveState = kInFrame;
				/* break; */
				/* FALL THROUGH TO kInFrame, receive first char */
				
			case kInFrame:
#if DEBUG_FRAME_STATE
				__puts("case kInFrame\r\n");
#endif

				if (c == PPP_FLAG) {
					
						/*
						 *	normal case, good packet
						 */
						if (ProcessFrame(&gFramingState)) {
							*bufferId = gFramingState.fBufferID;				
							gFramingState.fBufferID = kInvalidMessageBufferId;
							gFramingState.fBuffer = NULL;
							gFramingState.fReceiveState = kWaitFlag;
							return kNoError;
						}
						
						/* now start all over again */
						gFramingState.fReceiveState = kWaitFlag;
					break;
					
				} else if (c == PPP_ESCAPE) {
					
					gFramingState.fReceiveState = kInFrameEscaped; /* next character is escaped */
					break;
				}

kInFrame_Unescaped:
				/*
				 *	Append bytes to buffer and compute running FCS (checksum)
				 */
				err = MessageAdd_ui8(gFramingState.fBuffer, c);

				if (err == kMessageBufferOverflow)
				{
					
					/*
					** Overflow.  Send out a NAK immediately.  Go into an overflow state
					** which will wait for the end of the packet.
					*/
					
					StandardACK( gFramingState.fBuffer, kDSReplyNAK, kDSReplyOverflow );
					gFramingState.fReceiveState = kFrameOverflow;
					err = kNoError;
					break;
				}
				
				ASSERT(err == kNoError);
				
				gFramingState.fFCS = PPPFCS(gFramingState.fFCS, c);
				break;

			case kInFrameEscaped:
#if DEBUG_FRAME_STATE
				__puts("case kInFrameEscaped\r\n");
#endif
				if (c == PPP_FLAG)
				{
					/*
					 *	[..., escape, flag] is bad frame; drop it
					 */
					RejectFrame( &gFramingState, kDSReplyEscapeError );
				}
				else
				{
					c ^= PPP_TRANS;					/* change character to new code */
					gFramingState.fReceiveState = kInFrame; /* no longer escaped character */
					goto kInFrame_Unescaped;		/* process un-escaped character */
				}
				
				break;
				
			case kFrameOverflow:
			
				/*
				** Overflow occurred.  Throw away all characters and wait for a flag
				** to end the packet.
				*/
				
				if (c == PPP_FLAG)
					DiscardFrame( &gFramingState );
				break;
				
				
			default:
				ASSERT(!"Invalid State");
				break;
		};
		err = CHAR_LAYER_READ( nextLayer, &c );
	} /* while (err == kNoError) */

	/*
	** If we got here, then a complete message is not yet available.
	*/
	
	*bufferId = kInvalidMessageBufferId;
    return( err );
}

/******************************************************************************/
/*
 *	WriteRfc1662
 */
/******************************************************************************/

DSError WriteRfc1662( LayerParams *p, MessageBuffer *b )
{
	Rfc1662Params *lp = (Rfc1662Params *)p;
	CharLayer *clayer = lp->nextLayer;
	
	FCSType				fcs;
	FCSType				acc;
	int					i;
	DSError	err = kNoError;
		
	/*
	 *	Walk through buffer, computing frame check sequence
	 *	(checksum).
	 */
	fcs = PPPINITFCS;
	
	for (i = 0; i < b->fLength; i++) {
		fcs = PPPFCS(fcs, b->fData[i]);
#if DEBUG_PACKET_OUT		
		__puts("[");
		__puthex2(b->fData[i]);
		__puts("]");
	}
	__puts("\r\n");
#else
	}
#endif	

	fcs = fcs ^ PPPCOMPFCS;							/* finish by complementing */
	
#if DEBUG_PACKET_OUT		
	__puts("MessageSend. FCS = ");
	__puthex8(fcs);
	__puts("\r\n");
#endif
	

	/*
	 *	Send bytes out the UART:
	 *	[0x7E (frame flag)] [...data...] [FCS:8 or FCS:16 or FCS:32] [0x7E (frame flag)]
	 */
	
	/* FLAG */

	err = CHAR_LAYER_WRITE( clayer, PPP_FLAG );
					
	/* DATA */
	
	if (err == kNoError) {
		for (i = 0; i < b->fLength; i++) {
			unsigned char c = b->fData[i];
			
			if (c == PPP_FLAG || c == PPP_ESCAPE) {		/* escape if necessary */
				err = CHAR_LAYER_WRITE( clayer, PPP_ESCAPE );
				c ^= PPP_TRANS;
				if (err != kNoError)
					break;
			}
			err = CHAR_LAYER_WRITE( clayer, c );
			if (err != kNoError)
				break;
		} /* for */
	}
	
	
	/* FCS */
	/*
	 *	FCS always goes out in little endian (LSB, ... , MSB) order
	 */
	
	if (err == kNoError) {
		acc = fcs;									/* accumulator */
		for (i = 0; i < sizeof(FCSType); i++) {		/* 1, 2- or 4-bytes */
			unsigned char c = acc & 0xFF;
#if FCSBITSIZE != FCS8
			acc >>= 8;
#endif

			if (c == PPP_FLAG || c == PPP_ESCAPE) {		/* escape if necessary */
				err = CHAR_LAYER_WRITE( clayer, PPP_ESCAPE );
				c ^= PPP_TRANS;
				if (err != kNoError)
					break;
			}			
			err = CHAR_LAYER_WRITE( clayer, c );
			if (err != kNoError)
				break;
		} /* for */
	}
	
	/* FLAG */
	
	if (err == kNoError) {
		err = CHAR_LAYER_WRITE( clayer, PPP_FLAG );
		CHAR_LAYER_FLUSH( clayer );
	}
	
	return err;
}

/******************************************************************************/
/*
 *	IrqRfc1662
 */
/******************************************************************************/

int IrqRfc1662( LayerParams *p )
{
	Rfc1662Params *lp = (Rfc1662Params *)p;
	CharLayer *clayer = lp->nextLayer;
	
	/*
	** Defer to the sub-layer.
	*/
	
	return( LAYER_IRQ( clayer ) );
}

/******************************************************************************/
/*
	DiscardFrame
	
	Release resources used for the current frame.
*/
/******************************************************************************/

static void DiscardFrame(FramingState* state)
{
	if (state->fBufferID != kInvalidMessageBufferId) {
		ReleaseBuffer(state->fBufferID);
		state->fBufferID = kInvalidMessageBufferId;
	}
	state->fBuffer = NULL;
	state->fReceiveState = kWaitFlag;
}


/****************************************************************************/
/*
 *	RejectFrame
 *
 *	A message with initial frame byte and data arrived, but it is invalid.
 *
 */
/******************************************************************************/
static void RejectFrame(FramingState* framingState, DSReplyError replyError)
{
	MessageBuffer*	msgBuffer;
	
	__puts("RejectFrame\r\n");
		
	msgBuffer = framingState->fBuffer;
	ASSERT(msgBuffer);

	/*
	** Send a NAK with the given error.
	*/

	StandardACK( msgBuffer, kDSReplyNAK, replyError );

	/*
	** Throw away the data, release the frame.
	*/
	
	DiscardFrame( framingState );
}


/****************************************************************************/
/*
 *	ProcessFrame
 *
 *	A message with initial frame byte and data arrived.
 */
/******************************************************************************/
static bool ProcessFrame(FramingState* framingState)
{
	MessageBuffer*	msgBuffer;
	
#if DEBUG_FRAME_STATE || DEBUG_PACKET_IN
	__puts("ProcessFrame\r\n");
#endif
		
	msgBuffer = framingState->fBuffer;
	ASSERT(msgBuffer);
	
#if DEBUG_PACKET_IN
	{
		int i;

		__puts( "Received packet: " );
		
		for (i = 0; i < msgBuffer->fLength; i++) {
			__putchar( '[' );
			__puthex2( msgBuffer->fData[i] );
			__putchar( ']' );
		}
		
		__puts( "\r\n" );
	}
#endif /* DEBUG_PACKET_IN */
	
	
	/*
	 *	The buffer contains data in the following form:
	 *		[data] [FCS8]
	 *	or:
	 *		[data] [FCS16]
	 *	or:
	 *		[data] [FCS32]
	 *	
	 *	where data is arbitrary length and FCSxx is 1, 2 or 4 bytes
	 */
	if (msgBuffer->fLength < (DS_MIN_MSG_LENGTH + sizeof(FCSType))) {
		/*
		 *	data received is too short
		 */
		__puts("frame is too short\r\n");
		__puthex8(msgBuffer->fLength);
		__puts("\r\n");
		
		RejectFrame( framingState, kDSReplyPacketSizeError );
		
		return false;
	}
	
	/*
	 *	check for FCS
	 */
	
	if (framingState->fFCS != PPPGOODFCS) {
		/*
		 *	back checksum (FCS)
		 */
		RejectFrame( framingState, kDSReplyBadFCS );
		
		__puts("Bad FCS = ");
		__puthex8(framingState->fFCS);
		__puts("\r\n");
		return false;
	}

	/*
	 *	trim back packet, eliminating FCS
	 */
	msgBuffer->fPosition = 0;		/* rewind position to beginning*/
	msgBuffer->fLength -= sizeof(FCSType);
	
#if DEBUG_FRAME_STATE || DEBUG_PACKET_IN
	__puts("fLength = ");
	__puthex8(msgBuffer->fLength);
	__puts("\r\n");
#endif
	
	return true;
}



