/******************************************************************************************/
/*
	
DESCRIPTION

	Portable implementation of step logic.  Interfaces with target-specific
	routines defined in targimpl.h.
	
COPYRIGHT

	Copyright 2000-2003, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

/*
** TRK includes
*/

#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "targimpl.h"
#include "step.h"
#include "nubassrt.h"
#include "mainloop.h"

/* we cannot have stepping support through breaks without the target agent implementing breakpoint support */
#if TRK_SUPPORT_BREAKPOINTS

/*
** Set this to 1 to see detailed information about the invisible
** step ranges while stepping.
*/

#define DEBUG_INVISIBLE_STEP	(1)

/****************************************************************************/
/*
 *	Static Function Prototypes
 */
/****************************************************************************/
static DSError TargetDoStep( StepStatus *status, ui32 processId, ui32 threadId );
static bool TargetStepDone( StepStatus *status,
						InvisibleRangeInfo	*fInvisibleStepRangeList,
						bool *invisibleStop, ui32 processId, ui32 threadId );

/****************************************************************************/
/*
 *	Function Definitions
 */
/****************************************************************************/

/****************************************************************************/
/*
 *	TargetDoStep
 *
 *	General-purpose step handler.  Sets up the instruction trace and
 *  returns control to the target program.  Used to initiate the step
 *	and to continue the stepping as long as needed.
 */
/****************************************************************************/
static DSError TargetDoStep( StepStatus *status, ui32 processId, ui32 threadId )
{
	DSError err = kNoError;

	status->active = true;
	
	/*
	** Once the trace is enabled, just continue.  The target program
	** will break after each instruction.
	*/
	
	TargetSetupTrace( status, processId, threadId );
	
	status->traceStart = XTargetGetPC(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
								);


	/*
	** To continue, post a continue event.  The mainloop will do the
	** continue after it has handled any events that are already pending.
	*/

	PostContinueEvent(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
								);

	return err;
}

/****************************************************************************/
/*
 *	TargetSingleStep
 *
 *	Step over a specified number of instructions in the target
 *	thread.
 */
/****************************************************************************/
DSError TargetSingleStep( unsigned count, bool stepOver,
						ui32 processId,
						ui32 threadId )
{
	DSError err;
	StepStatii *statii;
	StepStatus *status;

	/*
	** A new step request from the host.  Cancel any previous in-progress
	** stepping for this thread.
	*/

	TargetCancelStep( processId, threadId );

	/*
	** Set up the step status data and call the general-purpose routine.
	*/
	
	statii = XTargetGetStepStatii(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
									
	ASSERT( statii != NULL );
	ASSERT( statii->current == 0 );
	status = &(statii->arr[ 0 ]);
	
	status->type = kDSStepIntoCount;
	status->count = count;
	status->stepOver = stepOver;
	
	err = TargetDoStep( status, processId, threadId );
	
	return( err );
}						


/****************************************************************************/
/*
 *	TargetStepOutOfRange
 *
 *	Step in the target thread until the PC is out of the
 *	specified address range.
 */
/****************************************************************************/
DSError TargetStepOutOfRange( ui32 start, ui32 end,
						bool stepOver,
						ui32 processId,
						ui32 threadId )
{
	DSError err;
	StepStatii *statii;
	StepStatus *status;

	/*
	** A new step request from the host.  Cancel any previous in-progress
	** stepping for this thread.
	*/

	TargetCancelStep( processId, threadId );

	/*
	** Set up the step status data and call the general-purpose routine.
	*/

	statii = XTargetGetStepStatii(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
									
	ASSERT( statii != NULL );
	ASSERT( statii->current == 0 );
	status = &(statii->arr[ 0 ]);

	status->type = kDSStepIntoRange;
	status->rangeStart = start;
	status->rangeEnd = end;
	status->stepOver = stepOver;
	
	err = TargetDoStep( status, processId, threadId );
	
	return( err );
}						


/****************************************************************************/
/*
 *	TargetCheckStep
 *
 *	General-purpose step handler - checks status and continues the
 *	instruction trace if the step is not complete.  Should be called
 *	whenever a breakpoint is hit.
 *
 *	Returns true if further stepping/execution activity remains,
 *	false otherwise.
 */
/****************************************************************************/
bool TargetCheckStep( InvisibleRangeInfo	*fInvisibleStepRangeList,
						bool *invisibleStop, ui32 processId, ui32 threadId )
{
	StepStatii *statii;
	StepStatus *status;
	bool tempDisabled;
	bool abort;
	bool release;
	bool result;

	result = false;
	release = true;
	
	*invisibleStop = false;

	statii = XTargetGetStepStatii(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
	ASSERT( statii != NULL );
	
	status = &(statii->arr[ statii->current ] );
	     

	/*
	** If there is a temporarily disabled breakpoint for this thread, then
	** we have just single-stepped over it.  Re-enable it.  This is a special,
	** internally generated step.
	*/

	if (status->tempDisableBreakpointId != kInvalidBreakpointId)
	{
		tempDisabled = true;
		EnableBreakpoint( GetBreakpoint( status->tempDisableBreakpointId ) );
		status->tempDisableBreakpointId = kInvalidBreakpointId;
	}
	else
	{
		tempDisabled = false;
	}

	/*
	** We hit a breakpoint or completed a single-instruction trace.  If a
	** step is in-process, continue it or stop it.
	*/

	if (status->active || tempDisabled)
	{
		abort = false;

		/*
		** Check to see if stopped at a trace breakpoint.  If
		** not, then abort the current step.
		*/
	
		if (!XAtTraceBreakPoint( status
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								, processId, threadId 
#endif
								))
		{
			abort = true;
			__puts( "Aborting step: not at trace breakpoint\r\n" );
		}
		
		if (status->active)
		{
			/*
			** If we're aborting, don't de-activate the step yet, but
			** indicate to the caller that we're done stepping for now.
			** If the host continues, the step will still be active.
			** If the host attempts a new step operation, it will
			** replace the current step operation.
			**
			** If we're finished, de-activate the step.  Otherwise,
			** continue with the next trace.
			*/
		
			if (abort)
			{
				release = false;
				result = false;
			}
			else 
			{
				/*
				** The trace must be cleaned up regardless of the error.
				*/
				
				TargetCleanupTrace( status, processId, threadId );

				if (TargetStepDone( status, fInvisibleStepRangeList,
						invisibleStop, processId, threadId ))
				{
					status->active = false;
					release = true;
					result = false;
					__puts( "Stepping complete\r\n" );
				}
				else
				{
					TargetDoStep( status, processId, threadId );
					release = false;
					result = true;
				}
			}
		}
		else if (tempDisabled)
		{
			/*
			** The trace must be cleaned up regardless of the error.
			*/
			
			TargetCleanupTrace( status, processId, threadId );

			/*
			** If no host-generated step command is active, but there was a 
			** temporarily disabled breakpoint, then we were attempting to
			** continue the thread when we single-stepped over that breakpoint.
			** To continue, post a continue event.  The mainloop will do the
			** continue after it has handled any events that are already pending.
			*/

			if (abort)
			{
				result = false;
				__puts( "Temp-disable step complete\r\n" );
			}
			else
			{
				PostContinueEvent(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
				result = true;
			}
			release = true;
		}

	}
	
		
	/*
	** The release flag will be true if the info stored at the current
	** step depth is no longer needed, so we can decrease the step depth.
	** However, if there is more than one level of step depth, the next
	** level up is not done.  If the result is false in this case, set
	** the result back to true.
	*/
	
	if (release && statii->current != 0)
	{
		statii->current--;
		result = true;
		
		__puts( "Decreasing step depth to " );
		__puthex2( statii->current );
		__puts( "\r\n" );
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	TargetStepDone
 *
 *	Determines whether step stop conditions are met.  If no trace is
 *	currently active, returns true.
 */
/****************************************************************************/
static bool TargetStepDone(StepStatus *status,
						InvisibleRangeInfo	*fInvisibleStepRangeList,
						bool *invisibleStop, ui32 processId, ui32 threadId )
{
	bool result = true;
	bool invisible = false;
	ui32 pc;
	/*Temp for list traversal*/
	InvisibleRangeInfo *tmp_RangeList;
	/*
	** If no step is active, then it is done by default.
	*/

	if (status->active)
	{

		pc = XTargetGetPC(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
								);
		
		/*
		** Invisible memory ranges are memory ranges that the user
		** doesn't want to know about, such as glue code that is
		** called when stepping into a shared library routine.
		** Don't check the step conditions or update the instruction
		** count while in one of these.
		*/
			
#if DEBUG_INVISIBLE_STEP
		__puts( "invisible step ranges:" );
#endif
	
	tmp_RangeList = fInvisibleStepRangeList;
	while (tmp_RangeList!= NULL)
	{
#if DEBUG_INVISIBLE_STEP
			__puts( " " );
			__puthex8(tmp_RangeList->InvisibleStepRanges.fStart );
			__puts( "-" );
			__puthex8(tmp_RangeList->InvisibleStepRanges.fEnd );
#endif
			if (pc >= tmp_RangeList->InvisibleStepRanges.fStart &&
				pc <= tmp_RangeList->InvisibleStepRanges.fEnd)
			{
				*invisibleStop = true;
				invisible = true;
				result = false;
				break;
			}
			tmp_RangeList = tmp_RangeList->next;
	}
#if DEBUG_INVISIBLE_STEP
		__puts( "\r\n" );
#endif

		if (!invisible)
		{
			switch (status->type)
			{

				/*
				** For single step, the trace is not done as long as the
				** count is larger than 0.
				*/

				case kDSStepIntoCount:
				case kDSStepOverCount:

					/*
					** Decrement the count each time.  Assume that this
					** function is called only once after each instruction
					** is executed.
					*/
					
					if (status->type == kDSStepIntoCount ||
						status->type == kDSStepOverCount)
						status->count--;

					if (status->count > 0)
						result = false;
					break;
					
				/*
				** For stepping out of a memory range, the trace is not
				** done as long as the PC is in the memory range, where
				** the range includes the start and end points.
				*/
				
				case kDSStepIntoRange:
				case kDSStepOverRange:
					if (pc >= status->rangeStart &&
						pc <= status->rangeEnd)
						result = false;
					break;
				
				default:
					break;
			}
		}
	}
	
	return( result );
}

/****************************************************************************/
/*
 *	TargetSetupTempDisableStep
 *
 *	Used to setup the step system in order to temporarily disable a
 *	breakpoint, step over it, and re-enable it.
 */
/****************************************************************************/
DSError TargetSetupTempDisableStep( BreakpointId tdbId,
									ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	Breakpoint *tdb;
	StepStatii *statii;
	StepStatus *status;
	ui32 pc;
	
	statii = XTargetGetStepStatii(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
	ASSERT( statii != NULL );
	ASSERT( statii->current == 0 );

	/*
	** First disable the breakpoint and store the ID.
	*/

	tdb = GetBreakpoint( tdbId );
	ASSERT( tdb != NULL );
	err = DisableBreakpoint( tdb );

	/*
	** Now determine what step depth we need.  In most cases only one
	** level is needed.  However, we need to add another level in the
	** case where we are stepping over a function call at the high level,
	** we hit an internally-generated breakpoint within that function, and
	** we want to auto-continue from that breakpoint.  We need to do a
	** temp-disable-step over that breakpoint without interfering with the
	** high level step operation.
	**
	** This happens, for example, when we have shared library load breakpoints
	** and we step over a function that loads a shared library.
	*/
	
	if (err == kNoError)
	{
		status = &(statii->arr[ statii->current ]);
		
		pc = XTargetGetPC(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
								);
		
		if (status->active && status->traceStart == pc)
		{
			/*
			** Stepping is already active and it was initiated at
			** this location, so an appropriate single-instruction
			** trace has already been setup.  Nothing special to do,
			** and the current step depth is fine.
			*/
		}
		else
		{
			/*
			** Need to set up a single-instruction trace.  If the 
			** current step depth is in use, go deeper.
			*/
			
			if (status->active ||
				status->tempDisableBreakpointId != kInvalidBreakpointId)
			{
				if (statii->current >= (kStepMaxDepth - 1))
				{
					err = kInternalError;
				}
				else
				{
					statii->current++;
					status = &(statii->arr[ statii->current ] );

					__puts( "Increasing step depth to " );
					__puthex2( statii->current );
					__puts( "\r\n" );
				}
			}
			
			if (err == kNoError)
			{
				err = TargetSetupTrace( status, processId, threadId );
				status->traceStart = pc;
			}
			
		}

		/*
		** Store the id of the disabled breakpoint at the appropriate
		** step depth.
		*/
			
		if (err == kNoError)
			status->tempDisableBreakpointId = tdbId;

		/*
		** If there was a problem, clean up.
		*/

		if (err != kNoError)
		{
			status->tempDisableBreakpointId = kInvalidBreakpointId;
			EnableBreakpoint( tdb );
		}
	}
	
	return( err );
}

/****************************************************************************/
/*
 *	TargetCancelStep
 *
 *	Cancels any step operations in progress.  Should be called any time
 *	the target thread stops for a reason that is definitely not part of
 *	an internal step operation.
 *
 *	Returns true if further stepping/execution activity remains,
 *	false otherwise.
 */
/****************************************************************************/
DSError TargetCancelStep( ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	StepStatii *statii;
	StepStatus *status;
	int i;
	bool tempDisabled;

	statii = XTargetGetStepStatii(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									processId, threadId
#endif
									);
	ASSERT( statii != NULL );
	
	for (i = statii->current; i >= 0; i--) 
	{
		status = &(statii->arr[ i ] );

		/*
		** If there is a temporarily disabled breakpoint for this thread, then
		** we have just single-stepped over it.  Re-enable it.  This is a special,
		** internally generated step.
		*/

		if (status->tempDisableBreakpointId != kInvalidBreakpointId)
		{
			tempDisabled = true;
			EnableBreakpoint( GetBreakpoint( status->tempDisableBreakpointId ) );
			status->tempDisableBreakpointId = kInvalidBreakpointId;
		}
		else
		{
			tempDisabled = false;
		}

		/*
		** We hit a breakpoint or completed a single-instruction trace.  Cleanup
		** any trace that was set up.
		*/

		if (status->active || tempDisabled)
		{
			TargetCleanupTrace( status, processId, threadId );
		}
		
		status->active = false;
	}
	
	statii->current = 0;
	
	return( err );
}


/****************************************************************************/
/*
 *	InitStepStatii
 *
 *	Initialize a new StepStatii object.
 */
/****************************************************************************/
DSError InitStepStatii( StepStatii *statii )
{
	int i;
	
	for (i = 0; i < kStepMaxDepth; i++)
		InitStepStatus( &(statii->arr[ i ]) );
		
	return( kNoError );
}

/****************************************************************************/
/*
 *	InitStepStatus
 *
 *	Initialize a new StepStatus object.
 */
/****************************************************************************/
DSError InitStepStatus( StepStatus *status )
{
	int i;
	
	status->active = false;
	status->stepOver = false;
	status->type = kDSStepIntoCount;
	status->count = 0;
	status->rangeStart = 0;
	status->rangeEnd = 0;
	status->tempDisableBreakpointId = kInvalidBreakpointId;
	
	for (i = 0; i < kTraceBreakpointCount; i++)
		status->traceBreakpoints[ i ] = kInvalidBreakpointId;
		
	return( kNoError );
}

#endif /* TRK_SUPPORT_BREAKPOINTS */
