/******************************************************************************************/
/*
	
DESCRIPTION

	Endian byte-swapping routines
	
COPYRIGHT

	Copyright 1999, 2002, 2003, 2006, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "SwapUtils.h"

//----------------------------------------------------------------------------
//	Swap2()
//
//	Byte-swaps a 2-byte value.  Used to convert to/from little/big endian.
//
//	Accepts:	unsigned short	source	- value to byte swap
//	Returns:	unsigned short			- swapped value
//----------------------------------------------------------------------------

unsigned short Swap2(unsigned short source)
{
	unsigned short dest;

	dest = source & 0x00FF;
	dest <<= 8;
	dest |= (source >> 8) & 0x00FF;
	
	return dest;
} 


//----------------------------------------------------------------------------
//	Swap2xN()
//
//	Byte-swaps N 2-byte values *in place*.  Used to convert to/from little/big
//	endian.
//
//	Accepts:	unsigned short*	source	- array of 16-bit values to swap
//				long			count	- # of 16-bit values to swap
//	Returns:	nothing
//----------------------------------------------------------------------------

void Swap2xN(unsigned short *source, long count)
{
	unsigned short temp;

	while (count--)
	{
		temp = Swap2(*source);
		*source++ = temp;
	}
} 


//----------------------------------------------------------------------------
//	Swap4()
//
//	Byte-swaps a 4-byte value.  Used to convert to/from little/big endian.
//
//	Accepts:	unsigned long	source	- value to byte swap
//	Returns:	unsigned long			- swapped value
//----------------------------------------------------------------------------

unsigned long Swap4(unsigned long source)
{
	unsigned long dest = 0;
	int i;
	
	for (i=0; i< 4; i++)
	{
		dest <<= 8;
		dest |= source & 0xFF;
		source >>= 8;
	}
	
	return dest;
} 


//----------------------------------------------------------------------------
//	Swap4xN()
//
//	Byte-swaps N 4-byte values *in place*.  Used to convert to/from little/big
//	endian.
//
//	Accepts:	unsigned long*	source	- array of 32-bit values to swap
//				long			count	- # of 32-bit values to swap
//	Returns:	nothing
//----------------------------------------------------------------------------

void Swap4xN(unsigned long *source, long count)
{
	unsigned long temp;

	while (count--)
	{
		temp = Swap4(*source);
		*source++ = temp;
	}
} 


//----------------------------------------------------------------------------
//	Swap8xN()
//
//	Byte-swaps N 8-byte values *in place*.  Used to convert to/from little/big
//	endian.
//
//	Accepts:	unsigned char*	source	- array of 64-bit values to swap
//				long			count	- # of 64-bit values to swap
//	Returns:	nothing
//----------------------------------------------------------------------------

void Swap8xN(unsigned char *source, long count)
{
	unsigned long *srcAsLong = (unsigned long*) source;
	unsigned long temp;

	while (count--)
	{
		// swap each half and then swap the halves
		temp = Swap4(srcAsLong[0]);
		srcAsLong[0] = Swap4(srcAsLong[1]);
		srcAsLong[1] = temp;

		srcAsLong += 2;
	}
} 


//----------------------------------------------------------------------------
//	SwapMBxN()
//
//	Byte-swaps N multi-byte values *in place*.  Used to convert to/from
//	little/big endian.
//
//	NOTE: the size of each element *MUST* be an even number!
//
//	Accepts:	unsigned char*	source		- array of values to swap
//				long			elemSize	- byte size of each value
//				long			numElems	- # of values to swap
//	Returns:	nothing
//----------------------------------------------------------------------------

void SwapMBxN(unsigned char *source, long elemSize, long numElems)
{
	unsigned short temp;

	// loop through all elements in array
	for (int i = 0; i < numElems; i++)
	{
		// point at the current element
		unsigned char *srcAsShort = (unsigned char*) (source + elemSize * i);
		long numShorts = elemSize / 2;

		while (numShorts > 1)
		{
			// swap first & last 16-bit words and exchange them
			temp = Swap2(srcAsShort[0]);
			srcAsShort[0] = Swap2(srcAsShort[numShorts-1]);
			srcAsShort[numShorts-1] = temp;

			srcAsShort++;
			numShorts -= 2;
		}

		if (numShorts == 1)
			*srcAsShort = Swap2(*srcAsShort);
	}
}
