/******************************************************************************************/
/*

DESCRIPTION

	CodeWarriorTRK's transport layer
	
COPYRIGHT

	Copyright 2000, 2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"		/* must be included first */

#include "transport.h"
#include "transport_tbl.h"

/****************************************************************************/
/* Index of current transport                                               */
/****************************************************************************/
#ifndef IS_SMP
int gCurTransport;
#endif

/****************************************************************************/
/* Global flag used to indicate that input from the host is waiting			*/
/****************************************************************************/
#if TRK_TRANSPORT_INT_DRIVEN
#ifndef IS_SMP
volatile unsigned char gInputPending;
#endif
#endif /* TRK_TRANSPORT_INT_DRIVEN */

/****************************************************************************/
/*                                                                          */
/* Function Name............: TransportInitialize		                    */
/*                                                                          */
/* Description..............: This function will initialize hw              */
/*                            interface. In other words, this func          */
/*                            will call initialize routine of each          */
/*                            available communication port.                 */
/*                                                                          */
/* Input....................: none                                          */
/*                                                                          */
/* Output...................: return the error from executing               */
/*                            Initialize routine                            */
/*                                                                          */
/****************************************************************************/
DSError InitializeTransport()
{
	int index;
	DSError err = kNoError;
	

	/************************************************************************/
	/* Initialize each port and channel                                     */
	/************************************************************************/
	for(index = 0; 
	    index < gTrkTransportNumCommPorts && err == kNoError; 
	    index++)
	{
		/*
		** Initialize all transport interfaces.  Each layer is 
		** responsible for initializing the layer beneath it, if
		** any; only the top layer need be initialized here.
		*/
		
		if (TRANSPORT_ACTIVE( index ))
			err = LAYER_INIT( TRANSPORT_LAYER( index ) );
		
	}
			
	/************************************************************************/
	/* Initialize global variable, gTransportCurrPort                       */
	/************************************************************************/
	gCurTransport = 0;
	TRANSPORT_USED(gCurTransport) = true;
		
	return err;
}

/****************************************************************************/
/*                                                                          */
/* Function Name............: TransportInitializeIndex		                */
/*                                                                          */
/* Description..............: Initialize comm channel w/specific table index.*/
/*                                                                          */
/* Input....................: index											*/
/*                                                                          */
/* Output...................: Error code returned by the init. func. for	*/
/*							  for that port.								*/
/*                                                                          */
/****************************************************************************/
DSError TransportInitializeIndex( int index )
{
	DSError err = kNoError;
	
	if (!TRANSPORT_ACTIVE(index))
	{
		/*
		** Initialize the specified channel.
		*/
		err = LAYER_INIT( TRANSPORT_LAYER( index ) );

		/*
		** Mark the channel active.
		*/
		if (err == kNoError)
			TRANSPORT_ACTIVE(index) = true;

	}

	return err;
}


/****************************************************************************/
/*                                                                          */
/* Function Name............: TransportTerminateIndex		                */
/*                                                                          */
/* Description..............: Terminate comm channel w/specific table index.*/
/*                                                                          */
/* Input....................: index											*/
/*                                                                          */
/* Output...................: Error code returned by the term. func. for	*/
/*							  for that port.								*/
/*                                                                          */
/****************************************************************************/
DSError TransportTerminateIndex( int index )
{
	DSError err = kNoError;
	
	if (TRANSPORT_ACTIVE(index))
	{
		/****************************************************************/		
		/* Deactivate the channel and mark it unused */
		/****************************************************************/		
		TRANSPORT_ACTIVE(index) = TRANSPORT_USED(index) = false;

		/****************************************************************/		
		/* Terminate the specified communication port                   */
		/****************************************************************/		
		err = LAYER_TERMINATE( TRANSPORT_LAYER( index ) );
	}

	return err;
}


/****************************************************************************/
/*                                                                          */
/* Function Name............: TransportAlternator				            */
/*                                                                          */
/* Description..............: Alternate through available channel           */
/*                                                                          */
/* Input....................: None                                          */
/*                                                                          */
/* Output...................: return kTransportConfigurationError           */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
void TransportAlternator()
{
	if (gTrkTransportNumCommPorts > 1)
	{
		/************************************************************************/
		/* Alternate serial polling if the number of active communication       */
		/* ports is more than 1.                                                */
		/************************************************************************/

		/************************************************************************/
		/* Set the current port unused                                          */
		/************************************************************************/
		TRANSPORT_USED(gCurTransport) = false;
		
		/************************************************************************/
		/* Search for another active port										*/
		/************************************************************************/
		do
		{
			gCurTransport++;
			if (gCurTransport == gTrkTransportNumCommPorts)
				gCurTransport = 0;
		} while (!TRANSPORT_ACTIVE( gCurTransport ));
			
			
		/************************************************************************/
		/* Mark it used                                                         */
		/************************************************************************/
		TRANSPORT_USED(gCurTransport) = true;
		
	}
}

/****************************************************************************/
/*                                                                          */
/* Function Name............: TransportSelector		 			            */
/*                                                                          */
/* Description..............: Determine through which port it should sent   */
/*                            message to host. It has default threshold     */
/*                            value and if a msg's length > threshold,      */
/*                            sent the message through the fastest port     */
/*                            available.                                    */
/*                                                                          */
/* Input....................: None                                          */
/*                                                                          */
/* Output...................: return kTransportConfigurationError           */
/*                                                                          */
/****************************************************************************/
void TransportSelector(MessageBuffer* buffer)
{
	/************************************************************************/
	/* If the current transport port is the primary port, return            */
	/************************************************************************/
	if (gCurTransport == gTrkTransportPrimaryIndex)
		return;
		
	if (   buffer->fLength > TRK_TRANSPORT_MSG_THRESHOLD 
	    && TRANSPORT_ACTIVE(gTrkTransportPrimaryIndex)
	    && !TRANSPORT_USED(gCurTransport))
	{
	    /********************************************************************/
		/* Mark current transport unused                                    */
	    /********************************************************************/
		TRANSPORT_USED(gCurTransport) = false;
		
	    /********************************************************************/
		/* Set current transport port to primary port                       */
	    /********************************************************************/
		gCurTransport = gTrkTransportPrimaryIndex;
		
	    /********************************************************************/
		/* Mark it used                                                     */
	    /********************************************************************/
		TRANSPORT_USED(gCurTransport) = true;
	}
}

/****************************************************************************/
/*
 * Function Name............: TransportIrq
 *
 * Description..............: Handle a transport IRQ event by calling the
 *							  IRQ handling routines for each active
 *							  transport channel.
 *
 * Input....................: None
 *
 * Output...................: return kTransportConfigurationError
 */
/****************************************************************************/
#if TRK_TRANSPORT_INT_DRIVEN
void TransportIrq()
{
	int i;
	
	for (i = 0; i < gTrkTransportNumCommPorts; i++)
	{
		if (TRANSPORT_ACTIVE(i))
		{
			/*
			** If an IRQ handler returns non-zero, then
			** new input has arrived.
			*/
			
			if (LAYER_IRQ( TRANSPORT_LAYER(i) ))
				gInputPending = true;
		}
	}
}
#endif /* #if TRK_TRANSPORT_INT_DRIVEN */
