/******************************************************************************************/
/*

DESCRIPTION

	Character-based transport layer header.  The character-based
	layer implements simple byte stream I/O and is intended 
	hardware communication drivers

COPYRIGHT

	Copyright 2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#ifndef __TRANSPORT_CLAYER_H__
#define __TRANSPORT_CLAYER_H__

/*
** transport_clayer.h is a very low-level header file.  It is intended
** to be included by hardware drivers *without* requiring any
** significant entanglement with the TRK sources.  All types in here
** are self-contained except the error type.  Fortunately, dserror.h
** itself is simple and self-contained, so including it is OK.
*/
#include "dserror.h"

/****************************************************************************/
/*
 * Type enum
 */
/****************************************************************************/

/* Transport layer error codes: see dserror.h */
enum {
	kTransportError = 0x0a00,	/* generic transport error */
	kTransportConfigError,
	kTransportNoData			/* returned from CharLayer read api */
};

/****************************************************************************/
/*
 * Type definitions
 */
/****************************************************************************/
typedef void LayerParams;

/****************************************************************************/
/*
 * Structure Name...........:
 *		CharLayerApi
 *
 * Description..............:
 * 		The structure contains char layer api (functional implementation).
 *
 * Fields...................:
 *
 *      The following structure fields contain function pointers
 *      to their respective functions of the specified character-
 *		based communication layer.  The required fields *must* contain
 *		valid pointers to appropriate functions.  The optional fields
 *		may be set to NULL if they are not implemented for a 
 *		particular layer.
 *
 *      init................: Initialize layer. Required.
 *      terminate...........: Terminate layer. Required.
 *      read................: Read a character, non-blocking.  Return
 *							  kTransportNoData if nothing is avaiable to
 *							  be read. Required.
 *      write...............: Write a character. Required.
 *		flush...............: Transmit any buffered output characters.
 *							  Optional.
 *      irq.................: Handle layer communication interrupt.  Return
 *							  0 if no new input arrived, 1 otherwise.  If in
 *							  doubt, return 1.
 *							  Optional.
 */
/****************************************************************************/
typedef DSError (LayerInitFunc)( LayerParams *params );
typedef DSError (LayerTerminateFunc)( LayerParams *params );
typedef DSError (LayerFlushFunc)( LayerParams *params );
typedef int (LayerIrqFunc)( LayerParams *params );
typedef DSError (CharLayerReadFunc)( LayerParams *params, char *c );
typedef DSError (CharLayerWriteFunc)( LayerParams *params, char c ); 
typedef struct CharLayerApi
{
	LayerInitFunc			*init;
	LayerTerminateFunc		*terminate;
	CharLayerReadFunc		*read;
	CharLayerWriteFunc		*write;
	LayerFlushFunc			*flush;
	LayerIrqFunc			*irq;
} CharLayerApi;

/****************************************************************************/
/*
 * Structure Name...........:
 *		CharLayer
 *
 * Description..............:
 * 		The structure contains a character-based layer instance
 *
 * Fields...................:
 *
 *		The following structure fields contain information about a specific
 *		character-based layer instance.
 *
 *      param...............: layer parameters
 *		api.................: layer api (functional implementation)
 */
/****************************************************************************/
typedef struct CharLayer
{
	LayerParams		*params;
	CharLayerApi	*api;
} CharLayer;

/****************************************************************************/
/*
 * Macros to access layer
 */
/****************************************************************************/
#define LAYER_PARAM(layer)				(layer)->param
#define LAYER_INIT(layer)				(*((layer)->api->init))((layer)->params)
#define LAYER_TERMINATE(layer)			(*((layer)->api->terminate))((layer)->params)
#define CHAR_LAYER_READ(layer,c)		(*((layer)->api->read))((layer)->params,c)
#define CHAR_LAYER_WRITE(layer,c)		(*((layer)->api->write))((layer)->params,c)
#define CHAR_LAYER_FLUSH(layer)			(((layer)->api->flush) ? (*((layer)->api->flush))((layer)->params) : 	kNoError)
#define LAYER_IRQ(layer)				(((layer)->api->irq) ? (*((layer)->api->irq))((layer)->params) : 0)
#define LAYER_SPECIFIC(layer)			(((layer)->api->specific) ? (*((layer)->api->specific))((layer)->params) : 	kNoError)

/****************************************************************************/
/*
 * Function Prototypes
 */
/****************************************************************************/
#endif /* __TRANSPORT_CLAYER_H__ */
