/******************************************************************************************/
/*
	UART.c

	Implementation file for legacy UART driver interface.  This allows
	UART driver libraries using the new interface to be backwards-compatible
	with the old interface.
	
COPYRIGHT

	Copyright 2003, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "UART.h"		/* abstract interface */


/****************************************************************************/
/*
	Global variables
*/
/****************************************************************************/
extern CharLayerApi gUartApi;
static UartParams gParams;

/****************************************************************************/
/*
	Static function declarations
*/
/****************************************************************************/

static unsigned long _strlen(const char * str);

/****************************************************************************/
/*
	Local "string length" function.
*/
/****************************************************************************/
unsigned long _strlen(const char * str)
{
	unsigned long	len = -1;
	
	unsigned char * p = (unsigned char *) str - 1;
	
	do
		len++;
	while (*++p);

	return(len);
}


/****************************************************************************/
/*
	InitializeUART
	
	Reset UART with the appropriate baud rate.
*/
/****************************************************************************/

UARTError InitializeUART(UARTBaudRate baudRate)
{
	gParams = (UartParams){ baudRate, 0, 0 };
	gUartApi.init( &gParams );
	return kUARTNoError;
}


UARTError InitializeIntDrivenUART( UARTBaudRate baudRate, 
								   unsigned char intDrivenInput,
								   unsigned char intDrivenOutput,
								   volatile unsigned char **inputPendingPtrRef )
{
	gParams = (UartParams){baudRate,intDrivenInput, intDrivenOutput };
	gUartApi.init(&gParams);
	return kUARTNoError;
}
/****************************************************************************/
/*
	TerminateUART
	
	Turn off the UART (as appropriate).
*/
/****************************************************************************/

UARTError TerminateUART(void)
{
	gUartApi.terminate( &gParams );
	return kUARTNoError;
}

/****************************************************************************/
/*
	WriteUARTFlush
*/
/****************************************************************************/

UARTError WriteUARTFlush(void)
{
	gUartApi.flush( &gParams );
	return kUARTNoError;	
}

/****************************************************************************/
/*
	ReadUARTPoll
	
	Read one character from the UART if it is available, otherwise
	return an error.
*/
/****************************************************************************/
UARTError ReadUARTPoll(char* c)
{
	DSError err;
	err = gUartApi.read( &gParams, c );
	return( (err == kNoError) ? kUARTNoError: kUARTNoData  );
}

/****************************************************************************/
/*
	ReadUART1
	
	Read one character from the UART.
*/
/****************************************************************************/

UARTError ReadUART1(char* c)
{
	DSError err;
	
	/*
	** Continue reading until something is returned.
	*/
	
	do
	{
		err = gUartApi.read( &gParams, c );
	} while (err == kTransportNoData);
	return( (err == kNoError) ? kUARTNoError: kUARTNoData  );
}

/****************************************************************************/
/*
	ReadUARTN
	
	Read N bytes from the UART.
	
	bytes			pointer to result buffer
	limit			size of buffer and # of bytes to read
*/
/****************************************************************************/
UARTError ReadUARTN(void* bytes, unsigned long limit)
{
	int count;
	UARTError err; 

	for (count = 0, err = kUARTNoError; 
		count < limit && err == kUARTNoError;
		count++)
        {
		err = ReadUART1( (char *)bytes + count );
        }

	return err;
}

/****************************************************************************/
/*
	ReadUARTString
	
	Read a string terminated by a special character into a buffer.  Do not
	include the terminate character (e.g. "0x0D") in the buffer.  Always
	return a zero-terminated buffer.
	
	Example:
	
	limit = 4, termChar = '\r'
	
	input           0  1  2  3   return value
	----------      -----------  ---------------
	a  b \r     ->  a  b \0 ??   kUARTNoError
	a  b  c \r  ->  a  b  c \0   kUARTNoError
	a  b  c  d  ->  a  b  c \0   kUARTBufferOverflow
	
	s				string (result) buffer
	limit			size of buffer
	termChar		character in the UART that terminates string entry
	ReadUARTString	kUARTBufferOverflow (and errors returned by ReadUART1)
*/
/****************************************************************************/
UARTError ReadUARTString(char* s, unsigned long limit, char termChar)
{
	UARTError err = kUARTNoError;
	
	unsigned long l = 0;			/* current length of buffer */
	char c;
		
	do {
		err = ReadUART1(&c);
		if (err)
			break;
		if (c == termChar)
			break;
		
		*s++ = c;
		l++;
	} while (l < limit);

	if (l == limit) {
		s--;						/* back up */
		if (err == kUARTNoError)	/* don't mask previous errors */
			err = kUARTBufferOverflow;
	}
	
	*s = 0;							/* always terminate string */
	return err;
}




/****************************************************************************/

UARTError WriteUART1(char c)
{
	gUartApi.write( &gParams, c );
	if (gUartApi.flush!=0){ /* zsz */
		gUartApi.flush( &gParams );
	}

	return kUARTNoError;
}

/****************************************************************************/

UARTError WriteUARTN(const void* bytes, unsigned long length)
{
	int count;
	UARTError err;

	for (count = 0, err = kUARTNoError;
		count < length && err == kUARTNoError;
		count++)
        {
		err = WriteUART1( ((char *)bytes)[ count ] );
        } 

	return err;
}

/****************************************************************************/

UARTError WriteUARTString(const char* string)
{
	unsigned long length = _strlen(string);
	
	return WriteUARTN((const void*) string, length);
}
