/******************************************************************************************/
/*

DESCRIPTION

	Target-specific implementation for debugging output.

	Development support functions for debugging output.  These functions
	can be used to write to a buffer in memory and later examined using
	a memory dump view, or by using a UART.
	
	The UART interface is defined in UART.h and linked in as a separate
	library.
	
	See the header file <usr_put.h>, which is a portable API for this file.

COPYRIGHT

	Copyright 1999-2003, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"
#include "target.h"
#include "usr_put.h"		// must be included before msgcmd.h
#include "support.h"
#include "transport_clayer.h"
#include "nubinit.h"

#include "usr_put_config.h"


void usr_put_initialize(void);
/* static void usr_put_initialize_serial(void);  TO BE DEFINED ??? */

#ifdef DEBUGIO_RAM
static void usr_put_initialize_ram(void);
static int usr_putchar_ram(int c);
static int usr_puts_ram(const char *s);
static char **mpp;			/* memory pointer */
#endif

#ifdef DEBUGIO_CUSTOM
static int usr_puts_custom(const char* s);
#endif

/* pointer to transport layer to be used for debug output */
#if defined( DEBUGIO_SERIAL ) || defined( DB_RAM_RESET_DUMP )
extern CharLayer *gUsrPutLayer;
#endif




/*
** During the process of dumping debug RAM contents to the
** console, the debug RAM must be locked so that it is not
** affected by any additional debug output that is generated.
** It is unlocked as soon as the console dump is complete.
*/
#if defined(DEBUGIO_RAM) && defined(DB_RAM_CONSOLE_DUMP)
	static bool db_ram_locked	= false;
#endif

/****************************************************************************/
/*
	usr_puts
*/
/****************************************************************************/

#ifdef DEBUGIO_CUSTOM

int usr_puts_custom(const char* s)
{
	int result = 0;
	char c;
	while ((result == 0) && (c = *s++) != 0) {
		usr_putchar_custom(c);
	}

	return result;
}

#endif /* #ifdef DEBUGIO_CUSTOM */


/****************************************************************************/
/*
	usr_putchar_serial
	
	Write one character out to serial
*/
/****************************************************************************/
extern int WriteSemiHost(char);
int usr_putchar_serial(int c)
{
	DSError err = kNoError;

#ifdef SemiHosting
	err = WriteSemiHost(c);
#else
#ifdef DEBUGIO_SERIAL
	err = CHAR_LAYER_WRITE( gUsrPutLayer, c );
	CHAR_LAYER_FLUSH( gUsrPutLayer );
#endif /* DEBUGIO_SERIAL */
#endif	

	return err;
}

/****************************************************************************/
/*
	usr_puts
*/
/****************************************************************************/

int usr_puts_serial(const char* s)
{
	int result = 0;
	char c;
	while ((result == 0) && (c = *s++) != 0) {
		result = usr_putchar_serial(c);
	}

	return result;
}


/****************************************************************************/

/****************************************************************************/
#ifdef DEBUGIO_RAM

static void usr_put_initialize_ram()
{
	char* cp;

#ifdef DB_RAM_RESET_DUMP
	char *cur = *(char **)DB_START;

	LAYER_INIT( gUsrPutLayer );

	if (DB_START <= cur && cur < DB_END)
	{
		for (cp = cur; cp < DB_END; cp++)
		{
			if (*cp != 0)
				CHAR_LAYER_WRITE( gUsrPutLayer, *cp );
		}
		for (cp = DB_START + sizeof( *mpp ); cp < cur; cp++)
		{
			if (*cp != 0)
				CHAR_LAYER_WRITE( gUsrPutLayer, *cp );
		}
		CHAR_LAYER_FLUSH( gUsrPutLayer );
	}
#endif

	/*
	** The beginning of the buffer always holds a pointer to the
	** current buffer position.  The initial buffer position
	** immediately follows that pointer.
	*/
	
	mpp = (char **)DB_START;
	*mpp = DB_START + sizeof( *mpp );

	for (cp = *mpp; cp < DB_END; cp++) {
		*cp = 0;
	}
}

/****************************************************************************/
static int usr_putchar_ram(int c)
{

#if defined(DB_RAM_CONSOLE_DUMP)
	if (!db_ram_locked)
#endif
	{
		*((*mpp)++) = c;
		if (*mpp >= (char *)(long) DB_END )
			*mpp = (char *)(long) DB_START + sizeof( mpp );
	}

	return 0;
}

/****************************************************************************/
static int usr_puts_ram(const char* p)
{
	int c;

	while ((c = *p++) != 0) {
		usr_putchar_ram(c);
	}
	
	return 0;
}

#endif	// #ifdef DEBUGIO_RAM


/****************************************************************************/
/*
	usr_putchar
	
	Initialize the serial/memory output buffer
*/
/****************************************************************************/
void usr_put_initialize()
{
#ifdef DEBUGIO_RAM
	usr_put_initialize_ram();
#endif

#ifdef DEBUGIO_SERIAL
	LAYER_INIT( gUsrPutLayer );
#endif

#ifdef DEBUGIO_CUSTOM
	usr_put_initialize_custom();
#endif
}

/****************************************************************************/
/*
	usr_putchar
	
	Put one character to serial or to memory
*/
/****************************************************************************/
int usr_putchar(int c)
{
#if !defined( DEBUGIO_RAM ) && !defined( DEBUGIO_SERIAL ) && !defined( DEBUGIO_CUSTOM )
	#if __MWERKS__
	#pragma unused( c )
	#endif /* __MWERKS__ */
#endif

	int result = 0;
#ifdef DEBUGIO_RAM
	result = usr_putchar_ram(c);
#endif

#ifdef DEBUGIO_SERIAL
	if (result == 0)
		usr_putchar_serial(c);
#endif

#ifdef DEBUGIO_CUSTOM
	if (result == 0)
		usr_putchar_custom(c);
#endif

	return result;
}

/****************************************************************************/
/*
	usr_puts
	
	Put a character string to serial or to memory.
*/
int usr_puts(const char* s)
{
#if !defined( DEBUGIO_RAM ) && !defined( DEBUGIO_SERIAL ) && !defined( DEBUGIO_CUSTOM )
	#if __MWERKS__
	#pragma unused( s )
	#endif /* __MWERKS__ */
#endif

	int result = 0;
#ifdef DEBUGIO_RAM
	result = usr_puts_ram(s);
#endif

#ifdef DEBUGIO_SERIAL
	if (result == 0)
		result = usr_puts_serial(s);
#endif

#ifdef DEBUGIO_CUSTOM
	if (result == 0)
		usr_puts_custom(s);
#endif

	return result;
}

/****************************************************************************/
#if defined(DEBUGIO_RAM) && defined(DB_RAM_CONSOLE_DUMP)

#define MAX_CONSOLE_DUMP_SIZE	250
void __do_console_dump()
{
	long total, current, index, temp;
	DSIOResult io_result;
	

	/*
	** Dump everything from the start of the RAM buffer to the
	** current position and reset the current position pointer
	** to the start.
	*/

	total = *mpp - (DB_START + sizeof( *mpp ));
	*mpp = DB_START + sizeof( *mpp );

	/*
	** Send it as multiple WriteFile requests, which should show up in the
	** debugger console window.  It attempts to send the output
	** one line at a time.
	**
	** Note: in some places where this is called, TRK is already 
	** using one message buffer, the WriteFile requires another,
	** and the reply would require a third.  By default TRK only
	** has two buffers.  When this feature is enabled, TRK must
	** have at least 3 buffers.  This is taken care of with 
	** #ifdef's in "msgbuf.h".
	*/
	
	db_ram_locked = true;

	for (index = 0; index < total; index += current)
	{

		/*
		** Get the size of the next thing to write.  Start with
		** the max dump size and work backwards to find the last
		** line ending in it.  We want a line-ending to be the
		** last character in the message because some debuggers
		** will add it anyway.  If we send random chunks of output,
		** we may get random line endings sprinkled throughout.
		** If there is no line ending, send the whole chunk.
		*/
		
		current = (MAX_CONSOLE_DUMP_SIZE < (total - index)) ?
			MAX_CONSOLE_DUMP_SIZE : (total - index);
		for (temp = current - 2; temp >= 0; --temp)
		{
			if ((*mpp)[ index + temp ] == '\n')
			{
				current = temp + 1;
				break;
			}
		}
	
		temp = current;
		SuppAccessFile( kDSStdout, (ui8 *)*mpp + index, (size_t *)&temp, &io_result, true, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
					, TRK_PROCESS_ID, TRK_THREAD_ID
#endif
					);
	}
		
	db_ram_locked = false;
}

#endif

/****************************************************************************/
/*
	__do_putchar
	
	Put one character to serial or to memory.  Don't call this function
	directly.  Use the __putchar macro instead.
*/
/****************************************************************************/
int __do_putchar(char c)
{
	return usr_putchar(c);
}

/****************************************************************************/
/*
	__do_puts
	
	Put a character string to serial or to memory.  Don't call this function
	directly.  Use the __puts macro instead.
*/
/****************************************************************************/
int __do_puts(const char* s)
{
	return usr_puts(s);
}

static const char gHexChars[] = "0123456789ABCDEF";

/****************************************************************************/
/*
	__do_puthex32
	
	Put a numeric value to serial or to memory in 32-digit hex format.
	Don't call this function directly.  Use the __puthex32 macro instead.
*/
/****************************************************************************/
int __do_puthex32(void* x)
{
	int i;
	char s[33];
	unsigned char copy[16];
	unsigned long *p;
	
	//deep copy the content of x to copy array
	for( i = 15; i >= 0 ; i-- )
		copy[i] = ((char*)x)[i];

	p = (unsigned long*)copy;
	
	for (i = 31; i >= 0; i--) {

		if (i == 23 || i == 15 || i == 7)	
			p += 1;

		s[i] = gHexChars[*p & 0xF];		/* convert binary to ASCII */
		*p >>= 4;
	}

	s[32] = 0;							/* terminating zero (cstring) */
	
	return usr_puts(s);
}


/****************************************************************************/
/*
	__do_puthex20
	
	Put a numeric value to serial or to memory in 20-digit hex format.
	Don't call this function directly.  Use the __puthex20 macro instead.
*/
/****************************************************************************/
int __do_puthex20(void *x)
{
	int i;
	int pi;
	ui8 *p;
	char s[21];
	
	p = (ui8 *)x;
	
	for (i = 19; i >= 0; i -= 2) {
		if (gBigEndian)
			pi = i >> 1;
		else
			pi = (19 - i) >> 1;
		s[i] = gHexChars[p[pi] & 0xF];		/* convert binary to ASCII */
		s[i - 1] = gHexChars[(p[pi] >> 4) & 0xF];
	}
	s[20] = 0;							/* terminating zero (cstring) */
	
	return usr_puts(s);
}

/****************************************************************************/
/*
	__do_puthex8
	
	Put a numeric value to serial or to memory in 8-digit hex format.
	Don't call this function directly.  Use the __puthex8 macro instead.
*/
/****************************************************************************/
int __do_puthex8(unsigned long x)
{
	int i;
	char s[9];
	for (i = 7; i >= 0; i--) {
		s[i] = gHexChars[x & 0xF];		/* convert binary to ASCII */
		x >>= 4;
	}
	s[8] = 0;							/* terminating zero (cstring) */
	
	return usr_puts(s);
}

/****************************************************************************/
/*
	__do_puthex16
	
	Put a numeric value to serial or to memory in 16-digit hex format.
	Don't call this function directly.  Use the __puthex16 macro instead.
*/
/****************************************************************************/

#if __MWERKS__
  #if __option( longlong )
int __do_puthex16(unsigned long long x)
	{
		int i;
		char s[17];
		for (i = 15; i >= 0; i--) {
			s[i] = gHexChars[x & 0xF];		/* convert binary to ASCII */
			x >>= 4;
		}
		s[16] = 0;							/* terminating zero (cstring) */
	
		return usr_puts(s);
	}
  #endif /* if/else __option( longlong ) */
#else /* NOT __MWERKS__ */
  #if __GNUC__
	int __do_puthex16(unsigned long long x)
	{
		int i;
		char s[17];
		for (i = 15; i >= 0; i--) {
			s[i] = gHexChars[x & 0xF];		/* convert binary to ASCII */
			x >>= 4;
		}
		s[16] = 0;							/* terminating zero (cstring) */
	
		return usr_puts(s);
	}
  #endif /* if/else __GNUC__ */
#endif /* if/else __MWERKS__ */


/****************************************************************************/
/*
	__do_puthex4
	
	Put a numeric value to serial or to memory in 4-digit hex format.
	Don't call this function directly.  Use the __puthex4 macro instead.
*/
/****************************************************************************/
int __do_puthex4(unsigned short x)
{
	int i;
	char s[5];
	for (i = 3; i >= 0; i--) {
		s[i] = gHexChars[x & 0xF];
		x >>= 4;
	}
	s[4] = 0;
	
	return usr_puts(s);
}

/****************************************************************************/
/*
	__do_puthex2
	
	Put a numeric value to serial or to memory in 2-digit hex format.
	Don't call this function directly.  Use the __puthex2 macro instead.
*/
/****************************************************************************/
int __do_puthex2(unsigned char x)
{
	int i;
	char s[3];
	for (i = 1; i >= 0; i--) {
		s[i] = gHexChars[x & 0xF];
		x >>= 4;
	}
	s[2] = 0;
	
	return usr_puts(s);
}
